#include <swift/data.h>

bool swift_init_array_size(swift_array_t *array, swift_cmp_fn_t cmp,
	size_t item_size, size_t size)
{
	if (!array || !item_size)
		return false;

	array->cmp = cmp;
	array->item_size = item_size;
	array->count = 0;
	array->size = size;

	array->data = calloc(array->size, array->item_size);

	return !!array->data;
}

bool swift_init_array(swift_array_t *array, swift_cmp_fn_t cmp,
	size_t item_size)
{
	return swift_init_array_size(array, cmp, item_size, 0);
}

void swift_term_array(swift_array_t *array)
{
	if (!array)
		return;

	if (array->data)
		SWIFT_FREE(array->data);
}

swift_array_t *swift_new_array_size(swift_cmp_fn_t cmp, size_t item_size,
	size_t size)
{
	swift_array_t *array;

	if (!item_size)
		return NULL;

	array = calloc(1, sizeof(swift_array_t));

	if (!array)
		return NULL;

	if (!swift_init_array_size(array, cmp, item_size, size))
		goto err_free_array;

	return array;

err_free_array:
	SWIFT_FREE(array);
	return NULL;
}

swift_array_t *swift_new_array(swift_cmp_fn_t cmp, size_t item_size)
{
	return swift_new_array_size(cmp, item_size, 0);
}

void swift_del_array(swift_array_t *array)
{
	if (!array)
		return;

	swift_term_array(array);
	SWIFT_FREE(array);
}

bool swift_array_resize(swift_array_t *array, size_t size)
{
	swift_array_t *tmp_array;

	if (!array)
		return false;

	if (size > 0 && size < 16)
		size = 16;

	if (array->size == size)
		return true;

	tmp_array = swift_new_array_size(array->cmp, array->item_size, size);

	if (!tmp_array)
		return false;

	tmp_array->count = SWIFT_MIN(tmp_array->size, array->count);
	memcpy(tmp_array->data, array->data, tmp_array->count *
		tmp_array->item_size);

	swift_swap(&array->data, &tmp_array->data, sizeof(void *));
	swift_swap(&array->count, &tmp_array->count, sizeof(size_t));
	swift_swap(&array->size, &tmp_array->size, sizeof(size_t));

	swift_del_array(tmp_array);

	return true;
}

bool swift_array_expand(swift_array_t *array, size_t count)
{
	if (!array)
		return false;

	if (count <= array->size)
		return true;

	return swift_array_resize(array, swift_next_pow2_size(count));
}

swift_array_iter_t swift_array_first(swift_array_t *array)
{
	if (!array || !array->count)
		return SWIFT_ARRAY_ITER(NULL, NULL, SIZE_MAX);

	return SWIFT_ARRAY_ITER(array, swift_array_get(array, 0), 0);
}

swift_array_iter_t swift_array_last(swift_array_t *array)
{
	size_t index;

	if (!array || !array->count)
		return SWIFT_ARRAY_ITER(NULL, NULL, SIZE_MAX);

	index = array->count - 1;

	return SWIFT_ARRAY_ITER(array, swift_array_get(array, index), index);
}

bool swift_array_next(swift_array_iter_t *iter)
{
	if (!iter || iter->index == SIZE_MAX)
		return false;

	iter->index = (iter->index < iter->array->count) ? (iter->index + 1) :
		SIZE_MAX;
	iter->data = swift_array_get(iter->array, iter->index);

	return !!iter->data;
}

bool swift_array_prev(swift_array_iter_t *iter)
{
	if (!iter || iter->index == SIZE_MAX)
		return false;

	iter->index = (iter->index > 0) ? (iter->index - 1) : SIZE_MAX;
	iter->data = swift_array_get(iter->array, iter->index);

	return !!iter->data;
}

bool swift_array_contains(swift_array_t *array, const void *key)
{
	swift_array_iter_t iter;

	if (!array)
		return false;

	for (iter = swift_array_first(array); iter.data;
		swift_array_next(&iter)) {
		if (array->cmp(key, iter.data) == 0)
			return true;
	}

	return false;
}

void *swift_array_get(swift_array_t *array, size_t index)
{
	if (!array || index >= array->count)
		return NULL;

	return array->data + index * array->item_size;
}

size_t swift_array_set(swift_array_t *array, size_t index, void *entry)
{
	size_t count;

	if (!array || !entry)
		return 0;

	index = SWIFT_MIN(index, array->count);
	count = SWIFT_MIN(1, array->count - index);

	if (count)
		memcpy(array->data + index * array->item_size, entry, array->item_size);

	return 1;
}

size_t swift_array_insert(swift_array_t *array, size_t index,
	void *entry)
{
	if (!array || !entry)
		return 0;

	if (!swift_array_expand(array, array->count + 1))
		return 0;

	index = SWIFT_MIN(index, array->count);

	if (index < array->count)
		memmove(array->data + (index + 1) * array->item_size,
			array->data + index * array->item_size,
				(array->count - index) * array->item_size);

	++array->count;
	swift_array_set(array, index, entry);

	return 1;
}

size_t swift_array_remove_range(swift_array_t *array, size_t index,
	size_t count)
{
	if (!array || !count)
		return 0;

	index = SWIFT_MIN(index, array->count);
	count = SWIFT_MIN(count, array->count - index);

	if (index + count < array->count)
		memmove(array->data + index * array->item_size,
			array->data + (index + count) * array->item_size,
				(array->count - (index + count)) * array->item_size);

	array->count -= count;

	return count;
}

size_t swift_array_remove_index(swift_array_t *array, size_t index)
{
	return swift_array_remove_range(array, index, 1);
}

size_t swift_array_remove(swift_array_t *array, const void *key)
{
	swift_array_iter_t iter;
	size_t ret = 0;

	if (!array)
		return 0;

	for (iter = swift_array_first(array); iter.data;
		swift_array_next(&iter)) {
		if (array->cmp(key, iter.data) == 0)
			ret += swift_array_remove_index(array, iter.index);
	}

	return ret;
}

size_t swift_array_push(swift_array_t *array, void *entry)
{
	return swift_array_insert(array, array->count, entry);
}

size_t swift_array_pop(swift_array_t *array)
{
	return swift_array_remove_index(array, array->count - 1);
}

bool swift_array_sort(swift_array_t *array)
{
	if (!array)
		return false;

	qsort(array->data, array->count, array->item_size, array->cmp);

	return true;
}
