#include <swift/data.h>

#define SWIFT_BIT_MASK(n) (1UL << (n))
#define SWIFT_BIT_INDEX(n) ((n) % SWIFT_BIT_SIZE(unsigned int))
#define SWIFT_WORD_INDEX(n) ((n) / SWIFT_BIT_SIZE(unsigned int))

bool swift_init_bitset_size(swift_bitset_t *set, size_t bit_size)
{
	if (!set)
		return false;

	set->bit_size = bit_size;
	set->size = (bit_size + SWIFT_BIT_SIZE(unsigned int) - 1) /
		SWIFT_BIT_SIZE(unsigned int);

	if (!set->size)
		return true;

	set->data = calloc(set->size, sizeof(unsigned int));

	return !!set->data;
}

bool swift_init_bitset(swift_bitset_t *set)
{
	return swift_init_bitset_size(set, 0);
}

void swift_term_bitset(swift_bitset_t *set)
{
	if (!set)
		return;

	if (set->data)
		SWIFT_FREE(set->data);
}

swift_bitset_t *swift_new_bitset_size(size_t bit_size)
{
	swift_bitset_t *set;

	set = calloc(1, sizeof(swift_bitset_t));

	if (!set)
		return NULL;

	if (!swift_init_bitset_size(set, bit_size))
		goto err_free_bitset;

	return set;

err_free_bitset:
	SWIFT_FREE(set);
	return NULL;
}

swift_bitset_t *swift_new_bitset(void)
{
	return swift_new_bitset_size(0);
}

void swift_del_bitset(swift_bitset_t *set)
{
	if (!set)
		return;

	swift_term_bitset(set);
	SWIFT_FREE(set);
}

bool swift_bitset_resize(swift_bitset_t *set, size_t bit_size)
{
	swift_bitset_t tmp_set;
	size_t size;

	if (!set)
		return false;

	size = (bit_size + SWIFT_BIT_SIZE(unsigned int) - 1) /
		SWIFT_BIT_SIZE(unsigned int);

	if (set->size == size)
		return true;

	if (!swift_init_bitset_size(&tmp_set, bit_size))
		return false;

	memcpy(tmp_set.data, set->data, tmp_set.size * sizeof(unsigned int));

	swift_swap(&set->data, &tmp_set.data, sizeof(void *));
	swift_swap(&set->bit_size, &tmp_set.bit_size, sizeof(size_t));
	swift_swap(&set->size, &tmp_set.size, sizeof(size_t));

	swift_term_bitset(&tmp_set);

	return true;
}

bool swift_bitset_expand(swift_bitset_t *set, size_t count)
{
	if (!set)
		return false;

	if (count <= set->size)
		return true;

	return swift_bitset_resize(set, swift_next_pow2_size(count));
}

size_t swift_bitset_count(swift_bitset_t *set)
{
	size_t index, ret = 0;

	if (!set)
		return 0;

	for (index = 0; index < set->size; ++index)
		ret += __builtin_popcount(set->data[index]);

	return ret;
}

size_t swift_bitset_size(swift_bitset_t *set)
{
	if (!set)
		return 0;

	return set->bit_size;
}

bool swift_is_bit_set(swift_bitset_t *set, size_t index)
{
	if (!set || index >= set->bit_size)
		return false;

	return set->data[SWIFT_WORD_INDEX(index)] & SWIFT_BIT_MASK(
		SWIFT_BIT_INDEX(index));
}

bool swift_set_bit(swift_bitset_t *set, size_t index)
{
	if (!set || index >= set->bit_size)
		return false;

	set->data[SWIFT_WORD_INDEX(index)] |= SWIFT_BIT_MASK(
		SWIFT_BIT_INDEX(index));

	return true;
}

bool swift_clear_bit(swift_bitset_t *set, size_t index)
{
	if (!set || index >= set->bit_size)
		return false;

	set->data[SWIFT_WORD_INDEX(index)] &= ~SWIFT_BIT_MASK(
		SWIFT_BIT_INDEX(index));

	return true;
}

bool swift_flip_bit(swift_bitset_t *set, size_t index)
{
	if (!set || index >= set->bit_size)
		return false;

	set->data[SWIFT_WORD_INDEX(index)] ^= SWIFT_BIT_MASK(
		SWIFT_BIT_INDEX(index));

	return true;
}

size_t swift_set_bit_range(swift_bitset_t *set, size_t from, size_t to)
{
	size_t index, count, mask_count, ret;
	unsigned int mask;

	if (!set)
		return 0;

	to = SWIFT_MIN(to, set->bit_size);
	from = SWIFT_MIN(from, to);
	ret = count = to - from;

	index = SWIFT_WORD_INDEX(from);
	mask = ~0UL << SWIFT_BIT_INDEX(from);
	mask_count = __builtin_popcount(mask);

	while (count > 0) {
		if (count < mask_count) {
			mask &= ~0UL >> (SWIFT_BIT_SIZE(unsigned int) -
				SWIFT_BIT_INDEX(to));
			mask_count = __builtin_popcount(mask);
		}

		set->data[index++] |= mask;

		count -= mask_count;
		mask = ~0UL;
		mask_count = SWIFT_BIT_SIZE(unsigned int);
	}

	return ret;
}

size_t swift_clear_bit_range(swift_bitset_t *set, size_t from, size_t to)
{
	size_t index, count, mask_count, ret;
	unsigned int mask;

	if (!set)
		return 0;

	to = SWIFT_MIN(to, set->bit_size);
	from = SWIFT_MIN(from, to);
	ret = count = to - from;

	index = SWIFT_WORD_INDEX(from);
	mask = ~0UL << SWIFT_BIT_INDEX(from);
	mask_count = __builtin_popcount(mask);

	while (count > 0) {
		if (count < mask_count) {
			mask &= ~0UL >> (SWIFT_BIT_SIZE(unsigned int) -
				SWIFT_BIT_INDEX(to));
			mask_count = __builtin_popcount(mask);
		}

		set->data[index++] &= ~mask;

		count -= mask_count;
		mask = ~0UL;
		mask_count = SWIFT_BIT_SIZE(unsigned int);
	}

	return ret;
}

size_t swift_flip_bit_range(swift_bitset_t *set, size_t from, size_t to)
{
	size_t index, count, mask_count, ret;
	unsigned int mask;

	if (!set)
		return 0;

	to = SWIFT_MIN(to, set->bit_size);
	from = SWIFT_MIN(from, to);
	ret = count = to - from;

	index = SWIFT_WORD_INDEX(from);
	mask = ~0UL << SWIFT_BIT_INDEX(from);
	mask_count = __builtin_popcount(mask);

	while (count > 0) {
		if (count < mask_count) {
			mask &= ~0UL >> (SWIFT_BIT_SIZE(unsigned int) -
				SWIFT_BIT_INDEX(to));
			mask_count = __builtin_popcount(mask);
		}

		set->data[index++] &= ~mask;

		count -= mask_count;
		mask = ~0UL;
		mask_count = SWIFT_BIT_SIZE(unsigned int);
	}

	return ret;
}

size_t swift_next_set_bit(swift_bitset_t *set, size_t from, size_t to)
{
	size_t index, count, mask_count;
	unsigned int mask, value;

	if (!set)
		return SIZE_MAX;

	to = SWIFT_MIN(to, set->bit_size);
	from = SWIFT_MIN(from, to);
	count = to - from;

	index = SWIFT_WORD_INDEX(from);
	mask = ~0UL << SWIFT_BIT_INDEX(from);
	mask_count = __builtin_popcount(mask);

	while (count > 0) {
		if (count < mask_count) {
			mask &= ~0UL >> (SWIFT_BIT_SIZE(unsigned int) -
				SWIFT_BIT_INDEX(to));
			mask_count = __builtin_popcount(mask);
		}

		value = set->data[index++] & mask;

		if (value)
			return (index - 1) * SWIFT_BIT_SIZE(unsigned int) +
				__builtin_ctz(value);

		count -= mask_count;
		mask = ~0UL;
		mask_count = SWIFT_BIT_SIZE(unsigned int);
	}

	return SIZE_MAX;
}

size_t swift_next_clear_bit(swift_bitset_t *set, size_t from, size_t to)
{
	size_t index, count, mask_count;
	unsigned int mask, value;

	if (!set)
		return SIZE_MAX;

	to = SWIFT_MIN(to, set->bit_size);
	from = SWIFT_MIN(from, to);
	count = to - from;

	index = SWIFT_WORD_INDEX(from);
	mask = ~0UL << SWIFT_BIT_INDEX(from);
	mask_count = __builtin_popcount(mask);

	while (count > 0) {
		if (count < mask_count) {
			mask &= ~0UL >> (SWIFT_BIT_SIZE(unsigned int) -
				SWIFT_BIT_INDEX(to));
			mask_count = __builtin_popcount(mask);
		}

		value = ~set->data[index++] & mask;

		if (value)
			return (index - 1) * SWIFT_BIT_SIZE(unsigned int) +
				__builtin_ctz(value);

		count -= mask_count;
		mask = ~0UL;
		mask_count = SWIFT_BIT_SIZE(unsigned int);
	}

	return SIZE_MAX;
}

size_t swift_prev_set_bit(swift_bitset_t *set, size_t from, size_t to)
{
	size_t index, count, mask_count;
	unsigned int mask, value;

	if (!set)
		return SIZE_MAX;

	to = SWIFT_MIN(to, set->bit_size);
	from = SWIFT_MIN(from, to);
	count = to - from;

	index = SWIFT_WORD_INDEX(to) + 1;
	mask = ~0UL >> (SWIFT_BIT_SIZE(unsigned int) - SWIFT_BIT_INDEX(to));
	mask_count = __builtin_popcount(mask);

	while (count > 0) {
		if (count < mask_count) {
			mask &= ~0UL << SWIFT_BIT_INDEX(from);
			mask_count = __builtin_popcount(mask);
		}

		value = set->data[--index] & mask;

		if (value)
			return (index + 1) * SWIFT_BIT_SIZE(unsigned int) -
				__builtin_clz(value) - 1;

		count -= SWIFT_MIN(mask_count, count);
		mask = ~0UL;
		mask_count = SWIFT_BIT_SIZE(unsigned int);
	}

	return SIZE_MAX;
}

size_t swift_prev_clear_bit(swift_bitset_t *set, size_t from, size_t to)
{
	size_t index, count, mask_count;
	unsigned int mask, value;

	if (!set)
		return SIZE_MAX;

	to = SWIFT_MIN(to, set->bit_size);
	from = SWIFT_MIN(from, to);
	count = to - from;

	index = SWIFT_WORD_INDEX(to) + 1;
	mask = ~0UL >> (SWIFT_BIT_SIZE(unsigned int) - SWIFT_BIT_INDEX(to));
	mask_count = __builtin_popcount(mask);

	while (count > 0) {
		if (count < mask_count) {
			mask &= ~0UL << SWIFT_BIT_INDEX(from);
			mask_count = __builtin_popcount(mask);
		}

		value = ~set->data[--index] & mask;

		if (value)
			return (index + 1) * SWIFT_BIT_SIZE(unsigned int) -
				__builtin_clz(value) - 1;

		count -= SWIFT_MIN(mask_count, count);
		mask = ~0UL;
		mask_count = SWIFT_BIT_SIZE(unsigned int);
	}

	return SIZE_MAX;
}

size_t swift_bitset_and(swift_bitset_t *dst, swift_bitset_t *src)
{
	size_t index = 0;

	if (!dst)
		return 0;

	if (!src) {
		swift_clear_bit_range(dst, 0, SIZE_MAX);
		return dst->bit_size;
	}

	swift_bitset_expand(dst, src->bit_size);

	for (index = 0; index < dst->size; ++index)
		dst->data[index] &= src->data[index];

	swift_clear_bit_range(dst, src->bit_size, SIZE_MAX);
	return dst->bit_size;
}

size_t swift_bitset_or(swift_bitset_t *dst, swift_bitset_t *src)
{
	size_t index = 0;

	if (!dst)
		return 0;

	if (!src)
		return dst->bit_size;

	swift_bitset_expand(dst, src->bit_size);

	for (index = 0; index < dst->size; ++index)
		dst->data[index] |= src->data[index];

	return dst->bit_size;
}

size_t swift_bitset_xor(swift_bitset_t *dst, swift_bitset_t *src)
{
	size_t index = 0;

	if (!dst)
		return 0;

	if (!src)
		return dst->bit_size;

	swift_bitset_expand(dst, src->bit_size);

	for (index = 0; index < dst->size; ++index)
		dst->data[index] ^= src->data[index];

	return dst->bit_size;
}


swift_bitset_iter_t swift_bitset_first_set(swift_bitset_t *set)
{
	if (!set)
		return SWIFT_BITSET_ITER(NULL, SIZE_MAX);

	return SWIFT_BITSET_ITER(set, swift_next_set_bit(set, 0, SIZE_MAX));
}

swift_bitset_iter_t swift_bitset_last_set(swift_bitset_t *set)
{
	if (!set)
		return SWIFT_BITSET_ITER(NULL, SIZE_MAX);

	return SWIFT_BITSET_ITER(set, swift_prev_set_bit(set, 0, SIZE_MAX));
}

bool swift_bitset_next_set(swift_bitset_iter_t *iter)
{
	if (!iter)
		return false;

	iter->index = swift_next_set_bit(iter->set, iter->index + 1, SIZE_MAX);

	return (iter->index != SIZE_MAX);
}

bool swift_bitset_prev_set(swift_bitset_iter_t *iter)
{
	if (!iter)
		return false;

	iter->index = swift_prev_set_bit(iter->set, 0, iter->index);

	return (iter->index != SIZE_MAX);
}

swift_bitset_iter_t swift_bitset_first_clear(swift_bitset_t *set)
{
	if (!set)
		return SWIFT_BITSET_ITER(NULL, SIZE_MAX);

	return SWIFT_BITSET_ITER(set, swift_next_clear_bit(set, 0, SIZE_MAX));
}

swift_bitset_iter_t swift_bitset_last_clear(swift_bitset_t *set)
{
	if (!set)
		return SWIFT_BITSET_ITER(NULL, SIZE_MAX);

	return SWIFT_BITSET_ITER(set, swift_prev_clear_bit(set, 0, SIZE_MAX));
}

bool swift_bitset_next_clear(swift_bitset_iter_t *iter)
{
	if (!iter)
		return false;

	iter->index = swift_next_clear_bit(iter->set, iter->index + 1, SIZE_MAX);

	return (iter->index != SIZE_MAX);
}

bool swift_bitset_prev_clear(swift_bitset_iter_t *iter)
{
	if (!iter)
		return false;

	iter->index = swift_prev_clear_bit(iter->set, 0, iter->index);

	return (iter->index != SIZE_MAX);
}
