#include <swift/data.h>

bool swift_bst_init_tree(swift_bst_tree_t *tree, swift_bst_cmp_fn_t cmp)
{
	if (!tree || !cmp)
		return false;

	tree->cmp = cmp;
	tree->root = NULL;

	return true;
}

void swift_bst_term_tree(swift_bst_tree_t *tree)
{
	if (!tree)
		return;

	tree->cmp = NULL;
	tree->root = NULL;
}

swift_bst_tree_t *swift_bst_new_tree(swift_bst_cmp_fn_t cmp)
{
	swift_bst_tree_t *tree;

	if (!cmp)
		return NULL;

	tree = malloc(sizeof(swift_bst_tree_t));

	if (!tree)
		return NULL;

	if (!swift_bst_init_tree(tree, cmp))
		goto err_free_tree;

	return tree;

err_free_tree:
	SWIFT_FREE(tree);
	return NULL;
}

void swift_bst_del_tree(swift_bst_tree_t *tree)
{
	if (!tree)
		return;

	swift_bst_term_tree(tree);
	SWIFT_FREE(tree);
}

swift_bst_node_t *swift_bst_outermost(swift_bst_node_t *node, int dir)
{
	if (!node)
		return NULL;

	while (node->child[dir])
		node = node->child[dir];

	return node;
}

swift_bst_iter_t swift_bst_first(swift_bst_tree_t *tree)
{
	if (!tree)
		return SWIFT_BST_ITER(NULL);

	return SWIFT_BST_ITER(swift_bst_outermost(tree->root,
		SWIFT_BST_DIR_LEFT));
}

swift_bst_iter_t swift_bst_last(swift_bst_tree_t *tree)
{
	if (!tree)
		return SWIFT_BST_ITER(NULL);

	return SWIFT_BST_ITER(swift_bst_outermost(tree->root,
		SWIFT_BST_DIR_RIGHT));
}

swift_bst_node_t *swift_bst_successor(swift_bst_node_t *node, int dir)
{
	swift_bst_node_t *parent;

	if (!node)
		return NULL;

	dir = !!dir;

	if (node->child[dir])
		return swift_bst_outermost(node->child[dir], !dir);

	parent = node->parent;

	while (parent && parent->child[dir] == node) {
		node = parent;
		parent = node->parent;
	}

	return parent;
}

bool swift_bst_next(swift_bst_iter_t *iter)
{
	if (!iter || !iter->node)
		return false;

	iter->node = swift_bst_successor(iter->node, SWIFT_BST_DIR_RIGHT);

	return !!iter->node;
}

bool swift_bst_prev(swift_bst_iter_t *iter)
{
	if (!iter || !iter->node)
		return false;

	iter->node = swift_bst_successor(iter->node, SWIFT_BST_DIR_LEFT);

	return !!iter->node;
}

void swift_bst_rotate(swift_bst_tree_t *tree, swift_bst_node_t *root,
	size_t dir)
{
	swift_bst_node_t *child, *parent;
	size_t child_dir;

	child = root->child[!dir];
	parent = root->parent;

	if (!parent) {
		tree->root = child;
	} else {
		child_dir = (parent->child[SWIFT_BST_DIR_RIGHT] == root);
		parent->child[child_dir] = child;
	}

	child->parent = parent;
	root->parent = child;
	root->child[!dir] = child->child[dir];

	if (root->child[!dir])
		root->child[!dir]->parent = root;

	child->child[dir] = root;
}

swift_bst_node_t *swift_bst_lookup(swift_bst_tree_t *tree,
	swift_bst_node_t *key, size_t *dir, swift_bst_node_t **parent)
{
	swift_bst_node_t *node = tree->root;
	int res;

	*parent = NULL;

	while (node) {
		res = tree->cmp(tree, key, node);

		if (res == 0)
			return node;

		*parent = node;
		*dir = (res > 0);
		node = node->child[*dir];
	}

	return NULL;
}

swift_bst_node_t *swift_bst_search(swift_bst_tree_t *tree,
	swift_bst_node_t *key)
{
	swift_bst_node_t *parent;
	size_t dir;

	return swift_bst_lookup(tree, key, &dir, &parent);
}

swift_bst_node_t *swift_bst_insert(swift_bst_tree_t *tree,
	swift_bst_node_t *node)
{
	swift_bst_node_t *parent;
	size_t dir;

	if (!tree || !node)
		return NULL;

	memset(node, 0, sizeof(swift_bst_node_t));

	if (swift_bst_lookup(tree, node, &dir, &parent))
		return NULL;

	*(parent ? &parent->child[dir] : &tree->root) = node;
	node->parent = parent;

	return node;
}

swift_bst_node_t *swift_bst_remove(swift_bst_tree_t *tree,
	swift_bst_node_t *key)
{
	swift_bst_node_t *parent, *node, *child;
	size_t dir, child_dir;

	node = swift_bst_lookup(tree, key, &dir, &parent);

	if (!node)
		return NULL;

	if (node->child[SWIFT_BST_DIR_LEFT] &&
		node->child[SWIFT_BST_DIR_RIGHT]) {
		child = node->child[SWIFT_BST_DIR_LEFT];
		child = swift_bst_outermost(child, SWIFT_BST_DIR_RIGHT);
		child = swift_bst_remove(tree, child);

		for (child_dir = SWIFT_BST_DIR_LEFT;
			child_dir <= SWIFT_BST_DIR_RIGHT; ++child_dir) {
			child->child[child_dir] = node->child[child_dir];

			if (node->child[child_dir])
				node->child[child_dir]->parent = child;
		}
	} else {
		child_dir = !node->child[SWIFT_BST_DIR_LEFT];
		child = node->child[child_dir];
	}

	*(parent ? &parent->child[dir] : &tree->root) = child;

	if (child)
		child->parent = parent;

	memset(node, 0, sizeof(swift_bst_node_t));

	return node;
}

swift_bst_node_t *swift_bst_shift(swift_bst_tree_t *tree)
{
	swift_bst_iter_t iter;

	if (!tree)
		return NULL;

	iter = swift_bst_first(tree);

	if (!iter.node)
		return NULL;

	swift_bst_remove(tree, iter.node);

	return iter.node;
}

swift_bst_node_t *swift_bst_pop(swift_bst_tree_t *tree)
{
	swift_bst_iter_t iter;

	if (!tree)
		return NULL;

	iter = swift_bst_last(tree);

	if (!iter.node)
		return NULL;

	swift_bst_remove(tree, iter.node);

	return iter.node;
}
