#include <swift/data.h>

bool swift_init_heap_size(swift_heap_t *heap, swift_cmp_fn_t cmp,
	bool heap_type, size_t item_size, size_t size)
{
	if (!heap || !item_size)
		return false;

	if (!swift_init_array_size(&heap->array, cmp, item_size, size))
		return false;

	heap->cmp = cmp;
	heap->heap_type = heap_type;

	return true;
}

bool swift_init_heap(swift_heap_t *heap, swift_cmp_fn_t cmp,
	bool heap_type, size_t item_size)
{
	return swift_init_heap_size(heap, cmp, heap_type, item_size, 0);
}

void swift_term_heap(swift_heap_t *heap)
{
	if (!heap)
		return;

	swift_term_array(&heap->array);
}

swift_heap_t *swift_new_heap_size(swift_cmp_fn_t cmp, bool heap_type,
	size_t item_size, size_t size)
{
	swift_heap_t *heap;

	if (!item_size)
		return NULL;

	heap = calloc(1, sizeof(swift_heap_t));

	if (!heap)
		return NULL;

	if (!swift_init_heap_size(heap, cmp, heap_type, item_size, size))
		goto err_free_heap;

	return heap;

err_free_heap:
	SWIFT_FREE(heap);
	return NULL;
}

swift_heap_t *swift_new_heap(swift_cmp_fn_t cmp, bool heap_type,
	size_t item_size)
{
	return swift_new_heap_size(cmp, heap_type, item_size, 0);
}

void swift_del_heap(swift_heap_t *heap)
{
	if (!heap)
		return;

	swift_term_heap(heap);
	SWIFT_FREE(heap);
}

bool swift_heap_resize(swift_heap_t *heap, size_t size)
{
	if (!heap)
		return false;

	return swift_array_resize(&heap->array, size);
}

bool swift_heap_expand(swift_heap_t *heap, size_t count)
{
	if (!heap)
		return false;

	if (count <= heap->array.size)
		return true;

	return swift_heap_resize(heap, swift_next_pow2_size(count));
}

void *swift_heap_get(swift_heap_t *heap)
{
	return swift_array_get(&heap->array, 0);
}

size_t swift_heap_push(swift_heap_t *heap, void *entry)
{
	void *parent;
	size_t parent_index, index;

	if (!heap || !entry)
		return 0;

	if (!swift_heap_expand(heap, heap->array.count + 1))
		return 0;

	index = heap->array.count++;
	parent_index = index / 2;

	while (index > 0) {
		parent = swift_array_get(&heap->array, parent_index);

		if (heap->cmp(entry, parent) * heap->heap_type > 0)
			break;

		swift_array_set(&heap->array, index, parent);
		index = parent_index;
		parent_index /= 2;
	}

	swift_array_set(&heap->array, index, entry);

	return 1;
}

size_t swift_heap_pop(swift_heap_t *heap)
{
	void *left, *right, *best, *entry;
	size_t left_index, right_index, best_index, index = 0;

	if (!heap || !heap->array.count)
		return 0;

	entry = swift_array_get(&heap->array, heap->array.count - 1);
	swift_array_set(&heap->array, index, entry);
	swift_array_pop(&heap->array);

	while (index < heap->array.count) {
		best_index = index;
		best = swift_array_get(&heap->array, best_index);

		left_index = 2 * index + 1;
		left = swift_array_get(&heap->array, left_index);

		if (left && heap->cmp(best, left) * heap->heap_type > 0) {
			best_index = left_index;
			best = left;
		}

		right_index = left_index + 1;
		right = swift_array_get(&heap->array, right_index);

		if (right && heap->cmp(best, right) * heap->heap_type > 0) {
			best_index = right_index;
			best = right;
		}

		if (index == best_index)
			break;

		swift_array_set(&heap->array, index, best);
		swift_array_set(&heap->array, best_index, entry);

		index = best_index;
	}

	return 1;
}
