#include <swift/data.h>

bool swift_init_htable_size(swift_htable_t *htable, swift_hash_fn_t hash,
	swift_eq_fn_t eq, size_t item_size, size_t size)
{
	if (!htable || !hash || !eq || !item_size)
		return false;

	htable->hash = hash;
	htable->eq = eq;
	htable->item_size = item_size;
	htable->count = 0;
	htable->size = size;

	if (!htable->size)
		return true;

	if (!swift_init_bitset_size(&htable->set, size))
		return false;

	htable->data = calloc(htable->size, htable->item_size);

	if (!htable->data)
		goto err_del_bitset;

	return true;

err_del_bitset:
	swift_term_bitset(&htable->set);
	return false;
}

bool swift_init_htable(swift_htable_t *htable, swift_hash_fn_t hash,
	swift_eq_fn_t eq, size_t item_size)
{
	return swift_init_htable_size(htable, hash, eq, item_size, 0);
}

void swift_term_htable(swift_htable_t *htable)
{
	if (!htable)
		return;

	if (!htable->data)
		SWIFT_FREE(htable->data);

	swift_term_bitset(&htable->set);
}

swift_htable_t *swift_new_htable_size(swift_hash_fn_t hash,
	swift_eq_fn_t eq, size_t item_size, size_t size)
{
	swift_htable_t *htable;

	if (!hash || !eq || !item_size)
		return NULL;

	htable = calloc(1, sizeof(swift_htable_t));

	if (!htable)
		return NULL;

	if (!swift_init_htable_size(htable, hash, eq, item_size, size))
		goto err_free_htable;

	return htable;

err_free_htable:
	SWIFT_FREE(htable);
	return NULL;
}

swift_htable_t *swift_new_htable(swift_hash_fn_t hash,
	swift_eq_fn_t eq, size_t item_size)
{
	return swift_new_htable_size(hash, eq, item_size, 0);
}

void swift_del_htable(swift_htable_t *htable)
{
	if (!htable)
		return;

	swift_term_htable(htable);
	SWIFT_FREE(htable);
}

bool swift_htable_resize(swift_htable_t *htable, size_t size)
{
	swift_htable_t tmp_htable;
	swift_htable_iter_t iter;

	if (!htable)
		return false;

	if (size > 0 && size < 16)
		size = 16;

	if (htable->size == size)
		return true;

	if (!swift_init_htable_size(&tmp_htable, htable->hash, htable->eq,
		htable->item_size, size))
		return false;

	for (iter = swift_htable_first(htable); iter.data;
		swift_htable_next(&iter)) {
		swift_htable_insert(&tmp_htable, iter.data);
	}

	swift_swap(&htable->data, &tmp_htable.data, sizeof(void *));
	swift_swap(&htable->set, &tmp_htable.set, sizeof(swift_bitset_t));
	swift_swap(&htable->count, &tmp_htable.count, sizeof(size_t));
	swift_swap(&htable->size, &tmp_htable.size, sizeof(size_t));
	swift_swap(&htable->item_size, &tmp_htable.item_size, sizeof(size_t));

	swift_term_htable(&tmp_htable);

	return true;
}

bool swift_htable_expand(swift_htable_t *htable, size_t count)
{
	if (!htable)
		return false;

	if (4 * count < 3 * htable->size)
		return true;

	return swift_htable_resize(htable, swift_next_pow2_size(count));
}

size_t swift_htable_count(swift_htable_t *htable)
{
	if (!htable)
		return 0;

	return htable->count;
}

size_t swift_htable_size(swift_htable_t *htable)
{
	if (!htable)
		return 0;

	return htable->size;
}

void *swift_htable_get(swift_htable_t *htable, size_t index)
{
	if (!htable || !swift_is_bit_set(&htable->set, index))
		return NULL;

	return htable->data + index * htable->item_size;
}

size_t swift_htable_set(swift_htable_t *htable, size_t index,
	void *entry)
{
	if (!htable || index >= htable->size)
		return 0;

	if (!entry)
		return swift_clear_bit(&htable->set, index);

	memcpy(htable->data + index * htable->item_size, entry, htable->item_size);

	return swift_set_bit(&htable->set, index);
}

size_t swift_htable_next_index(swift_htable_t *htable, size_t index)
{
	if (!htable || index >= htable->size)
		return SIZE_MAX;

	return swift_next_set_bit(&htable->set, index, SIZE_MAX);
}

swift_htable_iter_t swift_htable_first(swift_htable_t *htable)
{
	size_t index;

	if (!htable)
		return SWIFT_HTABLE_ITER(.index = SIZE_MAX);

	index = swift_htable_next_index(htable, 0);

	return SWIFT_HTABLE_ITER(htable, swift_htable_get(htable, index),
		index);
}

bool swift_htable_next(swift_htable_iter_t *iter)
{
	if (!iter)
		return false;

	iter->index = swift_htable_next_index(iter->htable, iter->index + 1);
	iter->data = swift_htable_get(iter->htable, iter->index);

	return (iter->index != SIZE_MAX);
}

bool swift_htable_contains(swift_htable_t *htable, const void *key)
{
	return (swift_htable_search_index(htable, key) != SIZE_MAX);
}

size_t swift_htable_search_index(swift_htable_t *htable, const void *key)
{
	void *entry;
	size_t hash, index;

	if (!htable || !htable->size || !key)
		return SIZE_MAX;

	hash = htable->hash(key) % htable->size;
	index = hash;

	while (swift_is_bit_set(&htable->set, index)) {
		entry = swift_htable_get(htable, index);

		if (htable->eq(key, entry))
			return index;

		index = (index + 1) % htable->size;

		if (index == hash)
			return SIZE_MAX;
	}

	return index;
}

void *swift_htable_search(swift_htable_t *htable, const void *key)
{
	return swift_htable_get(htable, swift_htable_search_index(htable, key));
}

size_t swift_htable_insert(swift_htable_t *htable, void *entry)
{
	size_t index;

	if (!swift_htable_expand(htable, htable->count + 1))
		return SIZE_MAX;

	index = swift_htable_search_index(htable, entry);

	if (index == SIZE_MAX)
		return SIZE_MAX;

	swift_htable_set(htable, index, entry);
	++htable->count;

	return index;
}

bool swift_htable_reorganise(swift_htable_t *htable, size_t index)
{
	void *entry;
	size_t empty_index, hash, new_dist, old_dist;

	if (!htable || index >= htable->size)
		return false;

	empty_index = index;
	index = (index + 1) % htable->size;

	while (swift_is_bit_set(&htable->set, index)) {
		entry = swift_htable_get(htable, index);
		hash = htable->hash(entry);

		new_dist = (htable->size + empty_index - hash) % htable->size;
		old_dist = (htable->size + index - hash) % htable->size;

		if (old_dist > new_dist) {
			swift_htable_set(htable, empty_index, entry);
			empty_index = index;
		}

		index = (index + 1) % htable->size;

		if (index == hash)
			break;
	}

	return true;
}

bool swift_htable_remove_index(swift_htable_t *htable, size_t index)
{
	if (!swift_htable_set(htable, index, NULL))
		return false;

	--htable->count;

	return swift_htable_reorganise(htable, index);
}

bool swift_htable_remove(swift_htable_t *htable, const void *key)
{
	size_t index = swift_htable_search_index(htable, key);

	if (index == SIZE_MAX)
		return false;

	return swift_htable_remove_index(htable, index);
}
