#include <swift/data.h>

static int swift_bst_cmp_rb(swift_bst_tree_t *tree_,
	swift_bst_node_t *lhs_, swift_bst_node_t *rhs_)
{
	swift_rb_tree_t *tree = SWIFT_RB_TREE(tree_);
	swift_rb_node_t *lhs = SWIFT_RB_NODE(lhs_);
	swift_rb_node_t *rhs = SWIFT_RB_NODE(rhs_);

	return tree->cmp(tree, lhs, rhs);
}

bool swift_rb_init_tree(swift_rb_tree_t *tree, swift_rb_cmp_fn_t cmp)
{
	if (!tree || !cmp)
		return false;

	if (!swift_bst_init_tree(&tree->bst_tree, swift_bst_cmp_rb))
		return false;

	tree->cmp = cmp;

	return true;
}

void swift_rb_term_tree(swift_rb_tree_t *tree)
{
	if (!tree)
		return;

	tree->cmp = NULL;

	swift_bst_term_tree(&tree->bst_tree);
}

swift_rb_tree_t *swift_rb_new_tree(swift_rb_cmp_fn_t cmp)
{
	swift_rb_tree_t *tree;

	if (!cmp)
		return NULL;

	tree = malloc(sizeof(swift_rb_tree_t));

	if (!tree)
		return NULL;

	if (!swift_rb_init_tree(tree, cmp))
		goto err_free_tree;

	return tree;

err_free_tree:
	SWIFT_FREE(tree);
	return NULL;
}

void swift_rb_del_tree(swift_rb_tree_t *tree)
{
	if (!tree)
		return;

	swift_rb_term_tree(tree);
	SWIFT_FREE(tree);
}

swift_rb_node_t *swift_rb_get_root(swift_rb_tree_t *tree)
{
	if (!tree)
		return NULL;

	return SWIFT_RB_NODE(tree->bst_tree.root);
}

swift_rb_node_t *swift_rb_get_parent(swift_rb_node_t *node)
{
	if (!node)
		return NULL;

	return SWIFT_RB_NODE(node->bst_node.parent);
}

swift_rb_node_t *swift_rb_get_child(swift_rb_node_t *node,
	size_t dir)
{
	if (!node)
		return NULL;

	return SWIFT_RB_NODE(node->bst_node.child[dir]);
}

swift_rb_node_t *swift_rb_outermost(swift_rb_node_t *node,
	int dir)
{
	if (!node)
		return NULL;

	return SWIFT_RB_NODE(swift_bst_outermost(&node->bst_node, dir));
}

swift_rb_iter_t swift_rb_first(swift_rb_tree_t *tree)
{
	if (!tree)
		return SWIFT_RB_ITER(NULL);

	return SWIFT_RB_ITER(swift_rb_outermost(swift_rb_get_root(tree),
		SWIFT_BST_DIR_LEFT));
}

swift_rb_iter_t swift_rb_last(swift_rb_tree_t *tree)
{
	if (!tree)
		return SWIFT_RB_ITER(NULL);

	return SWIFT_RB_ITER(swift_rb_outermost(swift_rb_get_root(tree),
		SWIFT_BST_DIR_RIGHT));
}

swift_rb_node_t *swift_rb_successor(
	swift_rb_node_t *node, int dir)
{
	if (!node)
		return NULL;

	return SWIFT_RB_NODE(swift_bst_successor(&node->bst_node, dir));
}

bool swift_rb_next(swift_rb_iter_t *iter)
{
	if (!iter || !iter->node)
		return false;

	iter->node = swift_rb_successor(iter->node, SWIFT_BST_DIR_RIGHT);

	return !!iter->node;
}

bool swift_rb_prev(swift_rb_iter_t *iter)
{
	if (!iter || !iter->node)
		return false;

	iter->node = swift_rb_successor(iter->node, SWIFT_BST_DIR_LEFT);

	return !!iter->node;
}

void swift_rb_rotate(swift_rb_tree_t *tree,
	swift_rb_node_t *root, size_t dir)
{
	if (!tree || !root)
		return;

	return swift_bst_rotate(&tree->bst_tree, &root->bst_node, dir);
}

swift_rb_node_t *swift_rb_search(swift_rb_tree_t *tree,
	swift_rb_node_t *key)
{
	if (!tree || !key)
		return NULL;

	return SWIFT_RB_NODE(swift_bst_search(&tree->bst_tree,
		&key->bst_node));
}

swift_rb_node_t *swift_rb_insert(swift_rb_tree_t *tree,
	swift_rb_node_t *node)
{
	swift_rb_node_t *root, *parent, *grand_parent, *uncle;
	size_t dir;

	if (!swift_bst_insert(&tree->bst_tree, &node->bst_node))
		return NULL;

	node->is_red = true;

	while ((parent = swift_rb_get_parent(node)) && parent->is_red) {
		grand_parent = swift_rb_get_parent(parent);

		dir = (swift_rb_get_child(grand_parent, SWIFT_BST_DIR_LEFT) ==
			parent);
		uncle = swift_rb_get_child(grand_parent, dir);

		if (uncle && uncle->is_red) {
			parent->is_red = false;
			uncle->is_red = false;
			grand_parent->is_red = true;
			node = grand_parent;

			continue;
		}

		if (swift_rb_get_child(parent, dir) == node) {
			swift_rb_rotate(tree, parent, !dir);
			node = parent;
			parent = swift_rb_get_parent(node);
		}

		parent->is_red = false;
		grand_parent->is_red = true;

		swift_rb_rotate(tree, grand_parent, dir);
	}

	root = swift_rb_get_root(tree);
	root->is_red = false;

	return node;
}

swift_rb_node_t *swift_rb_remove(swift_rb_tree_t *tree,
	swift_rb_node_t *key)
{
	swift_rb_node_t *node, *parent, *child, *sibling, *other, *ret;
	size_t dir, child_dir;

	if (!tree || !key)
		return NULL;

	node = swift_rb_search(tree, key);
	ret = node;

	if (!node)
		return NULL;

	child_dir = !swift_rb_get_child(node, SWIFT_BST_DIR_LEFT);
	child = swift_rb_get_child(node, child_dir);
	other = swift_rb_get_child(node, !child_dir);

	swift_bst_remove(&tree->bst_tree, &node->bst_node);

	if (child && other)
		return ret;

	if (node->is_red)
		return ret;

	node = child;

	if (node && node->is_red) {
		node->is_red = false;
		return ret;
	}

	do {
		parent = swift_rb_get_parent(node);

		if (!parent)
			break;

		dir = (swift_rb_get_child(parent, SWIFT_BST_DIR_LEFT) == node);
		sibling = swift_rb_get_child(parent, dir);

		if (sibling->is_red) {
			parent->is_red = true;
			sibling->is_red = false;
			swift_rb_rotate(tree, parent, !dir);
			sibling = swift_rb_get_child(parent, dir);
		}

		child = swift_rb_get_child(sibling, dir);
		other = swift_rb_get_child(sibling, !dir);

		if ((!child || !child->is_red) && (!other || !other->is_red)) {
			sibling->is_red = true;
			node = parent;
			parent = swift_rb_get_parent(parent);
			continue;
		}

		if (!child || !child->is_red) {
			other->is_red = false;
			sibling->is_red = true;
			swift_rb_rotate(tree, parent, dir);
			sibling = swift_rb_get_child(parent, dir);
		}

		child = swift_rb_get_child(sibling, dir);

		sibling->is_red = parent->is_red;
		parent->is_red = false;
		child->is_red = false;
		swift_rb_rotate(tree, parent, !dir);
		node = swift_rb_get_root(tree);
		break;
	} while (!node->is_red);

	if (node)
		node->is_red = false;

	return ret;
}

swift_rb_node_t *swift_rb_shift(swift_rb_tree_t *tree)
{
	swift_rb_iter_t iter;

	if (!tree)
		return NULL;

	iter = swift_rb_first(tree);

	if (!iter.node)
		return NULL;

	swift_rb_remove(tree, iter.node);

	return iter.node;
}

swift_rb_node_t *swift_rb_pop(swift_rb_tree_t *tree)
{
	swift_rb_iter_t iter;

	if (!tree)
		return NULL;

	iter = swift_rb_last(tree);

	if (!iter.node)
		return NULL;

	swift_rb_remove(tree, iter.node);

	return iter.node;
}
