#include <swift/data.h>
#include <swift/event.h>
#include <swift/thread.h>

#include "event.h"
#include "loop.h"
#include "source.h"

extern swift_ev_ctx_t *ev_ctx;

static bool swift_add_ev_loop(swift_ev_loop_t *loop)
{
	size_t ret;

	if (!loop)
		return false;

	swift_mutex_lock(ev_ctx->lock);
	ret = swift_htable_insert(ev_ctx->loops, &loop);
	swift_mutex_unlock(ev_ctx->lock);

	return (ret != SIZE_MAX);
}

static void swift_remove_ev_loop(swift_ev_loop_t *loop)
{
	if (!loop)
		return;

	swift_mutex_lock(ev_ctx->lock);
	swift_htable_remove(ev_ctx->loops, &loop);
	swift_mutex_unlock(ev_ctx->lock);
}

size_t swift_hash_ev_loop(const void *data)
{
	swift_ev_loop_t *loop = *(swift_ev_loop_t **)data;

	return swift_fnv1a_hash64(&loop->thread_id, sizeof(swift_thread_id_t));
}

bool swift_eq_ev_loops(const void *lhs_, const void *rhs_)
{
	swift_ev_loop_t *lhs = *(swift_ev_loop_t **)lhs_;
	swift_ev_loop_t *rhs = *(swift_ev_loop_t **)rhs_;

	return (lhs->thread_id == rhs->thread_id);
}

swift_ev_loop_t *swift_new_ev_loop(swift_thread_id_t thread_id)
{
	swift_htable_iter_t iter;
	swift_ev_loop_t *loop = calloc(1, sizeof(swift_ev_loop_t));;
	swift_ev_src_t *src;
	bool is_main_thread, add_src;

	if (!loop)
		return NULL;

	swift_init_list(&loop->msgs);
	loop->thread_id = thread_id;
	loop->srcs = swift_new_htable(swift_hash_ev_src, swift_eq_ev_srcs,
		sizeof(swift_ev_src_t));

	if (!loop->srcs)
		goto err_free_loop;

	swift_mutex_lock(ev_ctx->lock);
	is_main_thread = (thread_id == swift_get_main_thread_id());

	for (iter = swift_htable_first(ev_ctx->srcs); iter.data;
		swift_htable_next(&iter)) {
		src = *(swift_ev_src_t **)iter.data;

		add_src = (is_main_thread) ?
			(src->flags & SWIFT_MAIN_THREAD) :
			(src->flags & SWIFT_WORKER_THREAD);

		if (!add_src)
			continue;

		src = swift_ref_ev_src(src);

		if (swift_htable_insert(loop->srcs, &src) == SIZE_MAX)
			goto err_remove_src;
	}

	swift_mutex_unlock(ev_ctx->lock);

	return loop;

err_remove_src:
	for (iter = swift_htable_first(ev_ctx->srcs); iter.data;
		swift_htable_next(&iter)) {
		src = *(swift_ev_src_t **)iter.data;
		swift_htable_remove_index(loop->srcs, iter.index);
		swift_del_ev_src(src);
	}

	swift_mutex_unlock(ev_ctx->lock);
err_free_loop:
	SWIFT_FREE(loop);
	return NULL;
}

void swift_del_ev_loop(swift_ev_loop_t *loop)
{
	swift_htable_iter_t iter;
	swift_ev_src_t *src;

	if (!loop)
		return;

	for (iter = swift_htable_first(loop->srcs); iter.data;
		swift_htable_next(&iter)) {
		src = *(swift_ev_src_t **)iter.data;
		swift_htable_remove_index(loop->srcs, iter.index);
		swift_del_ev_src(src);
	}

	swift_del_htable(loop->srcs);
	swift_remove_ev_loop(loop);
	SWIFT_FREE(loop);
}

swift_ev_loop_t *swift_get_ev_loop(swift_thread_id_t thread_id)
{
	swift_ev_loop_t key_ = {.thread_id = thread_id,};
	swift_ev_loop_t *key = &key_;
	void *ret;

	swift_mutex_lock(ev_ctx->lock);
	ret = swift_htable_search(ev_ctx->loops, &key);
	swift_mutex_unlock(ev_ctx->lock);

	return ret ? *(swift_ev_loop_t **)ret : NULL;
}

swift_ev_loop_t *swift_get_cur_ev_loop(void)
{
	swift_thread_id_t thread_id = swift_get_current_thread_id();
	swift_ev_loop_t *loop = swift_get_ev_loop(thread_id);

	if (loop)
		return loop;

	if (!(loop = swift_new_ev_loop(thread_id)))
		return NULL;

	if (!swift_add_ev_loop(loop))
		goto err_del_loop;

	return loop;

err_del_loop:
	swift_del_ev_loop(loop);
	return NULL;
}

bool swift_init_main_ev_loop(void)
{
	return !!swift_get_cur_ev_loop();
}

bool swift_run_ev_loop_once(void)
{
	swift_htable_iter_t iter;
	swift_ev_loop_t *loop = swift_get_cur_ev_loop();
	swift_ev_src_t *src;

	if (!loop)
		return false;

	for (iter = swift_htable_first(loop->srcs); iter.data;
		swift_htable_next(&iter)) {
		src = *(swift_ev_src_t **)iter.data;

		src->poll_ev(src->user_data);
	}

	return ev_ctx->is_running;
}

void swift_run_ev_loop(void)
{
	swift_mutex_lock(ev_ctx->lock);
	ev_ctx->is_running = true;
	swift_mutex_unlock(ev_ctx->lock);

	while (swift_run_ev_loop_once())
		swift_yield_thread();
}

void swift_quit_ev_loop(void)
{
	swift_mutex_lock(ev_ctx->lock);
	ev_ctx->is_running = false;
	swift_mutex_unlock(ev_ctx->lock);
}
