#include <swift/atomic.h>
#include <swift/data.h>
#include <swift/event.h>
#include <swift/thread.h>

#include "event.h"
#include "loop.h"
#include "source.h"

extern swift_ev_ctx_t *ev_ctx;

static void swift_do_del_ev_src(swift_ref_t *ref)
{
	swift_ev_src_t *src;

	if (!ref)
		return;

	src = SWIFT_CONTAINER_OF(ref, swift_ev_src_t, ref);
	SWIFT_FREE(src);
}

static swift_ev_src_t *swift_get_ev_src(swift_poll_ev_fn_t poll_ev)
{
	swift_ev_src_t key_ = {.poll_ev = poll_ev,};
	swift_ev_src_t *key = &key_;
	void *ret;

	swift_mutex_lock(ev_ctx->lock);
	ret = swift_htable_search(ev_ctx->srcs, &key);
	swift_mutex_unlock(ev_ctx->lock);

	return ret ? *(swift_ev_src_t **)ret : NULL;
}

size_t swift_hash_ev_src(const void *data)
{
	const swift_ev_src_t *src = *(const swift_ev_src_t **)data;

	return swift_fnv1a_hash64(&src->poll_ev, sizeof(swift_poll_ev_fn_t));
}

bool swift_eq_ev_srcs(const void *lhs_, const void *rhs_)
{
	const swift_ev_src_t *lhs = *(const swift_ev_src_t **)lhs_;
	const swift_ev_src_t *rhs = *(const swift_ev_src_t **)rhs_;

	return (lhs->poll_ev == rhs->poll_ev);
}

swift_ev_src_t *swift_new_ev_src(swift_poll_ev_fn_t poll_ev, void *user_data,
	unsigned int flags)
{
	swift_ev_src_t *src;

	if (!poll_ev)
		return NULL;

	if (!(src = calloc(1, sizeof(swift_ev_src_t))))
		return NULL;

	swift_init_ref(&src->ref);
	src->poll_ev = poll_ev;
	src->user_data = user_data;
	src->flags = flags;

	return src;
}

void swift_del_ev_src(swift_ev_src_t *src)
{
	if (!src)
		return;

	swift_release_ref(&src->ref, swift_do_del_ev_src);
}

swift_ev_src_t *swift_ref_ev_src(swift_ev_src_t *src)
{
	if (!src)
		return NULL;

	swift_acquire_ref(&src->ref);

	return src;
}

bool swift_add_ev_src(swift_poll_ev_fn_t poll_ev, void *user_data,
	unsigned int flags)
{
	swift_htable_iter_t iter;
	swift_ev_src_t *src;
	swift_ev_loop_t *loop;
	bool add_src;

	if (!poll_ev || swift_get_ev_src(poll_ev))
		return false;

	/* Create a new event source object. */
	if (!(src = swift_new_ev_src(poll_ev, user_data, flags)))
		return false;

	swift_mutex_lock(ev_ctx->lock);

	/* Add the event source to the existing event loops. */
	for (iter = swift_htable_first(ev_ctx->loops); iter.data;
		swift_htable_next(&iter)) {
		loop = *(swift_ev_loop_t **)iter.data;

		/* Check if the event source should be added to the current
		 * event loop.
		 */
		add_src = (loop->thread_id == swift_get_main_thread_id()) ?
			(src->flags & SWIFT_MAIN_THREAD) :
			(src->flags & SWIFT_WORKER_THREAD);

		if (!add_src)
			continue;

		src = swift_ref_ev_src(src);

		if (swift_htable_insert(loop->srcs, &src) == SIZE_MAX)
			goto err_remove_src;
	}

	/* Add the event source to the list of available event sources. */
	if (swift_htable_insert(ev_ctx->srcs, &src) == SIZE_MAX)
		goto err_remove_src;

	swift_mutex_unlock(ev_ctx->lock);

	return true;

err_remove_src:
	for (iter = swift_htable_first(ev_ctx->loops); iter.data;
		swift_htable_next(&iter)) {
		loop = *(swift_ev_loop_t **)iter.data;

		swift_htable_remove(loop->srcs, &src);
		swift_del_ev_src(src);
	}

	swift_mutex_unlock(ev_ctx->lock);
	swift_del_ev_src(src);
	return false;
}

void swift_remove_ev_src(swift_poll_ev_fn_t poll_ev)
{
	swift_htable_iter_t iter;
	swift_ev_src_t *src;
	swift_ev_loop_t *loop;

	swift_mutex_lock(ev_ctx->lock);

	if (!(src = swift_get_ev_src(poll_ev)))
		goto err_unlock_ctx;

	/* Remove the event source from existing event loops. */
	for (iter = swift_htable_first(ev_ctx->loops); iter.data;
		swift_htable_next(&iter)) {
		loop = *(swift_ev_loop_t **)iter.data;
		swift_htable_remove(loop->srcs, &src);
	}

	/* Remove the event source from the list of available event sources. */
	swift_htable_remove(ev_ctx->srcs, &src);
	swift_del_ev_src(src);

	swift_mutex_unlock(ev_ctx->lock);

	return;

err_unlock_ctx:
	swift_mutex_unlock(ev_ctx->lock);
}
