#include <swift/data.h>
#include <swift/event.h>
#include <swift/time.h>

typedef struct swift_timer_t swift_timer_t;
typedef struct swift_timer_ctx_t swift_timer_ctx_t;

struct swift_timer_t {
	swift_rb_node_t node;
	swift_list_t list;
	swift_time_t time;
	swift_run_timer_fn_t run_timer;
	void *user_data;
};

struct swift_timer_ctx_t {
	swift_list_t *scheduled_timers;
	swift_rb_tree_t *running_timers;
	swift_mutex_t *lock;
};

static swift_timer_ctx_t *timer_ctx = NULL;

static int swift_cmp_timers(swift_rb_tree_t *tree, swift_rb_node_t *lhs_,
	swift_rb_node_t *rhs_)
{
	swift_timer_t *lhs = SWIFT_CONTAINER_OF(lhs_, swift_timer_t, node);
	swift_timer_t *rhs = SWIFT_CONTAINER_OF(rhs_, swift_timer_t, node);

	return swift_cmp_time(lhs->time, rhs->time);
}

static bool swift_init_timer(swift_timer_t *timer)
{
	if (!timer)
		return false;

	if (!swift_init_list(&timer->list))
		return false;

	return true;
}

static void swift_term_timer(swift_timer_t *timer)
{
	if (!timer)
		return;

	swift_term_list(&timer->list);
}

static swift_timer_t *swift_new_timer(void)
{
	swift_timer_t *timer;

	timer = calloc(1, sizeof(swift_timer_t));

	if (!timer)
		return NULL;

	if (!swift_init_timer(timer))
		goto err_free_timer;

	return timer;

err_free_timer:
	SWIFT_FREE(timer);
	return NULL;
}

static void swift_del_timer(swift_timer_t *timer)
{
	if (!timer)
		return;

	swift_term_timer(timer);
	SWIFT_FREE(timer);
}

static bool swift_queue_timer(void)
{
	swift_timer_t *timer;
	swift_list_t *entry;

	swift_mutex_lock(timer_ctx->lock);
	entry = swift_list_shift(timer_ctx->scheduled_timers);
	swift_mutex_unlock(timer_ctx->lock);

	if (!entry)
		return false;

	timer = SWIFT_CONTAINER_OF(entry, swift_timer_t, list);

	swift_mutex_lock(timer_ctx->lock);
	swift_rb_insert(timer_ctx->running_timers, &timer->node);
	swift_mutex_unlock(timer_ctx->lock);

	return true;
}

static bool swift_dequeue_timer()
{
	swift_time_t time;
	swift_timer_t *timer;
	swift_rb_node_t *node;

	swift_mutex_lock(timer_ctx->lock);
	node = swift_rb_shift(timer_ctx->running_timers);
	swift_mutex_unlock(timer_ctx->lock);

	if (!node)
		return false;

	timer = SWIFT_CONTAINER_OF(node, swift_timer_t, node);
	time = swift_get_monotonic_time();

	if (swift_cmp_time(time, timer->time) < 0)
		goto err_queue_timer;

	timer->run_timer(timer->user_data);
	swift_del_timer(timer);

	return true;

err_queue_timer:
	swift_mutex_lock(timer_ctx->lock);
	swift_rb_insert(timer_ctx->running_timers, &timer->node);
	swift_mutex_unlock(timer_ctx->lock);
	return false;
}

static void swift_poll_timers(void *user_data)
{
	if (!timer_ctx)
		return;

	while (swift_queue_timer());
	while (swift_dequeue_timer());
}

bool swift_init_timers(void)
{
	if (timer_ctx)
		return true;

	if (!(timer_ctx = malloc(sizeof(*timer_ctx))))
		return false;

	if (!(timer_ctx->lock = swift_new_mutex(SWIFT_MUTEX_RECURSIVE)))
		goto err_free_timer_ctx;

	if (!(timer_ctx->scheduled_timers = swift_new_list()))
		goto err_del_lock;

	if (!(timer_ctx->running_timers = swift_rb_new_tree(swift_cmp_timers)))
		goto err_del_scheduled_timers;

	if (!swift_add_ev_src(swift_poll_timers, NULL, SWIFT_MAIN_THREAD |
		SWIFT_WORKER_THREAD))
		goto err_del_running_timers;

	return true;

err_del_running_timers:
	swift_rb_del_tree(timer_ctx->running_timers);
err_del_scheduled_timers:
	swift_del_list(timer_ctx->scheduled_timers);
err_del_lock:
	swift_del_mutex(timer_ctx->lock);
err_free_timer_ctx:
	SWIFT_FREE(timer_ctx);
	return false;
}

void swift_term_timers(void)
{
	if (!timer_ctx)
		return;

	swift_remove_ev_src(swift_poll_timers);
	swift_rb_del_tree(timer_ctx->running_timers);
	swift_del_list(timer_ctx->scheduled_timers);
	swift_del_mutex(timer_ctx->lock);
	SWIFT_FREE(timer_ctx);
}

bool swift_timer_schedule(swift_time_t delay, swift_run_timer_fn_t run_timer,
	void *user_data)
{
	swift_timer_t *timer;
	bool ret;

	if (!timer_ctx || !run_timer)
		return false;

	timer = swift_new_timer();

	if (!timer)
		return false;

	timer->time = swift_add_time(swift_get_monotonic_time(), delay);
	timer->run_timer = run_timer;
	timer->user_data = user_data;

	swift_mutex_lock(timer_ctx->lock);
	ret = swift_list_push(timer_ctx->scheduled_timers, &timer->list);
	swift_mutex_unlock(timer_ctx->lock);

	if (!ret)
		goto err_del_timer;

	return true;

err_del_timer:
	swift_del_timer(timer);
	return false;
}
