#include <swift/fs.h>
#include <swift/string.h>
#include <swift/unicode.h>

#define WIN32_MEAN_AND_LEAN
#include <windows.h>

struct swift_dir_t {
	WIN32_FIND_DATA find_data;
	swift_str_t *name;
	HANDLE handle;
	bool has_next;
};

static bool swift_filetime_to_time(swift_time_t *time,
	const FILETIME *file_time)
{
	SYSTEMTIME utc_time, local_time;
	TIME_ZONE_INFORMATION time_zone_info;
	DWORD time_zone_id;
	struct tm time_info;

	if (!time)
		return false;

	if (!FileTimeToSystemTime(file_time, &utc_time))
		return false;

	time_zone_id = GetTimeZoneInformation(&time_zone_info);

	if (time_zone_id == TIME_ZONE_ID_INVALID)
		return false;

	if (!SystemTimeToTzSpecificLocalTime(&time_zone_info, &utc_time,
		&local_time))
		return false;

	time_info.tm_sec = local_time.wSecond;
	time_info.tm_min = local_time.wMinute;
	time_info.tm_hour = local_time.wHour;

	time_info.tm_mday = local_time.wDay;
	time_info.tm_mon = local_time.wMonth - 1;
	time_info.tm_year = local_time.wYear - 1900;

	time_info.tm_wday = -1;
	time_info.tm_yday = -1;

	time_info.tm_isdst = (time_zone_id == TIME_ZONE_ID_DAYLIGHT);

	*time = SWIFT_TIME(mktime(&time_info));

	return true;
}

bool swift_check_path(const char *path_)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	uint16_t *path;
	bool ret;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	return ret;
}

bool swift_is_file(const char *path_)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	uint16_t *path;
	bool ret;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	if (!ret)
		return false;

	return !(file_attr.dwFileAttributes & (FILE_ATTRIBUTE_DEVICE |
		FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_OFFLINE |
		FILE_ATTRIBUTE_REPARSE_POINT));
}

bool swift_is_dir(const char *path_)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	uint16_t *path;
	bool ret;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	if (!ret)
		return false;

	return !!(file_attr.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
}

bool swift_is_symlink(const char *path_)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	uint16_t *path;
	bool ret;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	if (!ret)
		return false;

	return !!(file_attr.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT);
}

bool swift_get_atime(const char *path_, swift_time_t *time)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	uint16_t *path;
	bool ret;

	if (!path_ || !time)
		return false;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	if (!ret)
		return false;

	return swift_filetime_to_time(time, &file_attr.ftLastAccessTime);
}

bool swift_get_ctime(const char *path_, swift_time_t *time)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	uint16_t *path;
	bool ret;

	if (!path_ || !time)
		return false;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	if (!ret)
		return false;

	return swift_filetime_to_time(time, &file_attr.ftCreationTime);
}

bool swift_get_mtime(const char *path_, swift_time_t *time)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	uint16_t *path;
	bool ret;

	if (!path_ || !time)
		return false;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	if (!ret)
		return false;

	return swift_filetime_to_time(time, &file_attr.ftLastWriteTime);
}

uint64_t swift_get_file_size(const char *path_)
{
	WIN32_FILE_ATTRIBUTE_DATA file_attr;
	ULARGE_INTEGER file_size;
	uint16_t *path;
	bool ret;

	if (!path_)
		return false;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	ret = GetFileAttributesEx(path, GetFileExInfoStandard, &file_attr);
	SWIFT_FREE(path);

	if (!ret)
		return false;

	file_size.HighPart = file_attr.nFileSizeHigh;
	file_size.LowPart = file_attr.nFileSizeLow;

	return file_size.QuadPart;
}

bool swift_new_dir(const char *path_)
{
	uint16_t *path;
	bool status;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	status = CreateDirectory(path, NULL);
	SWIFT_FREE(path);

	return status;
}

bool swift_del_dir(const char *path_)
{
	uint16_t *path;
	bool status;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	status = RemoveDirectory(path);
	SWIFT_FREE(path);

	return status;
}

swift_dir_t *swift_open_dir(const char *path_)
{
	swift_dir_t *dir;
	uint16_t *path;
	size_t count;

	path_ = path_ ? path_ : NULL;

	dir = calloc(1, sizeof(swift_dir_t));

	if (!dir)
		return NULL;

	dir->handle = INVALID_HANDLE_VALUE;
	dir->name = swift_new_cstr(path_);

	count = dir->name->count;

	if (count && dir->name->data[count - 1] != '/' &&
		dir->name->data[count - 1] != '\\')
		swift_append_uchar(dir->name, '/');

	swift_append_uchar(dir->name, '*');

	if (!dir->name)
		goto err_free_dir;

	path = swift_new_conv_str("utf-16", "utf-8", dir->name->data, SIZE_MAX);

	if (!path)
		goto err_del_name;

	dir->handle = FindFirstFile(path, &dir->find_data);
	SWIFT_FREE(path);

	if (dir->handle == INVALID_HANDLE_VALUE)
		goto err_del_name;

	dir->has_next = true;

	return dir;

err_del_name:
	swift_del_str(dir->name);
err_free_dir:
	SWIFT_FREE(dir);
	return NULL;
}

const char *swift_read_dir(swift_dir_t *dir)
{
	if (!dir || !dir->has_next)
		return NULL;

	if (!swift_set_from(dir->name, "utf-16", dir->find_data.cFileName,
		SIZE_MAX))
		return NULL;

	dir->has_next = FindNextFile(dir->handle, &dir->find_data);

	return dir->name->data;
}

void swift_close_dir(swift_dir_t *dir)
{
	if (!dir)
		return;

	FindClose(dir->handle);
	swift_del_str(dir->name);
	SWIFT_FREE(dir);
}

bool swift_move_file(const char *dst_path_, const char *src_path_)
{
	uint16_t *dst_path, *src_path;
	bool status;

	dst_path = swift_new_conv_str("utf-16", "utf-8", dst_path_, SIZE_MAX);

	if (!dst_path)
		return false;

	src_path = swift_new_conv_str("utf-16", "utf-8", src_path_, SIZE_MAX);

	if (!src_path)
		goto err_free_dst_path;

	status = MoveFileEx(src_path, dst_path, MOVEFILE_COPY_ALLOWED |
		MOVEFILE_REPLACE_EXISTING);

	SWIFT_FREE(src_path);
	SWIFT_FREE(dst_path);

	return status;

err_free_dst_path:
	SWIFT_FREE(dst_path);
	return false;
}

bool swift_del_file(const char *path_)
{
	uint16_t *path;
	bool status;

	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		return false;

	status = DeleteFile(path);
	SWIFT_FREE(path);

	return status;
}
