#include <swift/image.h>
#include <swift/io.h>

#include <png.h>

#include "api.h"
#include "main.h"

void codec_read(png_structp ctx, png_bytep data, png_size_t size)
{
	swift_stream_t *stream = png_get_io_ptr(ctx);

	swift_stream_read(stream, data, size);
}

void codec_write(png_structp ctx, png_bytep data, png_size_t size)
{
	swift_stream_t *stream = png_get_io_ptr(ctx);

	swift_stream_write(stream, data, size);
}

void codec_flush(png_structp ctx)
{
	swift_stream_t *stream = png_get_io_ptr(ctx);

	swift_stream_flush(stream);
}

static swift_image_t *codec_read_image(swift_stream_t *stream)
{
	swift_image_t *image;
	swift_pixfmt_t *fmt;
	png_bytep *rows;
	png_structp ctx;
	png_infop info;
	size_t row;
	int channels, depth, colour_type, interlace_method;
	png_uint_32 width, height;

	ctx = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL,
		NULL);

	if (!ctx)
		return false;

	info = png_create_info_struct(ctx);

	if (!info)
		goto err_free_ctx;

	if (setjmp(png_jmpbuf(ctx)))
		goto err_free_ctx;

	png_set_read_fn(ctx, stream, codec_read);

	png_read_info(ctx, info);
	png_get_IHDR(ctx, info, &width, &height, &depth, &colour_type,
		&interlace_method, NULL, NULL);
	png_set_strip_16(ctx);
	png_set_packing(ctx);

	if (colour_type == PNG_COLOR_TYPE_GRAY)
		png_set_expand(ctx);

	if (colour_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		png_set_gray_to_rgb(ctx);

	png_read_update_info(ctx, info);
	png_get_IHDR(ctx, info, &width, &height, &depth, &colour_type,
		&interlace_method, NULL, NULL);

	channels = png_get_channels(ctx, info);

	if (channels == 4) {
		fmt = swift_get_pixfmt("RGBA32");
	} else if (channels == 3) {
		fmt = swift_get_pixfmt("RGB24");
	} else {
		/* TODO: support these cases. */
		fprintf(stderr, "Png formate not supported");
		goto err_free_ctx;
	}

	if (!fmt)
		goto err_free_ctx;

	image = swift_new_image(fmt, width, height);

	if (!image)
		goto err_free_ctx;

	rows = (png_bytep *)calloc(image->height, sizeof(png_bytep));

	if (!rows)
		goto err_del_image;

	for (row = 0; row < image->height; ++row)
		rows[row] = image->data + row * image->pitch;

	png_read_image(ctx, rows);
	png_read_end(ctx, info);

	SWIFT_FREE(rows);
	png_destroy_read_struct(&ctx, &info, NULL);

	return image;

err_del_image:
	swift_del_image(image);
err_free_ctx:
	png_destroy_read_struct(&ctx, info ? &info : NULL, NULL);
	return NULL;
}

static bool codec_write_image(swift_stream_t *stream, swift_image_t *image)
{
	png_bytep *rows;
	png_structp ctx;
	png_infop info;
	size_t row;

	ctx = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL,
		NULL);

	if (!ctx)
		return false;

	info = png_create_info_struct(ctx);

	if (!info)
		goto err_free_ctx;

	if (setjmp(png_jmpbuf(ctx)))
		goto err_free_ctx;

	png_set_write_fn(ctx, stream, codec_write, codec_flush);

	png_set_IHDR(ctx, info, image->width, image->height, 8,
		PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	png_write_info(ctx, info);

	rows = (png_bytep *)calloc(image->height, sizeof(png_bytep));

	if (!rows)
		goto err_free_ctx;

	for (row = 0; row < image->height; ++row)
		rows[row] = image->data + row * image->pitch;

	png_write_image(ctx, rows);
	png_write_end(ctx, info);

	png_destroy_write_struct(&ctx, NULL);

	return true;

err_free_ctx:
	png_destroy_write_struct(&ctx, NULL);
	return false;
}

void swift_init_image_codec(void)
{
	swift_add_image_codec("png", "Portable Network Graphics",
		codec_read_image, codec_write_image);
}

void swift_term_image_codec(void)
{
	swift_remove_image_codec("png");
}
