#include <swift/data.h>
#include <swift/dylib.h>
#include <swift/fs.h>
#include <swift/image.h>
#include <swift/string.h>

typedef bool (* swift_init_image_codec_fn_t)(void);
typedef void (* swift_term_image_codec_fn_t)(void);

static swift_array_t *codecs_by_index;
static swift_htable_t *libs, *codecs_by_name;

static size_t swift_hash_ptr(const void *data)
{
	return swift_fnv1a_hash64(data, sizeof(void *));
}

static bool swift_eq_ptrs(const void *lhs_, const void *rhs_)
{
	const void *lhs = *(const void **)lhs_;
	const void *rhs = *(const void **)rhs_;

	return (lhs == rhs);
}

static size_t swift_hash_codec(const void *data)
{
	const swift_image_codec_t *codec = *(const swift_image_codec_t **)data;

	return swift_fnv1a_hash64(codec->name, strlen(codec->name));
}

static bool swift_eq_codecs(const void *lhs_, const void *rhs_)
{
	const swift_image_codec_t *lhs = *(const swift_image_codec_t **)lhs_;
	const swift_image_codec_t *rhs = *(const swift_image_codec_t **)rhs_;

	return (strcmp(lhs->name, rhs->name) == 0);
}

static bool swift_open_image_codec(const char *path)
{
	swift_init_image_codec_fn_t init_image_codec;
	swift_lib_t *lib;

	lib = swift_open_lib(path);

	if (!lib)
		return false;

	init_image_codec = (swift_init_image_codec_fn_t)swift_get_code_sym(lib,
		"swift_init_image_codec");

	if (!init_image_codec)
		goto err_close_lib;

	if (swift_htable_insert(libs, &lib) == SIZE_MAX)
		goto err_close_lib;

	init_image_codec();

	return true;

err_close_lib:
	swift_close_lib(lib);
	return false;
}

static void swift_close_image_codec(swift_lib_t *lib)
{
	swift_term_image_codec_fn_t term_image_codec;

	if (!lib)
		return;

	term_image_codec = (swift_term_image_codec_fn_t)swift_get_code_sym(lib,
		"swift_term_image_codec");

	if (term_image_codec)
		term_image_codec();

	swift_close_lib(lib);
}

static void swift_open_image_codecs(const char *path)
{
	swift_str_t *codec_path = swift_new_str();
	swift_dir_t *dir;
	const char *entry;

	dir = swift_open_dir(path);

	if (!dir)
		return;

	while ((entry = swift_read_dir(dir))) {
		swift_set_strf(codec_path, "%s/%s", path, entry);

		swift_open_image_codec(codec_path->data);
	}

	swift_del_str(codec_path);
	swift_close_dir(dir);
}

bool swift_init_image_codecs(const char *path)
{
	libs = swift_new_htable(swift_hash_ptr, swift_eq_ptrs,
		sizeof(swift_lib_t *));

	if (!libs)
		return false;

	codecs_by_name = swift_new_htable(swift_hash_codec, swift_eq_codecs,
		sizeof(swift_image_codec_t *));

	if (!codecs_by_name)
		goto err_del_libs;

	codecs_by_index = swift_new_array(NULL, sizeof(swift_image_codec_t *));

	if (!codecs_by_index)
		goto err_del_codecs_by_name;

	swift_open_image_codecs(path);

	return true;

err_del_codecs_by_name:
	swift_del_htable(codecs_by_name);
err_del_libs:
	swift_del_htable(libs);
	return false;
}

void swift_term_image_codecs(void)
{
	swift_htable_iter_t iter;
	swift_image_codec_t *codec;
	swift_lib_t *lib;

	for (iter = swift_htable_first(libs); iter.data;
		swift_htable_next(&iter)) {
		lib = *(swift_lib_t **)iter.data;
		swift_htable_remove_index(libs, iter.index);
		swift_close_image_codec(lib);
	}

	for (iter = swift_htable_first(codecs_by_name); iter.data;
		swift_htable_next(&iter)) {
		codec = *(swift_image_codec_t **)iter.data;
		SWIFT_FREE(codec->full_name);
		SWIFT_FREE(codec->name);
		SWIFT_FREE(codec);
	}

	swift_del_array(codecs_by_index);
	swift_del_htable(codecs_by_name);
	swift_del_htable(libs);
}

size_t swift_get_image_codec_count(void)
{
	return codecs_by_index->count;
}

swift_image_codec_t *swift_get_image_codec(size_t index)
{
	void *entry;

	entry = swift_array_get(codecs_by_index, index);

	if (!entry)
		return NULL;

	return *(swift_image_codec_t **)entry;
}

swift_image_codec_t *swift_search_image_codec(const char *name)
{
	swift_image_codec_t key_ = {
		.name = (char *)name,
	};
	swift_image_codec_t *key = &key_;
	void *entry;

	entry = swift_htable_search(codecs_by_name, &key);

	if (!entry)
		return NULL;

	return *(swift_image_codec_t **)entry;
}

bool swift_add_image_codec(const char *name, const char *full_name,
	swift_read_image_fn_t read_image, swift_write_image_fn_t write_image)
{
	swift_image_codec_t *codec;

	if (!name || !full_name)
		return false;

	codec = calloc(1, sizeof(swift_image_codec_t));

	if (!codec)
		return false;

	codec->name = strdup(name);
	codec->full_name = strdup(full_name);
	codec->read_image = read_image;
	codec->write_image = write_image;

	if (swift_htable_insert(codecs_by_name, &codec) == SIZE_MAX)
		goto err_free_codec;

	if (!swift_array_push(codecs_by_index, &codec))
		goto err_remove_codec;

	codec->index = codecs_by_index->count - 1;

	return true;

err_remove_codec:
	swift_htable_remove(codecs_by_name, &codec);
err_free_codec:
	SWIFT_FREE(codec->full_name);
	SWIFT_FREE(codec->name);
	SWIFT_FREE(codec);
	return false;
}

void swift_remove_image_codec(const char *name)
{
	swift_image_codec_t key_ = {
		.name = (char *)name,
	};
	swift_image_codec_t *codec, *key = &key_;
	size_t index;

	index = swift_htable_search_index(codecs_by_name, &key);

	if (index == SIZE_MAX)
		return;

	codec = *(swift_image_codec_t **)swift_htable_get(codecs_by_name,
		index);

	swift_array_remove_index(codecs_by_index, codec->index);
	swift_htable_remove_index(codecs_by_name, index);

	SWIFT_FREE(codec->full_name);
	SWIFT_FREE(codec->name);
	SWIFT_FREE(codec);
}

swift_image_t *swift_try_read_image(swift_stream_t *stream)
{
	swift_htable_iter_t iter;
	swift_image_codec_t *codec;
	swift_image_t *image;

	if (!stream)
		return NULL;

	for (iter = swift_htable_first(codecs_by_name); iter.data;
		swift_htable_next(&iter)) {
		codec = *(swift_image_codec_t **)iter.data;

		image = swift_read_image(codec, stream);

		if (image)
			return image;
	}

	return NULL;
}

swift_image_t *swift_try_read_image_from_file(const char *path)
{
	swift_htable_iter_t iter;
	swift_image_codec_t *codec;
	swift_image_t *image;

	if (!path)
		return NULL;

	for (iter = swift_htable_first(codecs_by_name); iter.data;
		swift_htable_next(&iter)) {
		codec = *(swift_image_codec_t **)iter.data;

		image = swift_read_image_from_file(codec, path);

		if (image)
			return image;
	}

	return NULL;
}

swift_image_t *swift_read_image(swift_image_codec_t *codec,
	swift_stream_t *stream)
{
	if (!codec || !codec->read_image || !stream)
		return NULL;

	return codec->read_image(stream);
}

swift_image_t *swift_read_image_from_file(swift_image_codec_t *codec,
	const char *path)
{
	swift_stream_t *stream;
	swift_image_t *image;

	if (!codec || !path)
		return NULL;

	stream = swift_open_file_stream(path, SWIFT_FILE_FLAGS_READ);

	if (!stream)
		return NULL;

	image = swift_read_image(codec, stream);
	swift_close_stream(stream);

	return image;
}

bool swift_write_image(swift_image_codec_t *codec, swift_stream_t *stream,
	swift_image_t *image)
{
	if (!codec || !codec->write_image || !stream || !image)
		return false;

	return codec->write_image(stream, image);
}

bool swift_write_image_to_file(swift_image_codec_t *codec, const char *path,
	swift_image_t *image)
{
	swift_stream_t *stream;
	bool ret;

	if (!codec || !path || !image)
		return false;

	stream = swift_open_file_stream(path, SWIFT_FILE_FLAGS_TRUNCATE);

	if (!stream)
		return NULL;

	ret = swift_write_image(codec, stream, image);
	swift_close_stream(stream);

	return ret;
}
