#include <swift/io.h>

#define MASK(bits) ((UINT64_C(1) << (bits)) - 1)

uint64_t swift_enc754_normal(double val, unsigned int bits,
	unsigned int exp_bits)
{
	double frac_mask, shift;
	uint64_t sign, exp, significand = 0;
	unsigned int significand_bits = bits - exp_bits - 1;
	unsigned int bit;

	/* Extract the sign and normalise the value. */
	sign = (signbit(val) > 0) ? 1 : 0;
	val = (sign) ? -val : val;

	/* Extract the exponent and add the bias. */
	shift = floor(log2(val));
	val /= pow(2.0, shift);
	exp = (UINT64_C(1) << (exp_bits - 1)) - 1 + shift;

	/* Extract the fraction and set the appropriate bits. */
	val = fmod(val, 1.0);
	frac_mask = 1.0;
	bit = significand_bits;

	do {
		--bit;
		frac_mask /= 2.0;

		if (val >= frac_mask) {
			val -= frac_mask;
			significand |= UINT64_C(1) << bit;
		}
	} while (bit && val);

	return (sign << (bits - 1)) | (exp << (bits - exp_bits - 1)) |
		significand;
}

uint64_t swift_enc754(double val, int val_class, unsigned int bits,
	unsigned int exp_bits)
{
	uint64_t sign, exp, significand = 0;
	unsigned int significand_bits = bits - exp_bits - 1;

	switch (val_class) {
		case FP_INFINITE: {
			exp = MASK(exp_bits);
			significand = 0;
		} break;

		case FP_NAN: {
			exp = MASK(exp_bits);
			significand = MASK(significand_bits);
		} break;

		case FP_NORMAL: {
			return swift_enc754_normal(val, bits, exp_bits);
		} break;

		case FP_SUBNORMAL:
		case FP_ZERO: {
			exp = 0;
			significand = 0;
		} break;
	}

	sign = (signbit(val) > 0) ? 1 : 0;

	return (sign << (bits - 1)) | (exp << (bits - exp_bits - 1)) |
		significand;
}

double swift_dec_normal754(uint64_t val, unsigned int bits,
	unsigned int exp_bits)
{
	double ret = 1.0, frac_mask;
	uint64_t sign, significand, exp;
	uint64_t bias;
	int64_t shift;
	unsigned significand_bits = bits - exp_bits - 1;
	unsigned bit;

	significand = val & MASK(significand_bits);
	val >>= significand_bits;

	exp = val & MASK(exp_bits);
	val >>= exp_bits;

	sign = val & MASK(1);

	/* Compute the fraction based on the field. */
	frac_mask = 1.0;
	bit = significand_bits;

	do {
		--bit;
		frac_mask /= 2.0;

		if (significand & (UINT64_C(1) << bit))
			ret += frac_mask;
	} while (bit);

	/* Shift the val appropriately based on the exponent field. */
	bias = (UINT64_C(1) << (exp_bits - 1)) - 1;
	shift = (int64_t)(exp - bias);
	ret *= pow(2.0, shift);

	/* Sign the val. */
	return (sign) ? -ret : ret;
}

double swift_dec754(uint64_t val, unsigned int bits,
	unsigned int exp_bits)
{
	uint64_t val_ = val;
	uint64_t sign, significand, exp;
	unsigned significand_bits = bits - exp_bits - 1;

	significand = val & MASK(significand_bits);
	val >>= significand_bits;

	exp = val & MASK(exp_bits);
	val >>= exp_bits;

	sign = val & MASK(1);

	if (exp == MASK(exp_bits)) {
		if (significand) {
			return (sign) ? -NAN : +NAN;
		} else {
			return (sign) ? -INFINITY : +INFINITY;
		}
	} else if (exp) {
		return swift_dec_normal754(val_, bits, exp_bits);
	}

	return 0.0;
}

uint32_t swift_enc754_32(float val)
{
	return swift_enc754(val, fpclassify(val), 32, 8);
}

uint64_t swift_enc754_64(double val)
{
	return swift_enc754(val, fpclassify(val), 64, 11);
}

float swift_dec754_32(uint32_t val)
{
	return swift_dec754(val, 32, 8);
}

double swift_dec754_64(uint64_t val)
{
	return swift_dec754(val, 64, 11);
}
