#include <swift/io.h>
#include <swift/unicode.h>

#define WIN32_MEAN_AND_LEAN
#include <windows.h>

struct swift_file_t {
	HANDLE handle;
};

swift_file_t *swift_open_file(const char *path_, unsigned int flags)
{
	swift_file_t *file;
	uint16_t *path;
	DWORD req_access, share_mode, open_mode, error_mode;

	file = malloc(sizeof(swift_file_t));

	if (!file)
		return NULL;

	/* If the user requests for the file to be the truncated and/or
	 * appended, then make sure the write flag is set as well.
	 */
	flags |= (flags & (SWIFT_FILE_FLAGS_TRUNCATE |
		SWIFT_FILE_FLAGS_APPEND)) ? SWIFT_FILE_FLAGS_WRITE : 0;

	/* Set the requested access accordingly. These flags tell whether the
	 * file is going to be read from and/or written to.
	 */
	req_access = (flags & SWIFT_FILE_FLAGS_READ) ? GENERIC_READ : 0 |
		(flags & SWIFT_FILE_FLAGS_WRITE) ? GENERIC_WRITE : 0;

	/* If the file isn't going to be written to, it is safe to share it for
	 * reading purposes.
	 */
	share_mode = (flags & SWIFT_FILE_FLAGS_WRITE) ? 0 : FILE_SHARE_READ;

	/* Set the open_mode accordingly. The open mode tells whether the file
	 * should always be created (truncate), should be created if it does not
	 * exist (write), should be opened only if it exists (read). */
	if (flags & SWIFT_FILE_FLAGS_TRUNCATE) {
		open_mode = CREATE_ALWAYS;
	} else if (flags & SWIFT_FILE_FLAGS_WRITE) {
		open_mode = OPEN_ALWAYS;
	} else if (flags & SWIFT_FILE_FLAGS_READ) {
		open_mode = OPEN_EXISTING;
	} else {
		goto err_free_file;
	}

	/* Convert the string to native format. */
	path = swift_new_conv_str("utf-16", "utf-8", path_, SIZE_MAX);

	if (!path)
		goto err_free_file;

	/* Disable the error dialogue when creating the file. */
	error_mode = SetErrorMode(SEM_NOOPENFILEERRORBOX |
		SEM_FAILCRITICALERRORS);

	/* Attempt to create and/or open the file. */
	file->handle = CreateFile(path, req_access, share_mode, NULL, open_mode,
		FILE_ATTRIBUTE_NORMAL, NULL);

	/* Restore the error mode. */
	SetErrorMode(error_mode);
	SWIFT_FREE(path);

	/* Check if the file was opened successfully. */
	if (file->handle == INVALID_HANDLE_VALUE)
		goto err_free_file;

	return file;

err_free_file:
	SWIFT_FREE(file);
	return NULL;
}

void swift_close_file(swift_file_t *file)
{
	if (!file)
		return;

	CloseHandle(file->handle);
	SWIFT_FREE(file);
}

int64_t swift_seek_file(swift_file_t *file, unsigned int whence_,
	int64_t pos_)
{
	LARGE_INTEGER ret_pos, pos = {
		.QuadPart = pos_,
	};
	DWORD whence;

	switch (whence_) {
		case SWIFT_FILE_SEEK_SET: {
			whence = FILE_BEGIN;
		} break;

		case SWIFT_FILE_SEEK_CURRENT: {
			whence = FILE_CURRENT;
		} break;

		case SWIFT_FILE_SEEK_END: {
			whence = FILE_END;
		} break;
	}

	if (!SetFilePointerEx(file->handle, pos, &ret_pos, whence))
		return -1;

	return ret_pos.QuadPart;
}

size_t swift_read_file(swift_file_t *file, void *buffer_, size_t size)
{
	unsigned char *buffer = (unsigned char *)buffer_;
	size_t ret = 0;
	DWORD req_size, chunk_size;

	if (!file || !buffer || !size)
		return 0;

	/* Attempt to read the buffer from the file. If the buffer is too large,
	 * consecutive chunks are being read in order to support large files.
	 */
	do {
		req_size = (DWORD)SWIFT_MIN(size, 0xFFFFFFFF);

		if (!ReadFile(file->handle, buffer, req_size, &chunk_size,
			NULL))
			return 0;

		buffer += chunk_size;
		size -= chunk_size;
		ret += chunk_size;
	} while (chunk_size == req_size && size);

	return ret;
}

size_t swift_write_file(swift_file_t *file, const void *buffer_,
	size_t size)
{
	unsigned char *buffer = (unsigned char *)buffer_;
	size_t ret = 0;
	DWORD req_size, chunk_size;

	if (!file || !buffer || !size)
		return 0;

	/* Attempt to write the buffer to the file. If the buffer is too large,
	 * consecutive chunks are being written in order to support large files.
	 */
	do {
		req_size = (DWORD)SWIFT_MIN(size, 0xFFFFFFFF);

		if (!WriteFile(file->handle, buffer, req_size, &chunk_size,
			NULL))
			return 0;

		buffer += chunk_size;
		size -= (size_t)chunk_size;
		ret += (size_t)chunk_size;
	} while (chunk_size == req_size && size);

	return ret;
}

bool swift_flush_file(swift_file_t *file)
{
	if (!file)
		return false;

	return FlushFileBuffers(file->handle);
}
