#define _LARGEFILE64_SOURCE

#include <swift/io.h>

#include <unistd.h>

#include <fcntl.h>

#include <sys/stat.h>
#include <sys/types.h>

#include "config.h"

struct swift_file_t {
	int handle;
};

swift_file_t *swift_open_file(const char *path, unsigned int flags_)
{
	swift_file_t *file;
	int flags;

	file = malloc(sizeof(swift_file_t));

	if (!file)
		return NULL;

	/* If the user requests for the file to be the truncated and/or
	 * appended, then make sure the write flag is set as well.
	 */
	flags = (flags_ & (SWIFT_FILE_FLAGS_TRUNCATE |
		SWIFT_FILE_FLAGS_APPEND)) ? SWIFT_FILE_FLAGS_WRITE : 0;

	/* Set the requested access accordingly. These flags tell whether the
	 * file is going to be read from and/or written to.
	 */
	if (flags_ & (SWIFT_FILE_FLAGS_READ | SWIFT_FILE_FLAGS_WRITE)) {
		flags_ |= O_RDWR;
	} else if (flags & SWIFT_FILE_FLAGS_WRITE) {
		flags_ |= O_WRONLY;
	} else if (flags & SWIFT_FILE_FLAGS_READ) {
		flags_ |= O_RDONLY;
	} else {
		goto err_free_file;
	}

	/* Set the truncate flag accordingly. */
	flags |= (flags_ & SWIFT_FILE_FLAGS_TRUNCATE) ? O_TRUNC : 0;

	/* Attempt to create and/or open the file. */
	file->handle = open(path, flags, S_IRWXU | S_IRWXG | S_IRWXO);

	/* Check if the file was opened successfully. */
	if (file->handle < 0)
		goto err_free_file;

	return file;

err_free_file:
	SWIFT_FREE(file);
	return NULL;
}

void swift_close_file(swift_file_t *file)
{
	if (!file)
		return;

	close(file->handle);
	SWIFT_FREE(file);
}

int64_t swift_seek_file(swift_file_t *file, unsigned int whence_, int64_t pos)
{
	int whence;

	if (!file)
		return -1;

	switch (whence_) {
		case SWIFT_FILE_SEEK_SET: {
			whence = SEEK_SET;
		} break;

		case SWIFT_FILE_SEEK_CURRENT: {
			whence = SEEK_CUR;
		} break;

		case SWIFT_FILE_SEEK_END: {
			whence = SEEK_END;
		} break;
	}

#if HAVE_LSEEK64
	return lseek64(file->handle, pos, whence);
#elif HAVE_LLSEEK
	return llseek(file->handle, pos, whence);
#elif HAVE_LSEEK
	return lseek(file->handle, pos, whence);
#else
#error your platform is not being supported.
#endif
}

size_t swift_read_file(swift_file_t *file, void *buffer, size_t size)
{
	if (!file || !buffer || !size)
		return 0;

	return read(file->handle, buffer, size);
}

size_t swift_write_file(swift_file_t *file, const void *buffer,
	size_t size)
{
	if (!file || !buffer || !size)
		return 0;

	return write(file->handle, buffer, size);
}

bool swift_flush_file(swift_file_t *file)
{
	if (!file)
		return false;

	return fsync(file->handle);
}
