#include <swift/io.h>

static void swift_do_close_stream(swift_ref_t *ref)
{
	swift_stream_t *stream = SWIFT_CONTAINER_OF(ref, swift_stream_t, ref);

	stream->stream_ops->close(stream);
}

void swift_close_stream(swift_stream_t *stream)
{
	if (!stream)
		return;

	swift_release_ref(&stream->ref, swift_do_close_stream);
}

swift_stream_t *swift_ref_stream(swift_stream_t *stream)
{
	if (!stream)
		return NULL;

	swift_acquire_ref(&stream->ref);

	return stream;
}

int64_t swift_stream_seek(swift_stream_t *stream, unsigned int whence,
	int64_t pos)
{
	if (!stream || !stream->stream_ops->seek)
		return -1;

	return stream->stream_ops->seek(stream, whence, pos);
}

size_t swift_stream_read(swift_stream_t *stream, void *buffer,
	size_t size)
{
	if (!stream || !stream->stream_ops->read || !buffer || !size)
		return 0;

	return stream->stream_ops->read(stream, buffer, size);
}

size_t swift_stream_write(swift_stream_t *stream, const void *buffer,
	size_t size)
{
	if (!stream || !stream->stream_ops->write || !buffer || !size)
		return 0;

	return stream->stream_ops->write(stream, buffer, size);
}

bool swift_stream_flush(swift_stream_t *stream)
{
	if (!stream || !stream->stream_ops->flush)
		return false;

	return stream->stream_ops->flush(stream);
}

bool swift_read_u8(swift_stream_t *stream, uint8_t *value)
{
	if (!stream || !value)
		return false;

	if (swift_stream_read(stream, value, sizeof(uint8_t)) < sizeof(uint8_t))
		return false;

	return true;
}

bool swift_read_u16_be(swift_stream_t *stream, uint16_t *value)
{
	if (!stream || !value)
		return false;

	if (swift_stream_read(stream, value, sizeof(uint16_t)) <
		sizeof(uint16_t))
		return false;

	*value = swift_bswap16_be(*value);

	return true;
}

bool swift_read_u16_le(swift_stream_t *stream, uint16_t *value)
{
	if (!stream || !value)
		return false;

	if (swift_stream_read(stream, value, sizeof(uint16_t)) <
		sizeof(uint16_t))
		return false;

	*value = swift_bswap16_le(*value);

	return true;
}

bool swift_read_u32_be(swift_stream_t *stream, uint32_t *value)
{
	if (!stream || !value)
		return false;

	if (swift_stream_read(stream, value, sizeof(uint32_t)) <
		sizeof(uint32_t))
		return false;

	*value = swift_bswap32_be(*value);

	return true;
}

bool swift_read_u32_le(swift_stream_t *stream, uint32_t *value)
{
	if (!stream || !value)
		return false;

	if (swift_stream_read(stream, value, sizeof(uint32_t)) <
		sizeof(uint32_t))
		return false;

	*value = swift_bswap32_le(*value);

	return true;
}

bool swift_read_u64_be(swift_stream_t *stream, uint64_t *value)
{
	if (!stream || !value)
		return false;

	if (swift_stream_read(stream, value, sizeof(uint64_t)) <
		sizeof(uint64_t))
		return false;

	*value = swift_bswap64_be(*value);

	return true;
}

bool swift_read_u64_le(swift_stream_t *stream, uint64_t *value)
{
	if (!stream || !value)
		return false;

	if (swift_stream_read(stream, value, sizeof(uint64_t)) <
		sizeof(uint64_t))
		return false;

	*value = swift_bswap64_le(*value);

	return true;
}

bool swift_write_u8(swift_stream_t *stream, uint8_t value)
{
	return (swift_stream_write(stream, &value, sizeof(uint8_t)) ==
		sizeof(uint8_t));
}

bool swift_write_u16_be(swift_stream_t *stream, uint16_t value)
{
	value = swift_bswap16_be(value);

	return (swift_stream_write(stream, &value, sizeof(uint16_t)) ==
		sizeof(uint16_t));
}

bool swift_write_u16_le(swift_stream_t *stream, uint16_t value)
{
	value = swift_bswap16_le(value);

	return (swift_stream_write(stream, &value, sizeof(uint16_t)) ==
		sizeof(uint16_t));
}

bool swift_write_u32_be(swift_stream_t *stream, uint32_t value)
{
	value = swift_bswap32_be(value);

	return (swift_stream_write(stream, &value, sizeof(uint32_t)) ==
		sizeof(uint32_t));
}

bool swift_write_u32_le(swift_stream_t *stream, uint32_t value)
{
	value = swift_bswap32_le(value);

	return (swift_stream_write(stream, &value, sizeof(uint32_t)) ==
		sizeof(uint32_t));
}

bool swift_write_u64_be(swift_stream_t *stream, uint64_t value)
{
	value = swift_bswap64_be(value);

	return (swift_stream_write(stream, &value, sizeof(uint64_t)) ==
		sizeof(uint64_t));
}

bool swift_write_u64_le(swift_stream_t *stream, uint64_t value)
{
	value = swift_bswap64_le(value);

	return (swift_stream_write(stream, &value, sizeof(uint64_t)) ==
		sizeof(uint64_t));
}

bool swift_read_f32_be(swift_stream_t *stream, float *value_)
{
	uint32_t value;

	if (!stream || !value_)
		return false;

	if (!swift_read_u32_be(stream, &value))
		return false;

	*value_ = swift_dec754_32(value);

	return true;
}

bool swift_read_f32_le(swift_stream_t *stream, float *value_)
{
	uint32_t value;

	if (!stream || !value_)
		return false;

	if (!swift_read_u32_le(stream, &value))
		return false;

	*value_ = swift_dec754_32(value);

	return true;
}

bool swift_read_f64_be(swift_stream_t *stream, double *value_)
{
	uint64_t value;

	if (!stream || !value_)
		return false;

	if (!swift_read_u64_be(stream, &value))
		return false;

	*value_ = swift_dec754_64(value);

	return true;
}

bool swift_read_f64_le(swift_stream_t *stream, double *value_)
{
	uint64_t value;

	if (!stream || !value_)
		return false;

	if (!swift_read_u64_le(stream, &value))
		return false;

	*value_ = swift_dec754_64(value);

	return true;
}

bool swift_write_f32_be(swift_stream_t *stream, double value)
{
	return swift_write_u32_be(stream, swift_enc754_32(value));
}

bool swift_write_f32_le(swift_stream_t *stream, float value)
{
	return swift_write_u32_le(stream, swift_enc754_32(value));
}

bool swift_write_f64_be(swift_stream_t *stream, double value)
{
	return swift_write_u64_be(stream, swift_enc754_64(value));
}

bool swift_write_f64_le(swift_stream_t *stream, double value)
{
	return swift_write_u64_le(stream, swift_enc754_64(value));
}

char *swift_read_cstr(swift_stream_t *stream)
{
	char *data;
	size_t size = 0;

	if (!stream)
		return NULL;

	swift_read_u32_be(stream, &size);

	if (!size)
		return NULL;

	if (!(data = malloc(size)))
		return NULL;

	swift_stream_read(stream, data, size);

	return data;
}

bool swift_write_cstr(swift_stream_t *stream, const char *data)
{
	if (!data) {
		swift_write_u32_be(stream, 0);
		return false;
	}
	else {
		size_t size = strlen(data) + 1;
		swift_write_u32_be(stream, size);
		return (swift_stream_write(stream, data, size) == size);
	}
}
