#include <swift/maths.h>

swift_mat4f_t swift_mat4f_zero(void)
{
	return SWIFT_MAT4F(0);
}

swift_mat4f_t swift_mat4f_identity(void)
{
	return SWIFT_MAT4F(
		.m.m00 = 1,
		.m.m11 = 1,
		.m.m22 = 1,
		.m.m33 = 1
	);
}

swift_mat4f_t swift_mat4f_from_mat3f(swift_mat3f_t mat)
{
	swift_mat4f_t ret = swift_mat4f_identity();
	size_t i;

	for (i = 0; i < 3; ++i)
		memcpy(ret.data + i * 4, mat.data + i * 3, 3 * sizeof(float));

	return ret;
}

swift_mat4f_t swift_mat4f_translation(swift_vec3f_t vec)
{
	swift_mat4f_t ret = swift_mat4f_identity();

	ret.m.m03 = vec.xyz.x;
	ret.m.m13 = vec.xyz.y;
	ret.m.m23 = vec.xyz.z;

	return ret;
}

swift_vec4f_t swift_mat4f_dot_vec4f(swift_mat4f_t lhs, swift_vec4f_t rhs)
{
	swift_vec4f_t ret = swift_vec4f_zero();
	size_t i, j;

	for (j = 0; j < 4; ++j)
		for (i = 0; i < 4; ++i)
			ret.data[j] += lhs.data[i * 4 + j] * rhs.data[i];

	return ret;
}


swift_mat4f_t swift_mat4f_dot(swift_mat4f_t lhs, swift_mat4f_t rhs)
{
	swift_mat4f_t ret = swift_mat4f_zero();
	size_t i, j, k;

	for (k = 0; k < 4; ++k)
		for (j = 0; j < 4; ++j)
			for (i = 0; i < 4; ++i)
				ret.data[i * 4 + j] += lhs.data[i * 4 + k] *
					rhs.data[k * 4 + j];

	return ret;
}

swift_mat4f_t swift_mat4f_translate(swift_mat4f_t mat, swift_vec3f_t vec)
{
	return swift_mat4f_dot(swift_mat4f_translation(vec), mat);
}

swift_mat4f_t swift_mat4f_scale(swift_mat4f_t mat, swift_vec3f_t vec)
{
	swift_mat4f_t ret = swift_mat4f_zero();
	ret.m.m00 = vec.xyz.x;
	ret.m.m11 = vec.xyz.y;
	ret.m.m22 = vec.xyz.z;
	ret.m.m33 = 1.0f;

	return swift_mat4f_dot(ret, mat);
}

swift_mat4f_t swift_mat4f_rotate(swift_mat4f_t mat, float angle,
	swift_vec3f_t axis)
{
	return swift_mat4f_dot(swift_mat4f_from_mat3f(swift_mat3f_rotation(angle,
		axis)), mat);
}

swift_mat4f_t swift_mat4f_ortho(float left, float right, float bottom,
	float top, float near, float far)
{
	return SWIFT_MAT4F(
		.m.m00 = 2 / (right - left),
		.m.m11 = 2 / (bottom - top),
		.m.m22 = -2 / (far - near),
		.m.m03 = -(right + left) / (right - left),
		.m.m13 = -(top + bottom) / (bottom - top),
		.m.m23 = -(far + near) / (far - near),
		.m.m33 = 1.0f,
	);
}

swift_mat4f_t swift_mat4f_frustum(float left, float right, float bottom,
	float top, float near, float far)
{
	return SWIFT_MAT4F(
		.m.m00 = (2 * near) / (right - left),
		.m.m11 = (2 * near) / (bottom - top),
		.m.m02 = (right + left) / (right - left),
		.m.m12 = (top + bottom) / (bottom - top),
		.m.m22 = -(far + near) / (far - near),
		.m.m23 = -(2 * far * near) / (far - near),
		.m.m32 = -1,
	);
}

swift_mat4f_t swift_mat4f_perspective(float fovy, float aspect, float near,
	float far)
{
	float width, height;

	height = tan(fovy * M_PI / 360.0) * near;
	width = height * aspect;

	return swift_mat4f_frustum(-width, width, -height, height, near, far);
}

swift_mat4f_t swift_mat4f_look_at(swift_vec3f_t eye, swift_vec3f_t center,
	swift_vec3f_t up)
{
	swift_mat4f_t ret = swift_mat4f_identity();
	swift_vec3f_t forward, side;

	forward = swift_vec3f_normalise(swift_vec3f_sub(eye, center));
	side = swift_vec3f_normalise(swift_vec3f_cross(up, forward));
	up = swift_vec3f_normalise(swift_vec3f_cross(forward, side));

	ret.m.m00 = side.xyz.x;
	ret.m.m01 = side.xyz.y;
	ret.m.m02 = side.xyz.z;

	ret.m.m10 = up.xyz.x;
	ret.m.m11 = up.xyz.y;
	ret.m.m12 = up.xyz.z;

	ret.m.m20 = forward.xyz.x;
	ret.m.m21 = forward.xyz.y;
	ret.m.m22 = forward.xyz.z;

	ret.m.m03 = swift_vec3f_dot(swift_vec3f_neg(side), eye);
	ret.m.m13 = swift_vec3f_dot(swift_vec3f_neg(up), eye);
	ret.m.m23 = swift_vec3f_dot(swift_vec3f_neg(forward), eye);

	return ret;
}

swift_mat4f_t swift_mat4f_mul_scalar(swift_mat4f_t mat, float scalar)
{
	size_t i;

	for (i = 0; i < 4 * 4; ++i)
		mat.data[i] *= scalar;

	return mat;
}

swift_mat4f_t swift_mat4f_add(swift_mat4f_t lhs, swift_mat4f_t rhs)
{
	swift_mat4f_t ret;
	size_t i;

	for (i = 0; i < 4 * 4; ++i)
		ret.data[i] = lhs.data[i] + rhs.data[i];

	return ret;
}
