#include <swift/net.h>

static void swift_close_buffer_stream(swift_stream_t *stream);
static int64_t swift_seek_buffer_stream(swift_stream_t *stream,
	unsigned int whence, int64_t pos);
static size_t swift_read_buffer_stream(swift_stream_t *stream,
	void *data, size_t size);
static size_t swift_write_buffer_stream(swift_stream_t *stream,
	const void *data, size_t size);
static bool swift_flush_buffer_stream(swift_stream_t *stream);

static struct swift_stream_ops_t buffer_stream_ops = {
	.close = swift_close_buffer_stream,
	.seek = swift_seek_buffer_stream,
	.read = swift_read_buffer_stream,
	.write = swift_write_buffer_stream,
	.flush = swift_flush_buffer_stream,
};

swift_stream_t *swift_create_buffer_stream(size_t size)
{
	swift_buffer_stream_t *buffer_stream;
	swift_stream_t *stream;

	if (!(buffer_stream = calloc(1, sizeof(swift_buffer_stream_t))))
		return NULL;

	if (size) {
		buffer_stream->size = size;

		if (!(buffer_stream->data = calloc(size, sizeof(unsigned char))))
			goto free_buffer_stream;
	}

	stream = &buffer_stream->stream;
	swift_init_ref(&stream->ref);
	stream->stream_ops = &buffer_stream_ops;

	return stream;

free_buffer_stream:
	SWIFT_FREE(buffer_stream);
	return NULL;
}

static void swift_close_buffer_stream(swift_stream_t *stream)
{
	swift_buffer_stream_t *buffer_stream = SWIFT_BUFFER_STREAM(stream);

	SWIFT_FREE(buffer_stream->data);
	SWIFT_FREE(buffer_stream);
	stream = NULL;
}

void *swift_get_buffer_stream_data(swift_stream_t *stream)
{
	swift_buffer_stream_t *buffer_stream;

	if (!stream)
		return NULL;

	buffer_stream = SWIFT_BUFFER_STREAM(stream);
	return buffer_stream->data;
}

size_t swift_get_buffer_stream_count(swift_stream_t *stream)
{
	swift_buffer_stream_t *buffer_stream;

	if (!stream)
		return 0;

	buffer_stream = SWIFT_BUFFER_STREAM(stream);
	return buffer_stream->count;
}

bool swift_resize_buffer_stream(swift_stream_t *stream, size_t size)
{
	swift_buffer_stream_t *tmp_buffer_stream, *buffer_stream;
	swift_stream_t *tmp_stream;

	if (!stream)
		return false;

	if (size > 0 && size < 16)
		size = 16;

	buffer_stream = SWIFT_BUFFER_STREAM(stream);

	if (buffer_stream->size == size)
		return true;

	if (!(tmp_stream = swift_create_buffer_stream(size)))
		return false;

	tmp_buffer_stream = SWIFT_BUFFER_STREAM(tmp_stream);

	tmp_buffer_stream->count = SWIFT_MIN(tmp_buffer_stream->size,
		buffer_stream->count);
	tmp_buffer_stream->pos = SWIFT_MIN(tmp_buffer_stream->count,
		buffer_stream->pos);
	memcpy(tmp_buffer_stream->data, buffer_stream->data,
		tmp_buffer_stream->count);

	swift_swap(&buffer_stream->data, &tmp_buffer_stream->data, sizeof(void *));
	swift_swap(&buffer_stream->pos, &tmp_buffer_stream->pos, sizeof(size_t));
	swift_swap(&buffer_stream->count, &tmp_buffer_stream->count,
		sizeof(size_t));
	swift_swap(&buffer_stream->size, &tmp_buffer_stream->size, sizeof(size_t));

	swift_close_stream(tmp_stream);

	return true;
}

bool swift_expand_buffer_stream(swift_stream_t *stream, size_t count)
{
	swift_buffer_stream_t *buffer_stream;

	if (!stream)
		return false;

	buffer_stream = SWIFT_BUFFER_STREAM(stream);

	if (count <= buffer_stream->size)
		return true;

	return swift_resize_buffer_stream(stream, swift_next_pow2_size(count));
}

static int64_t swift_seek_buffer_stream(swift_stream_t *stream,
	unsigned int whence, int64_t pos_)
{
	swift_buffer_stream_t *buffer_stream = SWIFT_BUFFER_STREAM(stream);
	int64_t pos;

	switch (whence) {
		case SWIFT_FILE_SEEK_SET: {
			pos = pos_;
		} break;

		case SWIFT_FILE_SEEK_CURRENT: {
			pos = buffer_stream->pos + pos_;
		} break;

		case SWIFT_FILE_SEEK_END: {
			pos = buffer_stream->size + pos_;
		} break;
	}

	buffer_stream->pos = SWIFT_MIN(pos, buffer_stream->size);

	return buffer_stream->pos;

}

static size_t swift_read_buffer_stream(swift_stream_t *stream,
	void *data, size_t size)
{
	swift_buffer_stream_t *buffer_stream = SWIFT_BUFFER_STREAM(stream);

	size = SWIFT_MIN(size, buffer_stream->count - buffer_stream->pos);
	memcpy(data, buffer_stream->data + buffer_stream->pos, size);
	buffer_stream->pos += size;

	return size;
}

static size_t swift_write_buffer_stream(swift_stream_t *stream,
	const void *data, size_t size)
{
	swift_buffer_stream_t *buffer_stream = SWIFT_BUFFER_STREAM(stream);

	if (!swift_expand_buffer_stream(stream, buffer_stream->count + size))
		return 0;

	memcpy(buffer_stream->data + buffer_stream->pos, data, size);
	buffer_stream->pos += size;
	buffer_stream->count = SWIFT_MAX(buffer_stream->count, buffer_stream->pos);

	return size;
}

static bool swift_flush_buffer_stream(swift_stream_t *stream)
{
	swift_buffer_stream_t *buffer_stream = SWIFT_BUFFER_STREAM(stream);

	memset(buffer_stream->data, 0, buffer_stream->count);
	buffer_stream->count = 0;
	buffer_stream->pos = 0;

	return true;
}
