#include <swift/data.h>
#include <swift/net.h>

#include <unistd.h>

#include <sys/epoll.h>

#include "epoll/epoll_monitor.h"
#include "posix/posix_sock.h"

static size_t sys_event_queue_hash(const void *data)
{
	const sys_event_queue_t *event_queue = data;

	return swift_sock_hash(event_queue->sock);
}

static bool sys_event_queue_eq(const void *lhs_, const void *rhs_)
{
	const sys_event_queue_t *lhs = lhs_;
	const sys_event_queue_t *rhs = rhs_;

	return swift_sock_eq(lhs->sock, rhs->sock);
}

sys_event_queue_t *sys_get_event_queue(swift_monitor_t *monitor,
	swift_sock_t *sock)
{
	sys_monitor_t *sys_monitor = monitor->data;
	sys_event_queue_t key = {
		.sock = sock,
	};

	return swift_htable_search(sys_monitor->event_queues, &key);
}

bool sys_register_event_queue(swift_monitor_t *monitor,
	swift_sock_t *sock)
{
	sys_monitor_t *sys_monitor = monitor->data;
	sys_event_queue_t *event_queue;

	event_queue = malloc(sizeof(*event_queue));

	if (!event_queue)
		return false;

	event_queue->sock = sock;
	swift_init_list(&event_queue->in_events);
	swift_init_list(&event_queue->out_events);

	if (swift_htable_insert(sys_monitor->event_queues, event_queue, false) ==
		SIZE_MAX)
		goto err_free_event_queue;

	return true;

err_free_event_queue:
	SWIFT_FREE(event_queue);
	return false;
}

void sys_deregister_event_queue(swift_monitor_t *monitor,
	swift_sock_t *sock)
{
	sys_monitor_t *sys_monitor = monitor->data;
	sys_event_queue_t *event_queue, key = {
		.sock = sock,
	};
	size_t index;

	index = swift_htable_search_index(sys_monitor->event_queues, &key);

	if (index == SIZE_MAX)
		return;

	event_queue = sys_monitor->event_queues->data[index];

	swift_htable_remove_index(sys_monitor->event_queues, index);
	SWIFT_FREE(event_queue);
}

bool sys_init_monitor(swift_monitor_t *monitor,
	unsigned long thread_count)
{
	sys_monitor_t *sys_monitor;

	sys_monitor = malloc(sizeof(*sys_monitor));

	if (!sys_monitor)
		return false;

	sys_monitor->handle = epoll_create1(0);

	if (sys_monitor->handle < 0)
		goto err_free_monitor;

	sys_monitor->event_queues = swift_new_htable(sys_event_queue_hash,
		sys_event_queue_eq);

	if (!sys_monitor->event_queues)
		goto err_close;

	monitor->data = sys_monitor;

	return true;

err_close:
	close(sys_monitor->handle);
err_free_monitor:
	SWIFT_FREE(sys_monitor);
	return false;
}

void sys_term_monitor(swift_monitor_t *monitor)
{
	sys_monitor_t *sys_monitor;

	sys_monitor = monitor->data;

	swift_del_htable(sys_monitor->event_queues);
	close(sys_monitor->handle);
	SWIFT_FREE(sys_monitor);
}

bool sys_monitor_sock(swift_monitor_t *monitor,
	swift_sock_t *sock)
{
	epoll_event ev = { 0 };
	sys_sock_t *sys_sock = sock->data;
	sys_monitor_t *sys_monitor = monitor->data;
	unsigned int flags;

	if (!sys_register_event_queue(monitor, sock))
		return false;

	flags = swift_sock_get_flags(sock);
	flags &= ~SWIFT_SOCK_WAIT;
	swift_sock_set_flags(sock, flags);

	ev.data.ptr = sock;
	ev.events = 0;

	return (epoll_ctl(sys_monitor->handle, EPOLL_CTL_ADD, sys_sock->handle,
		&ev) == 0);
}

void sys_deregister_sock(swift_monitor_t *monitor,
	swift_sock_t *sock)
{
	sys_sock_t *sys_sock = sock->data;
	sys_monitor_t *sys_monitor = monitor->data;

	epoll_ctl(sys_monitor->handle, EPOLL_CTL_DEL, sys_sock->handle, NULL);
	sys_deregister_event_queue(monitor, sock);
}

bool sys_update_sock(swift_sock_t *sock)
{
	epoll_event ev = { 0 };
	swift_monitor_t *monitor = sock->monitor;
	sys_sock_t *sys_sock = sock->data;
	sys_monitor_t *sys_monitor = monitor->data;
	sys_event_queue_t *event_queue;
	bool in_events, out_events;

	event_queue = sys_get_event_queue(sock->monitor, sock);

	in_events = !swift_list_is_empty(&event_queue->in_events);
	out_events = !swift_list_is_empty(&event_queue->out_events);

	if (!in_events && !out_events)
		return true;

	ev.data.ptr = sock;
	ev.events = EPOLLRDHUP | EPOLLET | EPOLLONESHOT;
	ev.events |= (in_events) ? EPOLLIN : 0;
	ev.events |= (out_events) ? EPOLLOUT : 0;

	return (epoll_ctl(sys_monitor->handle, EPOLL_CTL_MOD, sys_sock->handle,
		&ev) == 0);
}

void sys_poll_monitor(swift_monitor_t *monitor)
{
	sys_monitor_t *sys_monitor = monitor->data;
	swift_sock_t *sock;
	sys_sock_t *sys_sock;
	sys_event_queue_t *event_queue;
	sys_event_t *event;
	swift_list_t *node;
	epoll_event ev;

	while (epoll_wait(sys_monitor->handle, &ev, 1, 0) > 0) {
		sock = ev.data.ptr;
		sys_sock = sock->data;

		swift_ref_sock(sock);

		event_queue = sys_get_event_queue(monitor, sock);

		node = NULL;

		if (ev.events & (EPOLLHUP | EPOLLRDHUP)) {
			if (!swift_list_is_empty(&event_queue->in_events))
				node = swift_list_shift(&event_queue->in_events);
			else
				node = swift_list_shift(&event_queue->out_events);
		} else if (ev.events & EPOLLIN) {
			node = swift_list_shift(&event_queue->in_events);
		} else if (ev.events & EPOLLOUT) {
			node = swift_list_shift(&event_queue->out_events);
		}

		if (!node) {
			ev.events = EPOLLRDHUP | EPOLLET | EPOLLONESHOT;
			epoll_ctl(sys_monitor->handle, EPOLL_CTL_MOD, sys_sock->handle,
				&ev);

			continue;
		}

		event = SWIFT_CONTAINER_OF(node, sys_event_t, node);
		event->on_complete(true, sock, event);

		sys_update_sock(sock);
		swift_del_sock(sock);
	}
}
