#include <swift/data.h>
#include <swift/net.h>

#include <errno.h>
#include <unistd.h>

#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "monitor.h"

#include "epoll/epoll_monitor.h"
#include "epoll/epoll_sock.h"

#include "posix/posix_sock.h"

static void sys_on_accept(bool success, swift_sock_t *sock, sys_event_t *event)
{
	sys_accept_event_t *accept_event = SWIFT_CONTAINER_OF(event,
		sys_accept_event_t, event);
	swift_sock_t *peer = NULL;

	if (!success)
		goto err_skip;

	peer = swift_sock_accept(sock);

	if (!peer)
		goto err_free_event;

err_skip:
	accept_event->on_accept(success, sock, peer, accept_event->user_data);

err_free_event:
	swift_del_sock(accept_event->sock);
	SWIFT_FREE(accept_event);
}

static void sys_on_connect(bool success, swift_sock_t *sock, sys_event_t *event)
{
	sys_sock_t *sys_sock = sock->data;
	sys_connect_event_t *connect_event = SWIFT_CONTAINER_OF(event,
		sys_connect_event_t, event);
	socklen_t len = sizeof(int);
	int ret;

	if (!success)
		goto err_skip;

	if (getsockopt(sys_sock->handle, SOL_SOCKET, SO_ERROR, &ret, &len) < 0)
		goto err_free_event;

	if (ret != 0)
		goto err_free_event;

err_skip:
	connect_event->on_connect(success, sock, connect_event->user_data);

err_free_event:
	swift_del_addr(connect_event->addr);
	swift_del_sock(connect_event->sock);
	SWIFT_FREE(connect_event);
}

static void sys_on_recv(bool success, swift_sock_t *sock, sys_event_t *event)
{
	sys_event_queue_t *event_queue;
	sys_recv_event_t *recv_event = SWIFT_CONTAINER_OF(event,
		sys_recv_event_t, event);
	size_t ret;

	if (!success)
		goto err_callback;

	ret = swift_sock_recv(sock, recv_event->data, recv_event->size);

	if (ret) {
		recv_event->on_recv(true, sock, recv_event->data, ret,
			recv_event->user_data);
		goto err_free_event;
	}

	if (swift_get_errno() != SWIFT_EAGAIN) {
		success = false;
		goto err_callback;
	}

	event_queue = sys_get_event_queue(sock->monitor, sock);
	swift_list_push(&event_queue->in_events, &event->node);

	return;

err_callback:
	recv_event->on_recv(false, sock, NULL, 0, recv_event->user_data);
err_free_event:
	swift_del_sock(recv_event->sock);
	SWIFT_FREE(recv_event);
}

static void sys_on_send(bool success, swift_sock_t *sock, sys_event_t *event)
{
	sys_send_event_t *send_event = SWIFT_CONTAINER_OF(event,
		sys_send_event_t, event);
	size_t ret;

	if (!success)
		goto err_callback;

	ret = swift_sock_send(sock, send_event->data, send_event->size);

	if (ret)
		goto err_callback;

	if (swift_get_errno() != SWIFT_EAGAIN) {
		success = false;
		goto err_callback;
	}

err_callback:
	send_event->on_send(success, sock, send_event->data, ret,
		send_event->user_data);
	swift_del_sock(send_event->sock);
	SWIFT_FREE(send_event);
}

bool sys_queue_accept(swift_sock_t *sock, swift_on_accept_fn_t on_accept,
	void *user_data)
{
	swift_monitor_t *monitor = sock->monitor;
	sys_event_queue_t *event_queue;
	sys_accept_event_t *accept_event;
	sys_event_t *event;

	accept_event = malloc(sizeof(*accept_event));

	if (!accept_event)
		return false;

	accept_event->sock = swift_ref_sock(sock);
	accept_event->on_accept = on_accept;
	accept_event->user_data = user_data;

	event = &accept_event->event;
	event->on_complete = sys_on_accept;

	event_queue = sys_get_event_queue(monitor, sock);
	swift_list_push(&event_queue->in_events, &event->node);
	sys_update_sock(sock);

	return true;
}

bool sys_queue_connect(swift_sock_t *sock, swift_addr_t *addr,
	swift_on_connect_fn_t on_connect, void *user_data)
{
	swift_monitor_t *monitor = sock->monitor;
	sys_event_queue_t *event_queue;
	sys_connect_event_t *connect_event;
	sys_event_t *event;

	if (swift_sock_connect(sock, addr)) {
		on_connect(true, sock, user_data);
		return true;
	}

	if (swift_get_errno() != SWIFT_EINPROGRESS)
		return false;

	connect_event = malloc(sizeof(*connect_event));

	if (!connect_event)
		return false;

	connect_event->sock = swift_ref_sock(sock);
	connect_event->addr = swift_ref_addr(addr);
	connect_event->on_connect = on_connect;
	connect_event->user_data = user_data;

	event = &connect_event->event;
	event->on_complete = sys_on_connect;

	event_queue = sys_get_event_queue(monitor, sock);
	swift_list_push(&event_queue->out_events, &event->node);
	sys_update_sock(sock);

	return true;
}

bool sys_queue_recv(swift_sock_t *sock, void *data, size_t size,
	swift_on_recv_fn_t on_recv, void *user_data)
{
	swift_monitor_t *monitor = sock->monitor;
	sys_event_queue_t *event_queue;
	sys_recv_event_t *recv_event;
	sys_event_t *event;
	size_t ret;

	ret = swift_sock_recv(sock, data, size);

	if (ret) {
		on_recv(true, sock, data, ret, user_data);
		return true;
	}

	if (swift_get_errno() != SWIFT_EAGAIN)
		return false;

	recv_event = malloc(sizeof(*recv_event));

	if (!recv_event)
		return false;

	recv_event->sock = swift_ref_sock(sock);
	recv_event->data = data;
	recv_event->size = size;
	recv_event->on_recv = on_recv;
	recv_event->user_data = user_data;

	event = &recv_event->event;
	event->on_complete = sys_on_recv;

	event_queue = sys_get_event_queue(monitor, sock);
	swift_list_push(&event_queue->in_events, &event->node);
	sys_update_sock(sock);

	return true;
}

bool sys_queue_send(swift_sock_t *sock, const void *data, size_t size,
	swift_on_send_fn_t on_send, void *user_data)
{
	swift_monitor_t *monitor = sock->monitor;
	sys_event_queue_t *event_queue;
	sys_send_event_t *send_event;
	sys_event_t *event;
	size_t ret;

	ret = swift_sock_send(sock, data, size);

	if (ret) {
		on_send(true, sock, data, ret, user_data);
		return true;
	}

	if (swift_get_errno() != SWIFT_EAGAIN)
		return false;

	send_event = malloc(sizeof(*send_event));

	if (!send_event)
		return false;

	send_event->sock = swift_ref_sock(sock);
	send_event->data = data;
	send_event->size = size;
	send_event->on_send = on_send;
	send_event->user_data = user_data;

	event = &send_event->event;
	event->on_complete = sys_on_send;

	event_queue = sys_get_event_queue(monitor, sock);
	swift_list_push(&event_queue->out_events, &event->node);
	sys_update_sock(sock);

	return true;
}
