#include <swift/net.h>

#define WIN32_MEAN_AND_LEAN
#include <winsock2.h>
#include <windows.h>
#include <ws2tcpip.h>
#include <mswsock.h>

#include "msw_addr.h"
#include "msw_monitor.h"
#include "msw_sock.h"

static bool sys_post_init_sock(swift_sock_t *sock)
{
	GUID guid;
	sys_sock_t *sys_sock = sock->data;
	int len = sizeof(int);
	int value = 1;
	DWORD bytes_ret;

	setsockopt(sys_sock->handle, SOL_SOCKET, SO_REUSEADDR, (char *)&value,
		len);

	guid = (GUID)WSAID_CONNECTEX;
	WSAIoctl(sys_sock->handle, SIO_GET_EXTENSION_FUNCTION_POINTER,
		&guid, sizeof(guid), &sys_sock->connect,
		sizeof(sys_sock->connect), &bytes_ret, NULL, NULL);

	guid = (GUID)WSAID_ACCEPTEX;
	WSAIoctl(sys_sock->handle, SIO_GET_EXTENSION_FUNCTION_POINTER,
		&guid, sizeof(guid), &sys_sock->accept,
		sizeof(sys_sock->accept), &bytes_ret, NULL, NULL);

	guid = (GUID)WSAID_GETACCEPTEXSOCKADDRS;
	WSAIoctl(sys_sock->handle, SIO_GET_EXTENSION_FUNCTION_POINTER,
		&guid, sizeof(guid), &sys_sock->get_accept_sockaddrs,
		sizeof(sys_sock->get_accept_sockaddrs), &bytes_ret, NULL, NULL);

	return true;
}

static void sys_on_connect(DWORD error, DWORD transferred,
	sys_event_t *event)
{
	sys_connect_event_t *connect_event = SWIFT_CONTAINER_OF(event,
		sys_connect_event_t, event);
	sys_sock_t *sys_sock = connect_event->sock->data;

	if (setsockopt(sys_sock->handle, SOL_SOCKET, SO_UPDATE_CONNECT_CONTEXT,
		NULL, 0) < 0) {
		connect_event->on_connect(false, connect_event->sock,
			connect_event->user_data);

		goto err_free_event;
	}

	connect_event->on_connect(true, connect_event->sock,
		connect_event->user_data);

err_free_event:
	swift_del_addr(connect_event->addr);
	swift_del_sock(connect_event->sock);
	SWIFT_FREE(connect_event);
}

static void sys_on_accept(DWORD error, DWORD transferred,
	sys_event_t *event)
{
	sys_accept_event_t *accept_event = SWIFT_CONTAINER_OF(event,
		sys_accept_event_t, event);
	sys_sock_t *sys_sock = accept_event->sock->data;
	sys_sock_t *sys_peer = accept_event->peer->data;
	struct sockaddr *local_addr, *remote_addr;
	socklen_t local_addr_size, remote_addr_size;

	if (error != ERROR_SUCCESS)
		goto err_callback;

	if (setsockopt(sys_peer->handle, SOL_SOCKET, SO_UPDATE_ACCEPT_CONTEXT,
		(char *)&sys_sock->handle, sizeof(sys_sock->handle)) < 0)
		goto err_callback;

	sys_sock->get_accept_sockaddrs(accept_event->buffer, 0,
		accept_event->local_addr_size, accept_event->remote_addr_size,
		&local_addr, &local_addr_size, &remote_addr, &remote_addr_size);

	accept_event->on_accept(true, accept_event->sock, accept_event->peer,
		accept_event->user_data);

	goto err_free_event;

err_callback:
	accept_event->on_accept(false, accept_event->sock, NULL,
		accept_event->user_data);
err_free_event:
	swift_del_sock(accept_event->peer);
	SWIFT_FREE(accept_event->buffer);
	SWIFT_FREE(accept_event);
}

static void sys_on_recv(DWORD error, DWORD transferred,
	sys_event_t *event)
{
	sys_recv_event_t *recv_event = SWIFT_CONTAINER_OF(event,
		sys_recv_event_t, event);

	if (error != ERROR_SUCCESS)
		goto err_callback;

	if (transferred == 0)
		goto err_callback;

	recv_event->on_recv(true, recv_event->sock,
		recv_event->buf.buf, transferred, recv_event->user_data);

	goto err_free_event;

err_callback:
	recv_event->on_recv(false, recv_event->sock, NULL, 0,
		recv_event->user_data);
err_free_event:
	swift_del_sock(recv_event->sock);
	SWIFT_FREE(recv_event);
}

static void sys_on_send(DWORD error, DWORD transferred,
	sys_event_t *event)
{
	sys_send_event_t *send_event = SWIFT_CONTAINER_OF(event,
		sys_send_event_t, event);

	if (error != ERROR_SUCCESS)
		goto err_callback;

	if (transferred == 0)
		goto err_callback;

	send_event->on_send(true, send_event->sock,
		send_event->buf.buf, transferred, send_event->user_data);

	goto err_free_event;

err_callback:
	send_event->on_send(false, send_event->sock, NULL, 0,
		send_event->user_data);
err_free_event:
	swift_del_sock(send_event->sock);
	SWIFT_FREE(send_event);
}

bool sys_init_sock(swift_sock_t *sock, unsigned int addr_type,
	unsigned int proto_type)
{
	sys_sock_t *sys_sock;
	int domain, type, proto;

	switch (addr_type) {
		case SWIFT_ADDR_IPV4: {
			domain = AF_INET;
		} break;

		case SWIFT_ADDR_IPV6: {
			domain = AF_INET6;
		} break;

		default: {
			return false;
		} break;
	}

	switch (proto_type) {
		case SWIFT_PROTO_TCP: {
			type = SOCK_STREAM;
			proto = IPPROTO_TCP;
		} break;

		case SWIFT_PROTO_UDP: {
			type = SOCK_DGRAM;
			proto = IPPROTO_UDP;
		} break;

		default: {
			return false;
		} break;
	}

	sys_sock = malloc(sizeof(*sys_sock));

	if (!sys_sock)
		return false;

	sys_sock->handle = WSASocket(domain, type, proto, NULL, 0,
		WSA_FLAG_OVERLAPPED);

	if (sys_sock->handle == INVALID_SOCKET)
		goto err_free_sock;

	if (addr_type == SWIFT_ADDR_IPV6) {
		int value = 1;

		setsockopt(sys_sock->handle, IPPROTO_IPV6, IPV6_V6ONLY,
			(char *)&value, sizeof(value));
	}

	sock->data = sys_sock;

	if (!sys_post_init_sock(sock))
		goto err_close_sock;

	return true;

err_close_sock:
	closesocket(sys_sock->handle);
err_free_sock:
	SWIFT_FREE(sys_sock);
	return false;
}

bool sys_init_sock_from(swift_sock_t *sock, const void *data)
{
	sys_sock_t *sys_sock;

	sys_sock = malloc(sizeof(*sys_sock));

	if (!sys_sock)
		return false;

	sys_sock->handle = *((SOCKET *)data);

	if (sys_sock->handle == INVALID_SOCKET)
		goto err_free_sock;

	sock->data = sys_sock;

	if (!sys_post_init_sock(sock))
		goto err_free_sock;

	return true;

err_free_sock:
	SWIFT_FREE(sys_sock);
	return false;
}

void sys_term_sock(swift_sock_t *sock)
{
	sys_sock_t *sys_sock = sock->data;

	if (!(sock->flags & SWIFT_FOREIGN_SOCK))
		closesocket(sys_sock->handle);

	SWIFT_FREE(sys_sock);
}

size_t sys_sock_hash(const void *data)
{
	const swift_sock_t *sock = *(const swift_sock_t **)data;
	const sys_sock_t *sys_sock = sock->data;

	return swift_fnv1a_hash64(&sys_sock->handle, sizeof(sys_sock->handle));
}

bool sys_sock_eq(const void *lhs_, const void *rhs_)
{
	const swift_sock_t *ldata = *(const swift_sock_t **)lhs_;
	const swift_sock_t *rdata = *(const swift_sock_t **)rhs_;
	const sys_sock_t *lhs = ldata->data;
	const sys_sock_t *rhs = rdata->data;

	return (lhs->handle == rhs->handle);
}

size_t sys_peer_sock_hash(const void *data)
{
	const swift_peer_t *peer = *(const swift_peer_t **)data;
	const swift_sock_t *sock = peer->sock;
	const sys_sock_t *sys_sock = sock->data;

	return swift_fnv1a_hash64(&sys_sock->handle, sizeof(sys_sock->handle));
}

bool sys_peer_sock_eq(const void *lhs_, const void *rhs_)
{
	const swift_peer_t *lhs_peer = *(const swift_peer_t **)lhs_;
	const swift_peer_t *rhs_peer = *(const swift_peer_t **)rhs_;
	const swift_sock_t *lhs = lhs_peer->sock;
	const swift_sock_t *rhs = rhs_peer->sock;
	const sys_sock_t *sys_lhs = lhs->data;
	const sys_sock_t *sys_rhs = rhs->data;

	return (sys_lhs->handle == sys_rhs->handle);
}

unsigned int sys_sock_get_flags(swift_sock_t *sock)
{
	/* TODO: track blocking/non-blocking mode manually. */
	return 0;
}

bool sys_sock_set_flags(swift_sock_t *sock, unsigned int flags)
{
	sys_sock_t *sys_sock = sock->data;
	unsigned long mode;

	mode = (flags & SWIFT_SOCK_WAIT) ? 0 : 1;

	return (ioctlsocket(sys_sock->handle, FIONBIO, &mode) == 0);
}

bool sys_sock_local_addr(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;
	socklen_t size = sys_addr_sockaddr_size(addr);

	return (getsockname(sys_sock->handle, sys_addr_sockaddr(addr), &size) ==
		0);
}

bool sys_sock_remote_addr(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;
	socklen_t size = sys_addr_sockaddr_size(addr);

	return (getpeername(sys_sock->handle, sys_addr_sockaddr(addr), &size) ==
		0);
}

bool sys_queue_connect(swift_sock_t *sock, swift_addr_t *addr,
	swift_on_connect_fn_t on_connect, void *user_data)
{
	sys_sock_t *sys_sock = sock->data;
	sys_connect_event_t *connect_event;
	sys_event_t *event;
	DWORD transferred;

	connect_event = calloc(1, sizeof(*connect_event));

	if (!connect_event)
		return false;

	connect_event->sock = swift_ref_sock(sock);
	connect_event->addr = swift_ref_addr(addr);
	connect_event->on_connect = on_connect,
	connect_event->user_data = user_data;

	event = &connect_event->event;

	event->on_complete = sys_on_connect;

	if (!swift_sock_bind_name(sock, NULL, 0))
		goto err_free_event;

	if (sys_sock->connect(sys_sock->handle, sys_addr_sockaddr(addr),
		sys_addr_sockaddr_size(addr), NULL, 0, &transferred,
		&event->overlapped))
		return true;

	if (WSAGetLastError() != ERROR_IO_PENDING)
		goto err_free_event;

	return true;

err_free_event:
	swift_del_addr(connect_event->addr);
	swift_del_sock(connect_event->sock);
	SWIFT_FREE(connect_event);
	return false;
}

bool sys_sock_connect(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;

	return (connect(sys_sock->handle, sys_addr_sockaddr(addr),
		sys_addr_sockaddr_size(addr)) == 0);
}

bool sys_sock_bind(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;

	return (bind(sys_sock->handle, sys_addr_sockaddr(addr),
		sys_addr_sockaddr_size(addr)) == 0);
}

bool sys_sock_listen(swift_sock_t *sock, unsigned int back_log)
{
	sys_sock_t *sys_sock = sock->data;

	back_log = SWIFT_MIN(back_log, SOMAXCONN);

	return (listen(sys_sock->handle, back_log) == 0);
}

swift_sock_t *sys_sock_accept(swift_sock_t *sock)
{
	sys_sock_t *sys_sock = sock->data;
	SOCKET handle = accept(sys_sock->handle, NULL, 0);

	return (handle == INVALID_SOCKET) ? NULL : swift_new_sock_from(
		sock->addr_type, sock->proto_type, &handle);
}

bool sys_queue_accept(swift_sock_t *sock, swift_on_accept_fn_t on_accept,
	void *user_data)
{
	sys_sock_t *sys_sock = sock->data;
	sys_sock_t *sys_peer;
	sys_accept_event_t *accept_event;
	sys_event_t *event;
	swift_addr_t *addr;
	DWORD transferred;

	accept_event = calloc(1, sizeof(*accept_event));

	if (!accept_event)
		return false;

	accept_event->sock = swift_ref_sock(sock);
	accept_event->on_accept = on_accept;
	accept_event->user_data = user_data;

	addr = swift_new_addr(sock->addr_type);

	if (!addr)
		goto err_free_event;

	accept_event->local_addr_size = sys_addr_sockaddr_size(addr) + 16;
	accept_event->remote_addr_size = sys_addr_sockaddr_size(addr) + 16;

	swift_del_addr(addr);

	accept_event->buffer = calloc(accept_event->local_addr_size +
		accept_event->remote_addr_size, sizeof(char));

	if (!accept_event->buffer)
		goto err_free_event;

	accept_event->peer = swift_new_sock(sock->addr_type, sock->proto_type);

	if (!accept_event->peer)
		goto err_free_buffer;

	sys_peer = accept_event->peer->data;

	event = &accept_event->event;

	event->on_complete = sys_on_accept;

	if (sys_sock->accept(sys_sock->handle, sys_peer->handle,
		accept_event->buffer, 0, accept_event->local_addr_size,
		accept_event->remote_addr_size, &transferred,
		&event->overlapped))
		return true;

	if (WSAGetLastError() != ERROR_IO_PENDING)
		goto err_del_peer;

	return true;

err_del_peer:
	swift_del_sock(accept_event->peer);
err_free_buffer:
	SWIFT_FREE(accept_event->buffer);
err_free_event:
	SWIFT_FREE(accept_event);
	return false;
}

size_t sys_sock_recv(swift_sock_t *sock, void *data, size_t size)
{
	sys_sock_t *sys_sock = sock->data;
	int ret = recv(sys_sock->handle, data, size, 0);

	return (ret == SOCKET_ERROR) ? 0 : ret;
}

bool sys_queue_recv(swift_sock_t *sock, void *data, size_t size,
	swift_on_recv_fn_t on_recv, void *user_data)
{
	sys_sock_t *sys_sock = sock->data;
	sys_recv_event_t *recv_event;
	sys_event_t *event;
	DWORD transferred, flags = 0;

	recv_event = calloc(1, sizeof(*recv_event));

	if (!recv_event)
		return false;

	recv_event->sock = swift_ref_sock(sock);
	recv_event->buf.buf = data;
	recv_event->buf.len = size;
	recv_event->on_recv = on_recv,
	recv_event->user_data = user_data;

	event = &recv_event->event;

	event->on_complete = sys_on_recv;

	if (WSARecv(sys_sock->handle, &recv_event->buf, 1, &transferred,
		&flags, &event->overlapped, NULL) == 0)
		return true;

	if (WSAGetLastError() != ERROR_IO_PENDING)
		goto err_free_event;

	return true;

err_free_event:
	swift_del_sock(recv_event->sock);
	SWIFT_FREE(recv_event);
	return false;
}

size_t sys_sock_send(swift_sock_t *sock, const void *data, size_t size)
{
	sys_sock_t *sys_sock = sock->data;
	int ret = send(sys_sock->handle, data, size, 0);

	return (ret == SOCKET_ERROR) ? 0 : ret;
}

bool sys_queue_send(swift_sock_t *sock, void *data, size_t size,
	swift_on_send_fn_t on_send, void *user_data)
{
	sys_sock_t *sys_sock = sock->data;
	sys_send_event_t *send_event;
	sys_event_t *event;
	DWORD transferred, flags = 0;

	send_event = calloc(1, sizeof(*send_event));

	if (!send_event)
		return false;

	send_event->sock = swift_ref_sock(sock);
	send_event->buf.buf = data;
	send_event->buf.len = size;
	send_event->on_send = on_send,
	send_event->user_data = user_data;

	event = &send_event->event;

	event->on_complete = sys_on_send;

	if (WSASend(sys_sock->handle, &send_event->buf, 1, &transferred,
		flags, &event->overlapped, NULL) == 0)
		return true;

	if (WSAGetLastError() != ERROR_IO_PENDING)
		goto err_free_event;

	return true;

err_free_event:
	swift_del_sock(send_event->sock);
	SWIFT_FREE(send_event);
	return false;
}
