#include <limits.h>

#include <swift/data.h>
#include <swift/net.h>

typedef struct swift_send_data_t swift_send_data_t;

struct swift_send_data_t {
	swift_net_mgr_t *net_mgr;
	void *user_data;
};

static void swift_on_accept(bool success, swift_sock_t *sock,
	swift_sock_t *peer_sock, void *user_data);
static void swift_on_recv(bool success, swift_sock_t *sock,
	void *data, size_t size, void *user_data);
static void swift_on_send(bool success, swift_sock_t *sock,
	const void *data, size_t size, void *user_data);

static void swift_on_connect(bool success, swift_sock_t *sock,
	void *user_data)
{
	swift_net_mgr_t *net_mgr = user_data;
	swift_peer_t *peer;

	peer = swift_get_peer(net_mgr, sock);

	if (!success)
		goto err_del_peer;

	peer->is_connected = true;
	net_mgr->ops->on_connect(net_mgr, peer);

	if (!swift_queue_recv(peer->sock, peer->data, peer->size,
		swift_on_recv, net_mgr))
		goto err_del_peer;

	return;

err_del_peer:
	swift_del_peer(net_mgr, peer);
}

static void swift_on_accept(bool success, swift_sock_t *sock,
	swift_sock_t *peer_sock, void *user_data)
{
	swift_net_mgr_t *net_mgr = user_data;
	swift_peer_t *peer;

	if (!success)
		goto err_queue_accept;

	peer = swift_new_peer_from(net_mgr, peer_sock);

	if (!peer) {
		swift_del_sock(peer_sock);
		goto err_queue_accept;
	}

	peer->is_connected = true;
	net_mgr->ops->on_connect(net_mgr, peer);

	if (!swift_queue_recv(peer->sock, peer->data, peer->size, swift_on_recv,
		net_mgr))
		goto err_del_peer;

	swift_queue_accept(sock, swift_on_accept, net_mgr);

	return;

err_del_peer:
	swift_del_peer(net_mgr, peer);
err_queue_accept:
	swift_queue_accept(sock, swift_on_accept, net_mgr);
}

static void swift_on_recv(bool success, swift_sock_t *sock,
	void *data, size_t size, void *user_data)
{
	swift_net_mgr_t *net_mgr = user_data;
	swift_peer_t *peer;

	peer = swift_get_peer(net_mgr, sock);

	if (!success)
		goto err_del_peer;

	net_mgr->ops->on_recv(net_mgr, peer, data, size);

	if (!peer || !swift_queue_recv(peer->sock, peer->data, peer->size,
		swift_on_recv, net_mgr))
		goto err_del_peer;

	return;

err_del_peer:
	swift_del_peer(net_mgr, peer);
}

static void swift_on_send(bool success, swift_sock_t *sock,
	const void *data, size_t size, void *user_data)
{
	swift_send_data_t *send_data = user_data;
	swift_net_mgr_t *net_mgr = send_data->net_mgr;
	swift_peer_t *peer;

	user_data = send_data->user_data;
	SWIFT_FREE(send_data);

	peer = swift_get_peer(net_mgr, sock);

	if (!success)
		goto err_del_peer;

	net_mgr->ops->on_send(net_mgr, peer, data, size, user_data);

	return;

err_del_peer:
	swift_del_peer(net_mgr, peer);
}

swift_net_mgr_t *swift_new_net_mgr(swift_net_mgr_ops_t *ops,
	unsigned int thread_count, void *user_data)
{
	swift_net_mgr_t *net_mgr;

	if (!ops || !ops->on_connect || !ops->on_disconnect || !ops->on_recv ||
		!ops->on_send)
		return NULL;

	net_mgr = calloc(1, sizeof(*net_mgr));

	if (!net_mgr)
		return NULL;

	net_mgr->ops = ops;
	net_mgr->user_data = user_data;
	net_mgr->monitor = swift_new_monitor(thread_count);

	if (!net_mgr->monitor)
	if (!net_mgr->monitor)
		goto err_free_net_mgr;

	net_mgr->listeners = swift_new_htable(swift_sock_hash, swift_sock_eq,
		sizeof(swift_sock_t *));

	if (!net_mgr->listeners)
		goto err_del_monitor;

	net_mgr->peers = swift_new_htable(swift_peer_sock_hash, swift_peer_sock_eq,
		sizeof(swift_peer_t *));

	if (!net_mgr->peers)
		goto err_del_listeners;

	return net_mgr;

err_del_listeners:
	swift_del_htable(net_mgr->listeners);
err_del_monitor:
	swift_del_monitor(net_mgr->monitor);
err_free_net_mgr:
	SWIFT_FREE(net_mgr);
	return NULL;
}

void swift_del_net_mgr(swift_net_mgr_t *net_mgr)
{
	if (!net_mgr)
		return;

	swift_del_htable(net_mgr->peers);
	swift_del_htable(net_mgr->listeners);
	swift_del_monitor(net_mgr->monitor);
	SWIFT_FREE(net_mgr);
}

bool swift_add_listener(swift_net_mgr_t *net_mgr, swift_addr_t *addr)
{
	swift_sock_t *sock;

	if (!net_mgr || !addr)
		return false;

	sock = swift_new_sock(addr->addr_type, SWIFT_PROTO_TCP);

	if (!sock)
		return false;

	if (!swift_sock_bind(sock, addr))
		goto err_del_sock;

	if (!swift_sock_listen(sock, INT_MAX))
		goto err_del_sock;

	if (!swift_monitor_sock(net_mgr->monitor, sock))
		goto err_del_sock;

	if (swift_htable_insert(net_mgr->listeners, &sock) == SIZE_MAX)
		goto err_del_sock;

	if (!swift_queue_accept(sock, swift_on_accept, net_mgr))
		goto err_remove_listener;

	return true;

err_remove_listener:
	swift_htable_remove(net_mgr->listeners, sock);
err_del_sock:
	swift_del_sock(sock);
	return false;
}

size_t swift_add_listeners(swift_net_mgr_t *net_mgr, const char *host,
	uint16_t port)
{
	swift_array_iter_t arr_iter;
	swift_array_t *addrs;
	size_t count = 0;
	swift_addr_t *addr;

	if (!net_mgr)
		return 0;

	addrs = swift_resolve_addr(SWIFT_ADDR_ANY, host, port, 0);

	if (!addrs)
		return 0;


	for (arr_iter = swift_array_first(addrs); arr_iter.data;
		swift_array_next(&arr_iter)) {
		addr = *(swift_addr_t **)arr_iter.data;

		if (!swift_add_listener(net_mgr, addr))
			continue;

		++count;
	}

	swift_del_addrs(addrs);

	return count;
}

swift_peer_t *swift_new_peer_from(swift_net_mgr_t *net_mgr,
	swift_sock_t *sock)
{
	swift_peer_t *peer;

	if (!net_mgr || !sock)
		return NULL;

	peer = malloc(sizeof(*peer));

	if (!peer)
		return NULL;

	peer->is_connected = false;
	peer->sock = sock;
	peer->size = 8192;
	peer->data = calloc(peer->size, sizeof(char));

	if (!peer->data)
		goto err_free_peer;

	if (!swift_monitor_sock(net_mgr->monitor, peer->sock))
		goto err_free_data;

	if (swift_htable_insert(net_mgr->peers, &peer) == SIZE_MAX)
		goto err_free_data;

	return peer;

err_free_data:
	SWIFT_FREE(peer->data);
err_free_peer:
	SWIFT_FREE(peer);
	return NULL;
}

void swift_del_peer(swift_net_mgr_t *net_mgr, swift_peer_t *peer)
{
	if (!net_mgr || !peer)
		return;

	if (peer->is_connected)
		net_mgr->ops->on_disconnect(net_mgr, peer);

	swift_htable_remove(net_mgr->peers, &peer);
	SWIFT_FREE(peer->data);
	swift_del_sock(peer->sock);
	SWIFT_FREE(peer);
	peer = NULL;
}

swift_peer_t *swift_get_peer(swift_net_mgr_t *net_mgr, swift_sock_t *sock)
{
	void *entry;
	swift_peer_t key_ = {.sock = sock};
	swift_peer_t *key = &key_;

	entry = swift_htable_search(net_mgr->peers, &key);

	if (!entry)
		return NULL;

	return *(swift_peer_t **)entry;
}

bool swift_add_peer(swift_net_mgr_t *net_mgr, swift_addr_t *addr)
{
	swift_sock_t *sock;
	swift_peer_t *peer;

	if (!net_mgr || !addr)
		return false;

	sock = swift_new_sock(addr->addr_type, SWIFT_PROTO_TCP);

	if (!sock)
		return false;

	peer = swift_new_peer_from(net_mgr, sock);

	if (!peer)
		goto err_del_sock;

	if (!swift_queue_connect(peer->sock, addr, swift_on_connect, net_mgr))
		goto err_del_peer;

	return true;

err_del_peer:
	swift_del_peer(net_mgr, peer);
err_del_sock:
	swift_del_sock(sock);
	return false;
}

bool swift_add_peers(swift_net_mgr_t *net_mgr, const char *host,
	uint16_t port)
{
	swift_array_iter_t arr_iter;
	swift_array_t *addrs;
	swift_addr_t *addr;

	if (!net_mgr || !host)
		return false;

	addrs = swift_resolve_addr(SWIFT_ADDR_ANY, host, port, 0);

	if (!addrs)
		return false;

	for (arr_iter = swift_array_first(addrs); arr_iter.data;
		swift_array_next(&arr_iter)) {
		addr = *(swift_addr_t **)arr_iter.data;

		if (swift_add_peer(net_mgr, addr))
			return true;
	}

	swift_del_addrs(addrs);

	return false;
}

bool swift_peer_send(swift_net_mgr_t *net_mgr,
	swift_peer_t *peer, const void *data, size_t size,
	void *user_data)
{
	swift_send_data_t *send_data;

	if (!net_mgr || !peer || !data || !size)
		return false;

	send_data = calloc(1, sizeof(swift_send_data_t));

	if (!send_data)
		return false;

	send_data->net_mgr = net_mgr;
	send_data->user_data = user_data;

	if (!swift_queue_send(peer->sock, data, size, swift_on_send, send_data))
		goto err_free_send_data;

	return true;

err_free_send_data:
	SWIFT_FREE(send_data);
	return false;
}

size_t swift_peer_broadcast(swift_net_mgr_t *net_mgr, const void *data,
	size_t size, void *user_data)
{
	swift_htable_iter_t iter;
	swift_peer_t *peer;
	size_t count;

	if (!net_mgr || !data || !size)
		return false;

	for (iter = swift_htable_first(net_mgr->peers); iter.data;
		swift_htable_next(&iter)) {
		peer = *(swift_peer_t **)iter.data;

		if (swift_peer_send(net_mgr, peer, data, size, user_data))
			++count;
	}

	return count;
}
