#include <swift/string.h>
#include <swift/net.h>

#include <netdb.h>

#include <sys/socket.h>
#include <sys/types.h>

#include "posix/posix_addr.h"
#include "posix/posix_ipv4_addr.h"
#include "posix/posix_ipv6_addr.h"

bool sys_init_addr(swift_addr_t *addr, unsigned int addr_type)
{
	switch (addr_type) {
		case SWIFT_ADDR_IPV4: return sys_init_ipv4_addr(addr);
		case SWIFT_ADDR_IPV6: return sys_init_ipv6_addr(addr);
		default: break;
	}

	return false;
}

bool sys_init_addr_from(swift_addr_t *addr, const void *data)
{
	const struct sockaddr *sockaddr = data;

	switch (sockaddr->sa_family) {
		case AF_INET: return sys_init_ipv4_addr_from(addr, data);
		case AF_INET6: return sys_init_ipv6_addr_from(addr, data);
		default: break;
	}

	return false;
}

void sys_term_addr(swift_addr_t *addr)
{
	sys_addr_t *sys_addr = addr->data;

	sys_addr->ops->term_addr(addr);
}

struct sockaddr *sys_addr_sockaddr(swift_addr_t *addr)
{
	sys_addr_t *sys_addr = addr->data;

	return sys_addr->ops->addr_sockaddr(addr);
}

size_t sys_addr_sockaddr_size(swift_addr_t *addr)
{
	sys_addr_t *sys_addr = addr->data;

	return sys_addr->ops->addr_sockaddr_size(addr);
}

bool sys_resolve_addr(swift_array_t *addrs, unsigned int addr_type,
	const char *name, uint16_t port, unsigned int flags)
{
	struct addrinfo hints, *sys_addr, *sys_addrs;
	swift_addr_t *addr;
	swift_str_t *sys_port;
	int ret;

	memset(&hints, 0, sizeof(struct addrinfo));

	switch (addr_type) {
		case SWIFT_ADDR_ANY: {
			hints.ai_family = AF_UNSPEC;
		} break;

		case SWIFT_ADDR_IPV4: {
			hints.ai_family = AF_INET;
		} break;

		case SWIFT_ADDR_IPV6: {
			hints.ai_family = AF_INET6;
		} break;

		default: {
			return false;
		} break;
	}

	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	hints.ai_flags = 0;

	if (!name)
		hints.ai_flags |= AI_PASSIVE;

	sys_port = swift_new_strf("%" PRIu16, port);

	if (!sys_port)
		return false;

	ret = getaddrinfo(name, sys_port->data, &hints, &sys_addrs);
	swift_del_str(sys_port);

	if (ret != 0)
		return false;

	for (sys_addr = sys_addrs; sys_addr; sys_addr = sys_addr->ai_next) {
		addr = swift_new_addr_from(sys_addr->ai_addr);

		if (!addr)
			continue;

		swift_array_push(addrs, &addr);
	}

	return true;
}

swift_str_t *sys_resolve_name(swift_addr_t *addr, unsigned int flags)
{
	swift_str_t *name;
	int sys_flags = 0;

	name = swift_new_str_size(NI_MAXHOST);

	if (!name)
		return NULL;

	sys_flags |= (flags & SWIFT_NUMERIC_ADDR) ? NI_NUMERICHOST : 0;

	if (getnameinfo(sys_addr_sockaddr(addr), sys_addr_sockaddr_size(addr),
		name->data, name->size, NULL, 0, sys_flags) != 0)
		goto err_del_name;

	name->count = strlen(name->data);

	return name;

err_del_name:
	swift_del_str(name);
	return NULL;
}
