#include <swift/data.h>
#include <swift/net.h>

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <netinet/in.h>

#include <sys/socket.h>
#include <sys/types.h>

#include "posix_addr.h"
#include "posix_sock.h"

static bool sys_post_init_sock(swift_sock_t *sock)
{
	sys_sock_t *sys_sock = sock->data;
	socklen_t len = sizeof(int);
	int domain, proto;
	int value = 1;

	setsockopt(sys_sock->handle, SOL_SOCKET, SO_REUSEADDR, &value, len);

	return true;
}

bool sys_init_sock(swift_sock_t *sock, unsigned int addr_type,
	unsigned int proto_type)
{
	sys_sock_t *sys_sock;
	int domain, type, proto;

	switch (addr_type) {
		case SWIFT_ADDR_IPV4: {
			domain = AF_INET;
		} break;

		case SWIFT_ADDR_IPV6: {
			domain = AF_INET6;
		} break;

		default: {
			return false;
		} break;
	}

	switch (proto_type) {
		case SWIFT_PROTO_TCP: {
			type = SOCK_STREAM;
			proto = IPPROTO_TCP;
		} break;

		case SWIFT_PROTO_UDP: {
			type = SOCK_DGRAM;
			proto = IPPROTO_UDP;
		} break;

		default: {
			return false;
		} break;
	}

	sys_sock = malloc(sizeof(*sys_sock));

	if (!sys_sock)
		return false;

	sys_sock->handle = socket(domain, type, proto);

	if (sys_sock->handle < 0)
		goto err_free_sock;

	if (addr_type == SWIFT_ADDR_IPV6) {
		int value = 1;

		setsockopt(sys_sock->handle, IPPROTO_IPV6, IPV6_V6ONLY, &value,
			sizeof(value));
	}

	sock->data = sys_sock;

	if (!sys_post_init_sock(sock))
		goto err_close_sock;

	return true;

err_close_sock:
	close(sys_sock->handle);
err_free_sock:
	SWIFT_FREE(sys_sock);
	return false;
}

bool sys_init_sock_from(swift_sock_t *sock, const void *data)
{
	sys_sock_t *sys_sock;

	sys_sock = malloc(sizeof(*sys_sock));

	if (!sys_sock)
		return false;

	sys_sock->handle = *((int *)data);

	if (sys_sock->handle < 0)
		goto err_free_sock;

	sock->data = sys_sock;

	if (!sys_post_init_sock(sock))
		goto err_free_sock;

	return true;

err_free_sock:
	SWIFT_FREE(sys_sock);
	return false;
}

void sys_term_sock(swift_sock_t *sock)
{
	sys_sock_t *sys_sock = sock->data;

	if (!(sock->flags & SWIFT_FOREIGN_SOCK))
		close(sys_sock->handle);

	SWIFT_FREE(sys_sock);
}

size_t sys_sock_hash(const void *data)
{
	const swift_sock_t *sock = *(const swift_sock_t **)data;
	const sys_sock_t *sys_sock = sock->data;

	return swift_fnv1a_hash64(&sys_sock->handle, sizeof(sys_sock->handle));
}

bool sys_sock_eq(const void *lhs_, const void *rhs_)
{
	const swift_sock_t *ldata = *(const swift_sock_t **)lhs_;
	const swift_sock_t *rdata = *(const swift_sock_t **)rhs_;
	const sys_sock_t *lhs = ldata->data;
	const sys_sock_t *rhs = rdata->data;

	return (lhs->handle == rhs->handle);
}

size_t sys_peer_sock_hash(const void *data)
{
	const swift_peer_t *peer = *(const swift_peer_t **)data;
	const swift_sock_t *sock = peer->sock;
	const sys_sock_t *sys_sock = sock->data;

	return swift_fnv1a_hash64(&sys_sock->handle, sizeof(sys_sock->handle));
}

bool sys_peer_sock_eq(const void *lhs_, const void *rhs_)
{
	const swift_peer_t *lhs_peer = *(const swift_peer_t **)lhs_;
	const swift_peer_t *rhs_peer = *(const swift_peer_t **)rhs_;
	const swift_sock_t *lhs = lhs_peer->sock;
	const swift_sock_t *rhs = rhs_peer->sock;
	const sys_sock_t *sys_lhs = lhs->data;
	const sys_sock_t *sys_rhs = rhs->data;

	return (sys_lhs->handle != sys_rhs->handle);
}

unsigned int sys_sock_get_flags(swift_sock_t *sock)
{
	sys_sock_t *sys_sock = sock->data;
	unsigned int flags = 0;
	int sys_flags;

	sys_flags = fcntl(sys_sock->handle, F_GETFL);

	if (sys_flags < 0)
		return 0;

	flags |= (sys_flags & O_NONBLOCK) ? 0 : SWIFT_SOCK_WAIT;

	return (!(fcntl(sys_sock->handle, F_GETFL) & O_NONBLOCK));
}

bool sys_sock_set_flags(swift_sock_t *sock, unsigned int flags)
{
	sys_sock_t *sys_sock = sock->data;
	int sys_flags;

	sys_flags = fcntl(sys_sock->handle, F_GETFL);

	if (sys_flags < 0)
		return false;

	if (flags & SWIFT_SOCK_WAIT)
		sys_flags &= ~O_NONBLOCK;
	else
		sys_flags |= O_NONBLOCK;

	return (fcntl(sys_sock->handle, F_SETFL, sys_flags) == 0);
}

bool sys_sock_local_addr(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;
	socklen_t size = sys_addr_sockaddr_size(addr);

	return (getsockname(sys_sock->handle, sys_addr_sockaddr(addr), &size) ==
		0);
}

bool sys_sock_remote_addr(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;
	socklen_t size = sys_addr_sockaddr_size(addr);

	return (getpeername(sys_sock->handle, sys_addr_sockaddr(addr), &size) ==
		0);
}

bool sys_sock_connect(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;
	int ret;

	ret = connect(sys_sock->handle, sys_addr_sockaddr(addr),
		sys_addr_sockaddr_size(addr));

	if (ret == 0)
		return true;

	if (errno == EINPROGRESS)
		swift_set_errno(SWIFT_EINPROGRESS);

	return false;
}

bool sys_sock_bind(swift_sock_t *sock, swift_addr_t *addr)
{
	sys_sock_t *sys_sock = sock->data;

	return (bind(sys_sock->handle, sys_addr_sockaddr(addr),
		sys_addr_sockaddr_size(addr)) == 0);
}

bool sys_sock_listen(swift_sock_t *sock, unsigned int back_log)
{
	sys_sock_t *sys_sock = sock->data;

	back_log = SWIFT_MIN(back_log, SOMAXCONN);

	return (listen(sys_sock->handle, back_log) == 0);
}

swift_sock_t *sys_sock_accept(swift_sock_t *sock)
{
	sys_sock_t *sys_sock = sock->data;
	int ret = accept(sys_sock->handle, NULL, 0);

	if (ret >= 0)
		return swift_new_sock_from(sock->addr_type, sock->proto_type,
			&ret);

	if (errno == EAGAIN || errno == EWOULDBLOCK)
		swift_set_errno(SWIFT_EAGAIN);

	return NULL;
}

size_t sys_sock_recv(swift_sock_t *sock, void *data, size_t size)
{
	sys_sock_t *sys_sock = sock->data;
	int ret = recv(sys_sock->handle, data, size, 0);

	if (ret > 0)
		return ret;

	if (ret == 0)
		swift_set_errno(SWIFT_EHUP);
	else if (errno == EAGAIN || errno == EWOULDBLOCK)
		swift_set_errno(SWIFT_EAGAIN);

	return 0;
}

size_t sys_sock_send(swift_sock_t *sock, const void *data, size_t size)
{
	sys_sock_t *sys_sock = sock->data;
	int ret = send(sys_sock->handle, data, size, 0);

	if (ret > 0)
		return ret;

	if (ret == 0)
		swift_set_errno(SWIFT_EHUP);
	else if (errno == EAGAIN || errno == EWOULDBLOCK)
		swift_set_errno(SWIFT_EAGAIN);

	return 0;
}
