#include <swift/net.h>

struct swift_ring_buffer_t *swift_create_ring_buffer(size_t size)
{
	swift_ring_buffer_t *buffer;

	buffer = calloc(1, sizeof(swift_ring_buffer_t));

	if (!buffer)
		return NULL;

	if (!size)
		return buffer;

	buffer->size = size;
	buffer->data = calloc(size, sizeof(unsigned char));

	if (!buffer->data)
		goto free_buffer;

	swift_init_ref(&buffer->ref);

	return buffer;

free_buffer:
	SWIFT_FREE(buffer);
	return NULL;
}

static void swift_do_destroy_ring_buffer(swift_ref_t *ref)
{
	swift_ring_buffer_t *buffer =
		SWIFT_CONTAINER_OF(ref, swift_ring_buffer_t, ref);

	if (buffer->data)
		SWIFT_FREE(buffer->data);

	SWIFT_FREE(buffer);
}

void swift_destroy_ring_buffer(swift_ring_buffer_t *buffer)
{
	if (!buffer)
		return;

	swift_release_ref(&buffer->ref, swift_do_destroy_ring_buffer);
}

swift_ring_buffer_t *swift_copy_ring_buffer(swift_ring_buffer_t *buffer)
{
	if (!buffer)
		return NULL;

	swift_acquire_ref(&buffer->ref);

	return buffer;
}

bool swift_resize_ring_buffer(swift_ring_buffer_t *buffer, size_t size)
{
	swift_ring_buffer_t *tmp_buffer;
	size_t count;

	if (!buffer)
		return false;

	if (size > 0 && size < 16)
		size = 16;

	if (buffer->size == size)
		return true;

	tmp_buffer = swift_create_ring_buffer(size);

	if (!tmp_buffer)
		return false;

	tmp_buffer->count = SWIFT_MIN(tmp_buffer->size, buffer->count);
	tmp_buffer->tail = tmp_buffer->count;

	count = SWIFT_MIN(tmp_buffer->count, buffer->size - buffer->head);
	memcpy(tmp_buffer->data, buffer->data + buffer->head, count);
	memcpy(tmp_buffer->data + count, buffer->data, SWIFT_MIN(
		tmp_buffer->count - count, buffer->count - count));

	swift_swap(&buffer->data, &tmp_buffer->data, sizeof(void *));

	swift_swap(&buffer->head, &tmp_buffer->head, sizeof(size_t));
	swift_swap(&buffer->tail, &tmp_buffer->tail, sizeof(size_t));
	swift_swap(&buffer->count, &tmp_buffer->count, sizeof(size_t));
	swift_swap(&buffer->size, &tmp_buffer->size, sizeof(size_t));

	swift_destroy_ring_buffer(tmp_buffer);

	return true;
}

bool swift_expand_ring_buffer(swift_ring_buffer_t *buffer, size_t count)
{
	if (!buffer)
		return false;

	if (count <= buffer->size)
		return true;

	return swift_resize_ring_buffer(buffer, swift_next_pow2_size(count));
}

size_t swift_read_ring_buffer(swift_ring_buffer_t *buffer, void *data,
	size_t size)
{
	size_t count;

	if (!buffer || !data || !size)
		return 0;

	size = SWIFT_MIN(size, buffer->count);
	count = SWIFT_MIN(size, buffer->size - buffer->head);
	memcpy(data, buffer->data + buffer->head, count);
	memcpy(data + count, buffer->data, size - count);

	buffer->count -= size;
	buffer->head = (buffer->head + size) % buffer->size;

	return size;
}

size_t swift_write_ring_buffer(swift_ring_buffer_t *buffer, const void *data,
	size_t size)
{
	size_t count;

	if (!buffer || !data || !size)
		return 0;

	size = SWIFT_MIN(size, buffer->size - buffer->count);
	count = SWIFT_MIN(size, buffer->size - buffer->tail);
	memcpy(buffer->data + buffer->tail, data, count);
	memcpy(buffer->data, data + count, size - count);

	buffer->count += size;
	buffer->tail = (buffer->tail + size) % buffer->size;

	return size;
}
