#include <swift/net.h>

#include "monitor.h"
#include "sock.h"

static void swift_do_term_sock(swift_ref_t *ref)
{
	swift_sock_t *sock = SWIFT_CONTAINER_OF(ref, swift_sock_t,
		ref);

	if (sock->monitor)
		sys_deregister_sock(sock->monitor, sock);

	sys_term_sock(sock);
}

static void swift_do_del_sock(swift_ref_t *ref)
{
	swift_sock_t *sock = SWIFT_CONTAINER_OF(ref, swift_sock_t,
		ref);

	if (sock->monitor)
		sys_deregister_sock(sock->monitor, sock);

	sys_term_sock(sock);
	SWIFT_FREE(sock);
	sock = NULL;
}

size_t swift_sock_hash(const void *data)
{
	return sys_sock_hash(data);
}

bool swift_sock_eq(const void *lhs, const void *rhs)
{
	return sys_sock_eq(lhs, rhs);
}

size_t swift_peer_sock_hash(const void *data)
{
	return sys_peer_sock_hash(data);
}

bool swift_peer_sock_eq(const void *lhs, const void *rhs)
{
	return sys_peer_sock_eq(lhs, rhs);
}

bool swift_init_sock(swift_sock_t *sock, unsigned int addr_type,
	unsigned int proto_type)
{
	if (!sock)
		return false;

	if (!sys_init_sock(sock, addr_type, proto_type))
		return false;

	swift_init_ref(&sock->ref);
	sock->addr_type = addr_type;
	sock->proto_type = proto_type;

	return true;
}

bool swift_init_sock_from(swift_sock_t *sock, unsigned int addr_type,
	unsigned int proto_type, const void *data)
{
	if (!sock)
		return false;

	if (!sys_init_sock_from(sock, data))
		return false;

	swift_init_ref(&sock->ref);
	sock->flags = SWIFT_FOREIGN_SOCK;
	sock->addr_type = addr_type;
	sock->proto_type = proto_type;

	return true;
}

void swift_term_sock(swift_sock_t *sock)
{
	if (!sock)
		return;

	swift_release_ref(&sock->ref, swift_do_term_sock);
}

swift_sock_t *swift_new_sock(unsigned int addr_type,
	unsigned int proto_type)
{
	swift_sock_t *sock;

	sock = calloc(1, sizeof(*sock));

	if (!sock)
		return NULL;

	if (!swift_init_sock(sock, addr_type, proto_type))
		goto err_free_sock;

	return sock;

err_free_sock:
	SWIFT_FREE(sock);
	return NULL;
}

swift_sock_t *swift_new_sock_from(unsigned int addr_type,
	unsigned int proto_type, const void *data)
{
	swift_sock_t *sock;

	sock = calloc(1, sizeof(*sock));

	if (!sock)
		return NULL;

	if (!swift_init_sock_from(sock, addr_type, proto_type, data))
		goto err_free_sock;

	return sock;

err_free_sock:
	SWIFT_FREE(sock);
	return NULL;
}

void swift_del_sock(swift_sock_t *sock)
{
	if (!sock)
		return;

	swift_release_ref(&sock->ref, swift_do_del_sock);
}

swift_sock_t *swift_ref_sock(swift_sock_t *sock)
{
	if (!sock)
		return NULL;

	swift_acquire_ref(&sock->ref);

	return sock;
}

unsigned int swift_sock_get_flags(swift_sock_t *sock)
{
	if (!sock)
		return 0;

	return sys_sock_get_flags(sock);
}

bool swift_sock_set_flags(swift_sock_t *sock, unsigned int flags)
{
	if (!sock)
		return false;

	return sys_sock_set_flags(sock, flags);
}

swift_addr_t *swift_sock_local_addr(swift_sock_t *sock)
{
	swift_addr_t *addr;

	if (!sock)
		return NULL;

	addr = swift_new_addr(sock->addr_type);

	if (!addr)
		return NULL;

	if (!sys_sock_local_addr(sock, addr))
		goto err_del_addr;

	return addr;

err_del_addr:
	swift_del_addr(addr);
	return NULL;
}

swift_addr_t *swift_sock_remote_addr(swift_sock_t *sock)
{
	swift_addr_t *addr;

	if (!sock)
		return NULL;

	addr = swift_new_addr(sock->addr_type);

	if (!addr)
		return NULL;

	if (!sys_sock_remote_addr(sock, addr))
		goto err_del_addr;

	return addr;

err_del_addr:
	swift_del_addr(addr);
	return NULL;
}

bool swift_sock_connect(swift_sock_t *sock, swift_addr_t *addr)
{
	if (!sock || !addr)
		return false;

	return sys_sock_connect(sock, addr);
}

bool swift_queue_connect(swift_sock_t *sock, swift_addr_t *addr,
	swift_on_connect_fn_t on_connect, void *user_data)
{
	if (!sock || !sock->monitor || !addr || !on_connect)
		return false;

	return sys_queue_connect(sock, addr, on_connect, user_data);
}

bool swift_sock_bind(swift_sock_t *sock, swift_addr_t *addr)
{
	if (!sock || !addr)
		return false;

	return sys_sock_bind(sock, addr);
}

bool swift_sock_bind_name(swift_sock_t *sock, const char *name,
	uint16_t port)
{
	swift_array_t *addrs;
	swift_addr_t *addr;
	bool ret;

	if (!sock)
		return false;

	addrs = swift_resolve_addr(sock->addr_type, name, port, 0);

	if (!addrs || !addrs->count)
		return false;

	addr = *(swift_addr_t **)addrs->data;
	ret = swift_sock_bind(sock, addr);
	swift_del_addrs(addrs);

	return ret;
}

bool swift_sock_listen(swift_sock_t *sock, unsigned int back_log)
{
	if (!sock)
		return false;

	return sys_sock_listen(sock, back_log);
}

swift_sock_t *swift_sock_accept(swift_sock_t *sock)
{
	swift_sock_t *peer;

	if (!sock)
		return NULL;

	peer = sys_sock_accept(sock);

	if (!peer)
		return NULL;

	peer->flags &= ~SWIFT_FOREIGN_SOCK;

	return peer;
}

bool swift_queue_accept(swift_sock_t *sock,
	swift_on_accept_fn_t on_accept, void *user_data)
{
	if (!sock || !sock->monitor || !on_accept)
		return false;

	return sys_queue_accept(sock, on_accept, user_data);
}

size_t swift_sock_recv(swift_sock_t *sock, void *data, size_t size)
{
	if (!sock || !data || !size)
		return 0;

	return sys_sock_recv(sock, data, size);
}

bool swift_queue_recv(swift_sock_t *sock, void *data, size_t size,
	swift_on_recv_fn_t on_recv, void *user_data)
{
	if (!sock || !sock->monitor || !data || !size || !on_recv)
		return false;

	return sys_queue_recv(sock, data, size, on_recv, user_data);
}

size_t swift_sock_send(swift_sock_t *sock, const void *data, size_t size)
{
	if (!sock || !data || !size)
		return 0;

	return sys_sock_send(sock, data, size);
}

bool swift_queue_send(swift_sock_t *sock, const void *data, size_t size,
	swift_on_send_fn_t on_send, void *user_data)
{
	if (!sock || !sock->monitor || !data || !size || !on_send)
		return false;

	return sys_queue_send(sock, data, size, on_send, user_data);
}
