#include <swift/atomic.h>
#include <swift/string.h>
#include <swift/unicode.h>

static swift_ref_t ref = SWIFT_REF(0);

static void swift_do_term_strings(swift_ref_t *ref)
{
	swift_term_unicode();
}

bool swift_init_strings(void)
{
	if (swift_acquire_ref_unless_zero(&ref))
		return true;

	if (!swift_init_unicode())
		return false;

	swift_acquire_ref(&ref);

	return true;
}

void swift_term_strings(void)
{
	swift_release_ref(&ref, swift_do_term_strings);
}

swift_str_t *swift_new_str_size(size_t size)
{
	swift_str_t *str;

	str = calloc(1, sizeof(swift_str_t));

	str->size = size;

	if (!str->size)
		return str;

	str->data = calloc(str->size, sizeof(char));

	if (!str->data)
		goto err_free_str;

	return str;

err_free_str:
	SWIFT_FREE(str);
	return NULL;
}

swift_str_t *swift_new_str(void)
{
	return swift_new_str_size(0);
}

swift_str_t *swift_new_cstr(const char *cstr)
{
	swift_str_t *str;

	str = swift_new_str();

	if (!str)
		return NULL;

	if (!swift_set_cstr(str, cstr))
		goto err_del_str;

	return str;

err_del_str:
	swift_del_str(str);
	return NULL;
}

swift_str_t *swift_new_substr(size_t pos, size_t count, swift_str_t *src)
{
	swift_str_t *str;

	str = swift_new_str();

	if (!str)
		return NULL;

	if (!swift_set_substr(str, pos, count, src))
		goto err_del_str;

	return str;

err_del_str:
	swift_del_str(str);
	return NULL;
}

swift_str_t *swift_new_vstrf(const char *fmt, va_list args)
{
	swift_str_t *str;

	str = swift_new_str();

	if (!str)
		return NULL;

	if (!swift_set_vstrf(str, fmt, args))
		goto err_del_str;

	return str;

err_del_str:
	swift_del_str(str);
	return NULL;
}

swift_str_t *swift_new_strf(const char *fmt, ...)
{
	va_list args;
	swift_str_t *str;

	va_start(args, fmt);
	str = swift_new_vstrf(fmt, args);
	va_end(args);

	return str;
}

void swift_del_str(swift_str_t *str)
{
	if (!str)
		return;

	if (str->data)
		SWIFT_FREE(str->data);

	SWIFT_FREE(str);
	str = NULL;
}

bool swift_resize_str(swift_str_t *str, size_t size)
{
	swift_str_t *tmp_str;

	if (!str)
		return false;

	if (size > 0 && size < 16)
		size = 16;

	if (str->size == size)
		return true;

	tmp_str = swift_new_str_size(size);

	if (!tmp_str)
		return false;

	if (str->data)
		tmp_str->count = swift_conv_str(
			"utf-8", tmp_str->data, tmp_str->size,
			"utf-8", str->data, str->count + 1);

	if (tmp_str->count)
		--tmp_str->count;

	swift_swap(&str->data, &tmp_str->data, sizeof(void *));
	swift_swap(&str->count, &tmp_str->count, sizeof(size_t));
	swift_swap(&str->size, &tmp_str->size, sizeof(size_t));

	swift_del_str(tmp_str);

	return true;
}

bool swift_expand_str(swift_str_t *str, size_t count)
{
	if (!str)
		return false;

	if (count <= str->size)
		return true;

	return swift_resize_str(str, swift_next_pow2_size(count));
}

bool swift_set_str(swift_str_t *str, swift_str_t *src)
{
	if (!str)
		return false;

	swift_expand_str(str, src->count + 1);
	memcpy(str->data, src->data, src->count);
	str->count = src->count;
	swift_enc_utf8(str->data + str->count, 0);

	return true;
}

bool swift_set_substr(swift_str_t *str, size_t pos, size_t count,
	swift_str_t *src)
{
	if (!str)
		return false;

	pos = SWIFT_MIN(pos, src->count);
	count = SWIFT_MIN(count, src->count - pos);

	swift_expand_str(str, count + 1);
	memcpy(str->data, src->data + pos, count);
	str->count = count;
	swift_enc_utf8(str->data + str->count, 0);

	return true;
}

bool swift_set_cstr(swift_str_t *str, const char *data)
{
	size_t count;

	if (!str)
		return false;

	count = data ? strlen(data) : 0;
	swift_expand_str(str, count + 1);
	count = (data) ? (swift_conv_str(
		"utf-8", str->data, str->size,
		"utf-8", data, count + 1) - 1) : 0;

	str->count = count;
	swift_enc_utf8(str->data + str->count, 0);

	return true;
}

bool swift_set_from(swift_str_t *str, const char *src_fmt, const void *src,
	size_t src_size)
{
	size_t size;

	size = swift_conv_str_size("utf-8", src_fmt, src, src_size);

	if (!size)
		return swift_set_cstr(str, NULL);

	swift_expand_str(str, size);

	if (!swift_conv_str("utf-8", str->data, str->size, src_fmt, src,
		src_size))
		return swift_set_cstr(str, NULL);

	str->count = str->size - 1;

	return true;
}

bool swift_set_vstrf(swift_str_t *str, const char *fmt, va_list args)
{
	size_t count;

	if (!str)
		return false;

	count = vsnprintf(NULL, 0, fmt, args);
	swift_expand_str(str, count + 1);
	str->count = vsnprintf(str->data, count + 1, fmt, args);
	swift_enc_utf8(str->data + str->count, 0);

	return true;
}

bool swift_set_strf(swift_str_t *str, const char *fmt, ...)
{
	va_list args;
	bool status;

	if (!str)
		return false;

	va_start(args, fmt);
	status = swift_set_vstrf(str, fmt, args);
	va_end(args);

	return status;
}

bool swift_clear_str(swift_str_t *str)
{
	return swift_remove_str(str, 0, SIZE_MAX);
}

bool swift_append_uchar(swift_str_t *str, swift_uchar_t code)
{
	size_t count;

	if (!str)
		return false;

	count = swift_enc_utf8(NULL, code);
	swift_expand_str(str, str->count + count + 1);
	str->count += swift_enc_utf8(str->data + str->count, code);
	swift_enc_utf8(str->data + str->count, 0);

	return true;
}

bool swift_append_str(swift_str_t *str, swift_str_t *src)
{
	if (!str)
		return false;

	swift_expand_str(str, str->count + src->count + 1);
	memcpy(str->data + str->count, src->data, src->count);
	str->count += src->count;
	swift_enc_utf8(str->data + str->count, 0);

	return true;
}

bool swift_append_cstr(swift_str_t *str, const char *data)
{
	size_t count;

	if (!str || !data)
		return false;

	count = strlen(data);
	swift_expand_str(str, str->count + count + 1);
	count = swift_conv_str(
		"utf-8", str->data + str->count, str->size,
		"utf-8", data, count + 1);

	if (count)
		--count;

	str->count += count;

	return true;
}

size_t swift_find_str(swift_str_t *str, size_t pos,
	swift_str_t *key)
{
	if (!key)
		return SIZE_MAX;

	return swift_find_cstr(str, pos, key->data);
}

size_t swift_find_cstr(swift_str_t *str, size_t pos, const char *key)
{
	char *ret;

	if (!str || !key || pos >= str->count)
		return SIZE_MAX;

	ret = strstr(str->data + pos, key);

	if (!ret)
		return SIZE_MAX;

	return (size_t)(ret - str->data);
}

size_t swift_rfind_str(swift_str_t *str, size_t count,
	swift_str_t *key)
{
	size_t pos, i;

	if (!str || !key || !key->count)
		return SIZE_MAX;

	count = SWIFT_MIN(count, str->count);

	if (key->count > count)
		return SIZE_MAX;

	count -= key->count;

	for (i = 0; i <= count; ++i) {
		pos = count - i;

		if (strncmp(str->data + pos, key->data, key->count) == 0)
			return pos;
	}

	return SIZE_MAX;
}

size_t swift_rfind_cstr(swift_str_t *str, size_t count, const char *key)
{
	size_t pos, key_count, i;

	if (!str || !key)
		return SIZE_MAX;

	count = SWIFT_MIN(count, str->count);
	key_count = strlen(key);

	if (key_count > count)
		return SIZE_MAX;

	count -= key_count;

	for (i = 0; i <= count; ++i) {
		pos = count - i;

		if (strncmp(str->data + pos, key, key_count) == 0)
			return pos;
	}

	return SIZE_MAX;
}

bool swift_remove_str(swift_str_t *str, size_t pos, size_t count)
{
	if (!str)
		return false;

	pos = SWIFT_MIN(pos, str->count);
	count = SWIFT_MIN(count, str->count - pos);

	memcpy(str->data + pos, str->data + pos + count,
		str->count - pos - count);
	str->count -= count;
	swift_enc_utf8(str->data + str->count, 0);

	return true;
}

swift_array_t *swift_split_str(swift_str_t *str, swift_str_t *sep)
{
	if (!str || !sep || !sep->count)
		return NULL;

	return swift_split_cstr(str, sep->data);
}

swift_array_t *swift_split_cstr(swift_str_t *str, const char *sep)
{
	swift_array_t *items;
	swift_str_t *item;
	size_t pos = 0, prev_pos = 0, sep_count;

	if (!str || !sep)
		return NULL;

	sep_count = strlen(sep);
	items = swift_new_array(NULL, sizeof(swift_str_t *));

	if (!items)
		return NULL;

	do {
		pos = swift_find_cstr(str, pos, sep);
		item = swift_new_substr(prev_pos, pos - prev_pos, str);

		if (!item)
			goto err_del_items;

		if (!swift_array_push(items, &item)) {
			swift_del_str(item);
			goto err_del_items;
		}

		pos += SWIFT_MIN(SIZE_MAX - pos, sep_count);
		prev_pos = pos;
	} while (prev_pos != SIZE_MAX);

	return items;

err_del_items:
	swift_del_str_array(items);
	return NULL;
}

void swift_del_str_array(swift_array_t *str_array)
{
	swift_array_iter_t iter;
	swift_str_t *str;

	if (!str_array)
		return;

	for (iter = swift_array_first(str_array); iter.data != NULL;
		swift_array_next(&iter)) {
		str = *(swift_str_t **)iter.data;
		swift_del_str(str);
	}

	swift_del_array(str_array);
}

swift_uchar_t swift_str_pos_get(swift_str_t *str, size_t pos)
{
	swift_unicode_ret_t ret;
	unsigned char *src;
	size_t i = pos;

	src = (unsigned char *)str->data;

	if (i >= str->count)
		i = str->count - 5;

	src += pos;

	while(i < str->count) {
		ret = swift_dec_utf8(src, 4);
		src += ret.size;
		i += ret.size;

		if (ret.success)
			return ret.code;
	}

	return 0;
}

swift_str_iter_t swift_str_pos_next(swift_str_t *str, size_t pos)
{
	swift_unicode_ret_t ret;
	unsigned char *src = (unsigned char *)str->data;
	size_t i = pos, count;

	if (!str || !str->data || str->count == 0)
		return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);

	count = pos >= str->count ? str->count - 5 : pos;
	src += count;

	for (i = count; i < str->count; i += ret.size) {
		ret = swift_dec_utf8(src, 4);
		src += ret.size;

		if (ret.success)
			return SWIFT_STR_ITER(str, ret.code, i, ret.size);
	}

	return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);
}

swift_str_iter_t swift_str_pos_prev(swift_str_t *str, size_t pos)
{
	swift_unicode_ret_t ret, ret2;
	unsigned char *src = (unsigned char *)str->data;
	size_t i = 5, count = pos - 5;

	if (!str || !str->data || str->count == 0)
		return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);

	src += count;

	for (i = 5; i > 0; i -= ret.size) {
		ret = swift_dec_utf8(src, 4);
		src += ret.size;

		if (ret.success)
			ret2 = ret;
	}

	if (!ret2.success)
		return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);

	return SWIFT_STR_ITER(str, ret2.code, count - ret2.size, ret2.size);
}

swift_str_iter_t swift_str_first(swift_str_t *str)
{
	swift_unicode_ret_t ret;
	unsigned char *src = (unsigned char *)str->data;

	if (!str || !src || str->count == 0)
		return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);

	ret = swift_dec_utf8(src, 4);

	if (ret.success)
		return SWIFT_STR_ITER(str, ret.code, 0, ret.size);

	return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);
}

swift_str_iter_t swift_str_last(swift_str_t *str)
{
	swift_unicode_ret_t ret, ret2;
	unsigned char *src = (unsigned char *)str->data;
	size_t i = 5, count = str->count - 5;

	if (!str || !str->data || str->count == 0)
		return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);

	src += count;

	for (i = 5; i > 0; i -= ret.size) {
		ret = swift_dec_utf8(src, 4);
		src += ret.size;

		if (ret.success)
			ret2 = ret;
	}

	if (!ret2.success)
		return SWIFT_STR_ITER(NULL, SWIFT_UNICODE_UNKNOWN, 0, 0);

	return SWIFT_STR_ITER(str, ret2.code, str->count - ret2.size, ret2.size);
}

bool swift_str_next(swift_str_iter_t *iter)
{
	swift_unicode_ret_t ret;
	unsigned char *src = (unsigned char *)iter->str->data;

	if (!iter || !iter->str || !src || iter->str->count == 0 ||
		iter->pos + iter->size >= iter->str->count)
		return false;

	src += iter->pos + iter->size;

	ret = swift_dec_utf8(src, 4);

	if (!ret.success)
		return false;

	iter->pos += iter->size;
	iter->size = ret.size;
	iter->code = ret.code;

	return true;
}

bool swift_str_prev(swift_str_iter_t *iter)
{
	swift_unicode_ret_t ret, ret2;
	unsigned char *src = (unsigned char *)iter->str->data;
	size_t i = 5, count;

	if (!iter || !iter->str || !src|| iter->str->count == 0|| iter->pos <= 0)
		return false;

	count = (iter->pos - 5 > 0) ? (iter->pos - 5) : 0;
	src += count;

	for (i = 5; i > 0; i -= ret.size) {
		ret = swift_dec_utf8(src, 4);
		src += ret.size;

		if (ret.success)
			ret2 = ret;
	}

	if (!ret2.success)
		return false;

	iter->code = ret2.code;
	iter->pos -= ret2.size;
	iter->size = ret2.size;

	return true;
}

bool is_digit(size_t c)
{
	return (c >= '0' && c <= '9');
}

//TODO: Setup so it can get multiple numbers from a string if seprated etc.
int64_t swift_str_to_int(swift_str_t *str)
{
	bool isminus = false;
	swift_str_iter_t iter = swift_str_first(str);
	int64_t ret = 0;

	if (!iter.str)
		return 0;

	if (iter.code == '-') {
		isminus = true;
		if (!swift_str_next(&iter))
			return 0;
	}

	do {
		if (!is_digit(iter.code))
			return 0;

		ret = (ret * 10) + (iter.code - '0');
	} while (swift_str_next(&iter));

	if (isminus)
		ret = -ret;

	return ret;
}

//fix to make sure 1 string is not longer than the other.
bool swift_str_comp(swift_str_t *str1, swift_str_t *str2)
{
	swift_str_iter_t lhs = swift_str_first(str1);
	swift_str_iter_t rhs = swift_str_first(str2);

	if (!lhs.str || !rhs.str)
		return false;

	do {
		if (lhs.code != rhs.code)
			return false;

	} while (swift_str_next(&lhs) && swift_str_next(&rhs));

	return true;
}
