#include <swift/atomic.h>
#include <swift/thread.h>
#include <swift/time.h>

#include "generic/generic_cond.h"

bool native_init_cond(swift_cond_t *cond)
{
	native_cond_t *cond_data;

	if (!cond)
		return false;

	cond_data = malloc(sizeof(native_cond_t));

	if (!cond_data)
		return false;

	cond_data->lock = swift_new_mutex(SWIFT_MUTEX_RECURSIVE);

	if (!cond_data->lock)
		goto err_free_cond_data;

	cond_data->wait = swift_new_sem(0);

	if (!cond_data->wait)
		goto err_del_lock;

	cond_data->ready = swift_new_sem(0);

	if (!cond_data->ready)
		goto err_del_wait;

	cond->data = cond_data;

	return true;

err_del_wait:
	swift_del_sem(cond_data->wait);
err_del_lock:
	swift_del_mutex(cond_data->lock);
err_free_cond_data:
	SWIFT_FREE(cond_data);
	return false;
}

void native_term_cond(swift_cond_t *cond)
{
	native_cond_t *cond_data;

	if (!cond)
		return;

	cond_data = cond->data;
	swift_del_sem(cond_data->ready);
	swift_del_sem(cond_data->wait);
	swift_del_mutex(cond_data->lock);
	SWIFT_FREE(cond_data);
}

bool native_cond_wait(swift_cond_t *cond, swift_mutex_t *mutex)
{
	native_cond_t *cond_data;

	if (!cond)
		return false;

	cond_data = cond->data;

	swift_mutex_lock(cond_data->lock);
	++cond_data->wait_count;
	swift_mutex_unlock(cond_data->lock);

	swift_mutex_unlock(mutex);

	if (!swift_sem_wait(cond_data->wait))
		goto err_lock_mutex;

	swift_mutex_lock(cond_data->lock);

	if (cond_data->signal_count) {
		swift_sem_signal(cond_data->ready);
		--cond_data->signal_count;
	}

	--cond_data->wait_count;

	swift_mutex_unlock(cond_data->lock);
	swift_mutex_lock(mutex);

	return true;

err_lock_mutex:
	swift_mutex_lock(mutex);
	swift_mutex_lock(cond_data->lock);
	--cond_data->wait_count;
	swift_mutex_unlock(cond_data->lock);

	return false;
}

bool native_cond_timed_wait(swift_cond_t *cond, swift_mutex_t *mutex,
	swift_time_t time)
{
	native_cond_t *cond_data;

	if (!cond)
		return false;

	cond_data = cond->data;

	swift_mutex_lock(cond_data->lock);
	++cond_data->wait_count;
	swift_mutex_unlock(cond_data->lock);

	swift_mutex_unlock(mutex);

	if (!swift_sem_timed_wait(cond_data->wait, time))
		goto err_lock_mutex;

	swift_mutex_lock(cond_data->lock);

	if (cond_data->signal_count) {
		swift_sem_signal(cond_data->ready);
		--cond_data->signal_count;
	}

	--cond_data->wait_count;

	swift_mutex_unlock(cond_data->lock);
	swift_mutex_lock(mutex);

	return true;

err_lock_mutex:
	swift_mutex_lock(mutex);
	swift_mutex_lock(cond_data->lock);
	--cond_data->wait_count;
	swift_mutex_unlock(cond_data->lock);

	return false;
}

bool native_cond_signal(swift_cond_t *cond)
{
	native_cond_t *cond_data;

	if (!cond)
		return false;

	cond_data = cond->data;

	swift_mutex_lock(cond_data->lock);

	if (cond_data->wait_count <= cond_data->signal_count) {
		swift_mutex_unlock(cond_data->lock);

		return true;
	}

	++cond_data->signal_count;

	swift_sem_signal(cond_data->wait);
	swift_mutex_unlock(cond_data->lock);
	swift_sem_wait(cond_data->ready);

	return true;
}

bool native_cond_broadcast(swift_cond_t *cond)
{
	native_cond_t *cond_data;
	unsigned long signal, signal_count;

	if (!cond)
		return false;

	cond_data = cond->data;

	swift_mutex_lock(cond_data->lock);

	if (cond_data->wait_count <= cond_data->signal_count) {
		swift_mutex_unlock(cond_data->lock);

		return true;
	}

	signal_count = cond_data->wait_count - cond_data->signal_count;
	cond_data->signal_count += signal_count;

	for (signal = 0; signal < signal_count; ++signal)
		swift_sem_signal(cond_data->wait);

	swift_mutex_unlock(cond_data->lock);

	for (signal = 0; signal < signal_count; ++signal)
		swift_sem_wait(cond_data->ready);

	return true;
}
