#include <swift/atomic.h>
#include <swift/data.h>
#include <swift/thread.h>
#include <swift/time.h>

#include "mutex.h"

static void swift_do_del_mutex(swift_ref_t *ref)
{
	swift_mutex_t *mutex;

	if (!ref)
		return;

	mutex = SWIFT_CONTAINER_OF(ref, swift_mutex_t, ref);

	native_term_mutex(mutex);

	if (mutex->on_heap)
		SWIFT_FREE(mutex);
}

bool swift_init_mutex(swift_mutex_t *mutex, unsigned int flags)
{
	if (!mutex)
		return false;

	mutex->count = 0;
	mutex->owner_id = 0;
	mutex->flags = flags;

	if (!native_init_mutex(mutex))
		return false;

	swift_init_ref(&mutex->ref);

	return true;
}

void swift_term_mutex(swift_mutex_t *mutex)
{
	if (!mutex)
		return;

	swift_release_ref(&mutex->ref, swift_do_del_mutex);
}

swift_mutex_t *swift_new_mutex(unsigned int flags)
{
	swift_mutex_t *mutex;

	mutex = calloc(1, sizeof(swift_mutex_t));

	if (!mutex)
		return NULL;

	mutex->on_heap = true;

	if (!swift_init_mutex(mutex, flags))
		goto err_free_mutex;

	return mutex;

err_free_mutex:
	SWIFT_FREE(mutex);
	return NULL;
}

void swift_del_mutex(swift_mutex_t *mutex)
{
	if (!mutex)
		return;

	swift_release_ref(&mutex->ref, swift_do_del_mutex);
}

swift_mutex_t *swift_copy_mutex(swift_mutex_t *mutex)
{
	if (!mutex)
		return NULL;

	swift_acquire_ref(&mutex->ref);

	return mutex;
}

bool swift_mutex_lock(swift_mutex_t *mutex)
{
	swift_thread_id_t current_id = swift_get_current_thread_id();

	if (!mutex)
		return false;

	if (mutex->owner_id == current_id) {
		if (!(mutex->flags & SWIFT_MUTEX_RECURSIVE))
			return false;

		++mutex->count;

		return true;
	}

	if (!native_mutex_lock(mutex))
		return false;

	mutex->owner_id = current_id;
	mutex->count = 0;

	return true;
}

bool swift_mutex_try_lock(swift_mutex_t *mutex)
{
	swift_thread_id_t current_id = swift_get_current_thread_id();

	if (!mutex)
		return false;

	if (mutex->owner_id == current_id) {
		if (!(mutex->flags & SWIFT_MUTEX_RECURSIVE))
			return false;

		++mutex->count;

		return true;
	}

	if (!native_mutex_try_lock(mutex))
		return false;

	mutex->owner_id = current_id;
	mutex->count = 0;

	return true;
}

bool swift_mutex_timed_lock(swift_mutex_t *mutex, swift_time_t time)
{
	swift_thread_id_t current_id = swift_get_current_thread_id();

	if (!mutex)
		return false;

	if (mutex->owner_id == current_id) {
		if (!(mutex->flags & SWIFT_MUTEX_RECURSIVE))
			return false;

		++mutex->count;

		return true;
	}

	if (!native_mutex_timed_lock(mutex, time))
		return false;

	mutex->owner_id = current_id;
	mutex->count = 0;

	return true;
}

bool swift_mutex_unlock(swift_mutex_t *mutex)
{
	swift_thread_id_t current_id = swift_get_current_thread_id();

	if (!mutex || mutex->owner_id != current_id)
		return false;

	if ((mutex->flags & SWIFT_MUTEX_RECURSIVE) && mutex->count) {
		--mutex->count;

		return true;
	}

	mutex->owner_id = 0;

	return native_mutex_unlock(mutex);
}
