#include <swift/atomic.h>
#include <swift/thread.h>
#include <swift/time.h>

#include <errno.h>
#include <pthread.h>

#include "posix/posix_cond.h"
#include "posix/posix_mutex.h"

bool native_init_cond(swift_cond_t *cond)
{
	native_cond_t *cond_data;

	if (!cond)
		return false;

	cond_data = malloc(sizeof(native_cond_t));

	if (!cond_data)
		return false;

	if (pthread_cond_init(&cond_data->handle, NULL) < 0)
		goto err_free_cond_data;

	cond->data = cond_data;

	return true;

err_free_cond_data:
	SWIFT_FREE(cond_data);
	return false;
}

void native_term_cond(swift_cond_t *cond)
{
	native_cond_t *cond_data;

	if (!cond)
		return;

	cond_data = cond->data;
	pthread_cond_destroy(&cond_data->handle);
	SWIFT_FREE(cond_data);
}

bool native_cond_wait(swift_cond_t *cond, swift_mutex_t *mutex)
{
	native_cond_t *cond_data;
	native_mutex_t *mutex_data;
	int ret;

	if (!cond || !mutex)
		return false;

	cond_data = cond->data;
	mutex_data = mutex->data;

	do {
		ret = pthread_cond_wait(&cond_data->handle, &mutex_data->handle);
	} while (ret == EINTR);

	return (ret == 0);
}

bool native_cond_timed_wait(swift_cond_t *cond, swift_mutex_t *mutex,
	swift_time_t time_)
{
	struct timespec time;
	native_cond_t *cond_data;
	native_mutex_t *mutex_data;
	int ret;

	if (!cond || !mutex)
		return false;

	cond_data = cond->data;
	mutex_data = mutex->data;

	time_ = swift_add_time(swift_get_monotonic_time(), time_);
	time.tv_sec = time_.s;
	time.tv_nsec = time_.ns;

	do {
		ret = pthread_cond_timedwait(&cond_data->handle, &mutex_data->handle,
			&time);
	} while (ret == EINTR);

	return (ret == 0);
}

bool native_cond_signal(swift_cond_t *cond)
{
	native_cond_t *cond_data;

	if (!cond)
		return false;

	cond_data = cond->data;

	return (pthread_cond_signal(&cond_data->handle) == 0);
}

bool native_cond_broadcast(swift_cond_t *cond)
{
	native_cond_t *cond_data;

	if (!cond)
		return false;

	cond_data = cond->data;

	return (pthread_cond_broadcast(&cond_data->handle) == 0);
}
