#include <swift/atomic.h>
#include <swift/thread.h>
#include <swift/time.h>

#include <errno.h>
#include <pthread.h>

#include "posix/posix_mutex.h"

bool native_init_mutex(swift_mutex_t *mutex)
{
	pthread_mutexattr_t attr;
	native_mutex_t *mutex_data;
	int ret;

	if (!mutex)
		return false;

	mutex_data = malloc(sizeof(native_mutex_t));

	if (!mutex_data)
		return false;

	if (pthread_mutexattr_init(&attr) < 0)
		goto err_free_mutex_data;

	ret = pthread_mutex_init(&mutex_data->handle, &attr);
	pthread_mutexattr_destroy(&attr);

	if (ret < 0)
		goto err_free_mutex_data;

	mutex->data = mutex_data;

	return true;

err_free_mutex_data:
	SWIFT_FREE(mutex_data);
	return false;
}

void native_term_mutex(swift_mutex_t *mutex)
{
	native_mutex_t *mutex_data;

	if (!mutex)
		return;

	mutex_data = mutex->data;
	pthread_mutex_destroy(&mutex_data->handle);
	SWIFT_FREE(mutex_data);
}

bool native_mutex_lock(swift_mutex_t *mutex)
{
	native_mutex_t *mutex_data;
	int ret;

	if (!mutex)
		return false;

	mutex_data = mutex->data;

	do {
		ret = pthread_mutex_lock(&mutex_data->handle);
	} while (ret == EINTR);

	return (ret == 0);
}

bool native_mutex_try_lock(swift_mutex_t *mutex)
{
	native_mutex_t *mutex_data;
	int ret;

	if (!mutex)
		return false;

	mutex_data = mutex->data;

	do {
		ret = pthread_mutex_trylock(&mutex_data->handle);
	} while (ret == EINTR);

	return (ret == 0);
}

bool native_mutex_timed_lock(swift_mutex_t *mutex, swift_time_t time_)
{
	struct timespec time;
	native_mutex_t *mutex_data;
	int ret;

	if (!mutex)
		return false;

	mutex_data = mutex->data;

	time_ = swift_add_time(swift_get_monotonic_time(), time_);
	time.tv_sec = time_.s;
	time.tv_nsec = time_.ns;

	do {
		ret = pthread_mutex_timedlock(&mutex_data->handle, &time);
	} while (ret == EINTR);

	return (ret == 0);
}

bool native_mutex_unlock(swift_mutex_t *mutex)
{
	native_mutex_t *mutex_data;

	if (!mutex)
		return false;

	mutex_data = mutex->data;

	return (pthread_mutex_unlock(&mutex_data->handle) == 0);
}
