#include <swift/atomic.h>
#include <swift/thread.h>

#include <pthread.h>
#include <sched.h>
#include <signal.h>

#include "thread.h"

#include "posix/posix_thread.h"

const int signals[] = {
	SIGALRM, SIGCHLD, SIGHUP, SIGINT, SIGPIPE, SIGPROF, SIGQUIT, SIGTERM,
	SIGVTALRM, SIGWINCH,
};

static void *native_thread_run(void *user_data)
{
	swift_thread_run(user_data);

	return NULL;
}

bool native_init_threads(void)
{
	return true;
}

void native_term_threads(void)
{
}

swift_thread_id_t native_get_current_thread_id(void)
{
	return (swift_thread_id_t)pthread_self();
}

void native_yield_thread(void)
{
	sched_yield();
}

void native_exit_thread(int status)
{
	pthread_exit((void *)&status);
}

bool native_init_thread(swift_thread_t *thread, void *user_data,
	size_t stack_size, unsigned int flags)
{
	pthread_attr_t attr;
	native_thread_t *thread_data;

	if (!thread)
		return false;

	thread_data = malloc(sizeof(native_thread_t));

	if (!thread_data)
		return false;

	if (pthread_attr_init(&attr) < 0)
		goto err_free_thread_data;

	if (pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM) < 0)
		goto err_destroy_attr;

	if (stack_size > 0) {
		if (pthread_attr_setstacksize(&attr, stack_size) < 0)
			goto err_destroy_attr;
	}

	if (flags & SWIFT_THREAD_DETACHED) {
		if (pthread_attr_setdetachstate(&attr,
			PTHREAD_CREATE_DETACHED) < 0)
			goto err_destroy_attr;
	}

	if (pthread_create(&thread_data->handle, &attr, native_thread_run,
		user_data) < 0)
		goto err_destroy_attr;

	thread->data = thread_data;
	pthread_attr_destroy(&attr);

	return true;

err_destroy_attr:
	pthread_attr_destroy(&attr);
err_free_thread_data:
	SWIFT_FREE(thread_data);
	return false;
}

bool native_init_current_thread(swift_thread_t *thread)
{
	native_thread_t *thread_data;

	if (!thread)
		return false;

	thread_data = malloc(sizeof(native_thread_t));

	if (!thread_data)
		return false;

	thread_data->handle = pthread_self();

	return true;
}

void native_term_thread(swift_thread_t *thread)
{
	native_thread_t *thread_data;

	if (!thread)
		return;

	thread_data = thread->data;

	SWIFT_FREE(thread_data);
}

bool native_thread_setup(swift_thread_t *thread)
{
	sigset_t mask;
	size_t signal;
	int type;

	if (!thread)
		return false;

	sigemptyset(&mask);

	for (signal = 0; signal < SWIFT_ARRAY_SIZE(signals); ++signal)
		sigaddset(&mask, signals[signal]);

	pthread_sigmask(SIG_BLOCK, &mask, 0);

	return (pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &type) == 0);
}

bool native_thread_wait(swift_thread_t *thread)
{
	native_thread_t *thread_data;

	if (!thread)
		return false;

	thread_data = thread->data;

	return (pthread_join(thread_data->handle, 0) == 0);
}

bool native_thread_cancel(swift_thread_t *thread)
{
	native_thread_t *thread_data;

	if (!thread)
		return false;

	thread_data = thread->data;

	return (pthread_cancel(thread_data->handle) == 0);
}

bool native_thread_set_prio(swift_thread_t *thread, unsigned int prio)
{
	struct sched_param sched_param;
	native_thread_t *thread_data;
	int policy, min, max, range;

	if (!thread)
		return false;

	thread_data = thread->data;

	if (pthread_getschedparam(thread_data->handle, &policy, &sched_param) <
		0)
		return false;

	min = sched_get_priority_min(policy);
	max = sched_get_priority_max(policy);
	range = abs(max - min);

	sched_param.sched_priority = (100 * min + prio * range) / 100;

	return (pthread_setschedparam(thread_data->handle, policy,
		&sched_param) == 0);
}
