#include <swift/atomic.h>
#include <swift/data.h>
#include <swift/thread.h>
#include <swift/time.h>

#include "sem.h"

static void swift_do_del_sem(swift_ref_t *ref)
{
	swift_sem_t *sem;

	if (!ref)
		return;

	sem = SWIFT_CONTAINER_OF(ref, swift_sem_t, ref);

	native_term_sem(sem);

	if (sem->on_heap)
		SWIFT_FREE(sem);
}

bool swift_init_sem(swift_sem_t *sem, int count)
{
	if (!sem)
		return false;

	if (!native_init_sem(sem, count))
		return false;

	swift_init_ref(&sem->ref);
	swift_atomic_set(&sem->count, count);

	return true;
}

void swift_term_sem(swift_sem_t *sem)
{
	if (!sem)
		return;

	swift_release_ref(&sem->ref, swift_do_del_sem);
}

swift_sem_t *swift_new_sem(int count)
{
	swift_sem_t *sem;

	sem = calloc(1, sizeof(swift_sem_t));

	if (!sem)
		return NULL;

	sem->on_heap = true;

	if (!swift_init_sem(sem, count))
		goto err_free_sem;

	return sem;

err_free_sem:
	SWIFT_FREE(sem);
	return NULL;
}

void swift_del_sem(swift_sem_t *sem)
{
	if (!sem)
		return;

	swift_release_ref(&sem->ref, swift_do_del_sem);
}

swift_sem_t *swift_copy_sem(swift_sem_t *sem)
{
	if (!sem)
		return NULL;

	swift_acquire_ref(&sem->ref);

	return sem;
}

bool swift_sem_wait(swift_sem_t *sem)
{
	if (!sem)
		return false;

	swift_atomic_dec(&sem->count);

	if (!native_sem_wait(sem))
		goto err_inc_count;

	return true;

err_inc_count:
	swift_atomic_inc(&sem->count);
	return false;
}

bool swift_sem_try_wait(swift_sem_t *sem)
{
	if (!sem)
		return false;

	swift_atomic_dec(&sem->count);

	if (!native_sem_try_wait(sem))
		goto err_inc_count;

	return true;

err_inc_count:
	swift_atomic_inc(&sem->count);
	return false;
}

bool swift_sem_timed_wait(swift_sem_t *sem, swift_time_t time)
{
	if (!sem || time.s < 0)
		return false;

	swift_atomic_dec(&sem->count);

	if (!native_sem_timed_wait(sem, time))
		goto err_inc_count;

	return true;

err_inc_count:
	swift_atomic_inc(&sem->count);
	return false;
}

bool swift_sem_signal(swift_sem_t *sem)
{
	if (!sem)
		return false;

	swift_atomic_inc(&sem->count);

	if (!native_sem_signal(sem))
		goto err_dec_count;

	return true;

err_dec_count:
	swift_atomic_dec(&sem->count);
	return false;
}

bool swift_sem_get_count(swift_sem_t *sem, int *count)
{
	if (!sem || !count)
		return false;

	*count = swift_atomic_get(&sem->count);

	return true;
}
