#include <swift/atomic.h>
#include <swift/data.h>
#include <swift/thread.h>

#include "thread.h"

typedef struct swift_thread_info_t swift_thread_info_t;

struct swift_thread_info_t {
	swift_run_thread_fn_t run_thread;
	void *user_data;
	swift_thread_t *thread;
	swift_sem_t *ready;
};

static swift_thread_id_t main_thread_id;
static swift_htable_t *threads;
static swift_mutex_t *threads_lock;
static swift_ref_t ref = SWIFT_REF(0);

static size_t swift_hash_thread(const void *data)
{
	const swift_thread_t *thread = *(const swift_thread_t **)data;

	return swift_fnv1a_hash64(&thread->thread_id, sizeof(swift_thread_id_t));
}

static bool swift_eq_threads(const void *lhs_, const void *rhs_)
{
	const swift_thread_t *lhs = *(const swift_thread_t **)lhs_;
	const swift_thread_t *rhs = *(const swift_thread_t **)rhs_;

	return (lhs->thread_id == rhs->thread_id);
}

static bool swift_add_thread(swift_thread_t *thread)
{
	size_t index;

	swift_mutex_lock(threads_lock);
	index = swift_htable_insert(threads, &thread);
	swift_mutex_unlock(threads_lock);

	return (index != SIZE_MAX);
}

static void swift_do_del_thread(swift_ref_t *ref)
{
	swift_thread_t *thread;

	if (!ref)
		return;

	thread = SWIFT_CONTAINER_OF(ref, swift_thread_t, ref);
	swift_mutex_lock(thread->lock);
	swift_htable_remove(threads, &thread);

	if (!(thread->flags & (SWIFT_THREAD_EXITED | SWIFT_THREAD_FOREIGN)))
		native_thread_cancel(thread);

	native_term_thread(thread);

	if (thread->on_heap)
		SWIFT_FREE(thread);
}

static void swift_do_term_threads(swift_ref_t *ref)
{
	swift_htable_iter_t iter;
	swift_thread_t *thread;

	native_term_threads();

	for (iter = swift_htable_first(threads); iter.data;
		swift_htable_next(&iter)) {
		thread = *(swift_thread_t **)iter.data;

		swift_del_thread(thread);
	}

	swift_del_htable(threads);
	swift_del_mutex(threads_lock);
}

bool swift_init_threads(void)
{
	main_thread_id = swift_get_current_thread_id();

	if (swift_acquire_ref_unless_zero(&ref))
		return true;

	threads_lock = swift_new_mutex(SWIFT_MUTEX_RECURSIVE);

	if (!threads_lock)
		return false;

	threads = swift_new_htable(swift_hash_thread, swift_eq_threads,
		sizeof(swift_thread_t));

	if (!threads)
		goto err_del_threads_lock;

	if (!native_init_threads())
		goto err_del_threads;

	swift_acquire_ref(&ref);

	return true;

err_del_threads:
	swift_del_htable(threads);
err_del_threads_lock:
	swift_del_mutex(threads_lock);
	return false;
}

void swift_term_threads(void)
{
	swift_release_ref(&ref, swift_do_term_threads);
}

swift_thread_id_t swift_get_main_thread_id(void)
{
	return main_thread_id;
}

swift_thread_id_t swift_get_current_thread_id(void)
{
	return native_get_current_thread_id();
}

swift_thread_t *swift_get_thread(swift_thread_id_t thread_id)
{
	swift_thread_t key_ = {
		.thread_id = thread_id,
	};
	swift_thread_t *key = &key_;
	void *ret;

	swift_mutex_lock(threads_lock);
	ret = swift_htable_search(threads, &key);
	swift_mutex_unlock(threads_lock);

	return ret ? *(swift_thread_t **)ret : NULL;
}

swift_thread_t *swift_get_current_thread(void)
{
	swift_thread_t *thread;
	swift_thread_id_t thread_id = swift_get_current_thread_id();

	thread = swift_get_thread(thread_id);

	if (thread)
		return thread;

	thread = calloc(1, sizeof(swift_thread_t));

	if (!thread)
		return NULL;

	thread->on_heap = true;

	if (!native_init_current_thread(thread))
		goto err_free_thread;

	if (!swift_add_thread(thread))
		goto err_term_thread;

	return thread;

err_term_thread:
err_free_thread:
	SWIFT_FREE(thread);
	return NULL;
}

void swift_yield_thread(void)
{
	native_yield_thread();
}

void swift_exit_thread(int status)
{
	swift_thread_t *thread;

	thread = swift_get_current_thread();

	if (!thread)
		return;

	swift_mutex_lock(thread->lock);
	thread->flags |= SWIFT_THREAD_EXITED;
	thread->status = status;
	swift_mutex_unlock(thread->lock);

	native_exit_thread(status);
}

bool swift_init_thread(swift_thread_t *thread, swift_run_thread_fn_t run_thread,
	void *user_data, size_t stack_size, unsigned int flags)
{
	swift_thread_info_t *thread_info;
	swift_sem_t *ready;

	if (!thread)
		return false;

	swift_init_ref(&thread->ref);
	thread->flags = (thread->flags & SWIFT_THREAD_DETACHED) ?
		SWIFT_THREAD_DETACHED : SWIFT_THREAD_TO_BE_JOINED;
	thread->lock = swift_new_mutex(SWIFT_MUTEX_RECURSIVE);

	if (!thread->lock)
		goto err_del_ready;

	ready = swift_new_sem(0);

	if (!ready)
		return false;

	thread_info = calloc(1, sizeof(swift_thread_info_t));

	if (!thread_info)
		goto err_del_thread_lock;

	thread_info->run_thread = run_thread;
	thread_info->user_data = user_data;
	thread_info->thread = thread;
	thread_info->ready = swift_copy_sem(ready);

	swift_mutex_lock(thread->lock);

	if (!native_init_thread(thread, thread_info, stack_size, flags))
		return false;

	swift_mutex_unlock(thread->lock);
	swift_sem_wait(ready);
	swift_del_sem(ready);
	SWIFT_FREE(thread_info);

	return true;

err_del_thread_lock:
	swift_del_mutex(thread->lock);
err_del_ready:
	swift_del_sem(ready);
	return false;
}

void swift_term_thread(swift_thread_t *thread)
{
	if (!thread)
		return;

	swift_release_ref(&thread->ref, swift_do_del_thread);
}

swift_thread_t *swift_new_thread(swift_run_thread_fn_t run_thread,
	void *user_data, size_t stack_size, unsigned int flags)
{
	swift_thread_t *thread;

	thread = calloc(1, sizeof(swift_thread_t));

	if (!thread)
		return NULL;

	thread->on_heap = true;

	if (!swift_init_thread(thread, run_thread, user_data, stack_size,
		flags))
		goto err_free_thread;

	return thread;

err_free_thread:
	SWIFT_FREE(thread);
	return NULL;
}

void swift_del_thread(swift_thread_t *thread)
{
	if (!thread)
		return;

	swift_release_ref(&thread->ref, swift_do_del_thread);
}

void swift_thread_run(void *user_data)
{
	swift_thread_info_t *thread_info = user_data;
	swift_sem_t *ready;
	swift_run_thread_fn_t run_thread;
	int status;

	if (!thread_info)
		goto err_exit_thread;

	if (!native_thread_setup(thread_info->thread))
		goto err_exit_thread;

	thread_info->thread->thread_id = swift_get_current_thread_id();

	if (!swift_add_thread(thread_info->thread))
		goto err_exit_thread;

	ready = thread_info->ready;
	run_thread = thread_info->run_thread;
	user_data = thread_info->user_data;

	swift_sem_signal(ready);
	swift_del_sem(ready);

	status = run_thread(user_data);
	swift_exit_thread(status);

	return;

err_exit_thread:
	swift_sem_signal(thread_info->ready);
	swift_del_sem(thread_info->ready);
	swift_exit_thread(EXIT_FAILURE);
}

bool swift_thread_wait(swift_thread_t *thread, int *status)
{
	if (!thread)
		return false;

	swift_mutex_lock(thread->lock);

	if (thread->flags & SWIFT_THREAD_DETACHED)
		goto err_unlock_thread_lock;

	if (thread->flags & SWIFT_THREAD_TO_BE_JOINED) {
		swift_mutex_unlock(thread->lock);

		if (!native_thread_wait(thread))
			return false;

		swift_mutex_lock(thread->lock);

		thread->flags &= ~SWIFT_THREAD_TO_BE_JOINED;
	}

	if (status)
		*status = thread->status;

	return true;

err_unlock_thread_lock:
	swift_mutex_unlock(thread->lock);
	return false;
}

bool swift_thread_get_prio(swift_thread_t *thread, unsigned int *prio)
{
	if (!thread || !prio)
		return false;

	*prio = thread->prio;

	return true;
}

bool swift_thread_set_prio(swift_thread_t *thread, unsigned int prio)
{
	if (!thread)
		return false;

	prio = SWIFT_MIN(prio, 100);

	if (!native_thread_set_prio(thread, prio))
		return false;

	thread->prio = prio;

	return true;
}
