#include <swift/data.h>
#include <swift/thread.h>
#include <swift/time.h>

typedef struct swift_thread_pool_t swift_thread_pool_t;
typedef struct swift_job_t swift_job_t;

struct swift_thread_pool_t {
	swift_list_t jobs;
	swift_array_t *threads;
	swift_mutex_t *lock;
};

struct swift_job_t {
	swift_list_t list;
	swift_run_job_fn_t run_job;
	void *user_data;
};

static swift_thread_pool_t *thread_pool;

static int swift_thread_pool_run(void *user_data)
{
	swift_thread_pool_t *thread_pool = user_data;
	swift_job_t *job;
	swift_list_t *entry;

	if (!thread_pool)
		return EXIT_FAILURE;

	while (true) {
		swift_mutex_lock(thread_pool->lock);
		entry = swift_list_shift(&thread_pool->jobs);
		swift_mutex_unlock(thread_pool->lock);

		if (!entry) {
			swift_yield_thread();
			continue;
		}

		job = SWIFT_CONTAINER_OF(entry, swift_job_t, list);
		job->run_job(job->user_data);
		SWIFT_FREE(job);
	}

	return EXIT_SUCCESS;
}

bool swift_init_thread_pool(unsigned int count)
{
	swift_array_iter_t iter;
	swift_thread_t *thread;
	unsigned int i;

	thread_pool = calloc(1, sizeof(swift_thread_pool_t));

	if (!thread_pool)
		return false;

	swift_init_list(&thread_pool->jobs);
	thread_pool->lock = swift_new_mutex(SWIFT_MUTEX_RECURSIVE);

	if (!thread_pool->lock)
		goto err_free_thread_pool;

	thread_pool->threads = swift_new_array(NULL, sizeof(swift_thread_t *));

	if (!thread_pool->threads)
		goto err_del_lock;

	for (i = 0; i < count; ++i) {
		thread = swift_new_thread(swift_thread_pool_run, thread_pool,
			8192, 0);

		if (!thread)
			goto err_del_threads;

		swift_array_push(thread_pool->threads, &thread);
	}

	return true;

err_del_threads:
	for (iter = swift_array_first(thread_pool->threads); iter.data;
		swift_array_next(&iter)) {
		thread = *(swift_thread_t **)iter.data;
		swift_del_thread(thread);
	}

	swift_del_array(thread_pool->threads);
err_del_lock:
	swift_del_mutex(thread_pool->lock);
err_free_thread_pool:
	SWIFT_FREE(thread_pool);
	return false;
}

void swift_term_thread_pool(void)
{
	swift_list_iter_t job_iter;
	swift_array_iter_t iter;
	swift_job_t *job;
	swift_thread_t *thread;

	swift_mutex_lock(thread_pool->lock);

	for (job_iter = swift_list_first(&thread_pool->jobs); job_iter.node;
		swift_list_next(&job_iter)) {
		job = SWIFT_CONTAINER_OF(job_iter.node, swift_job_t, list);
		SWIFT_FREE(job);
	}

	for (iter = swift_array_first(thread_pool->threads); iter.data;
		swift_array_next(&iter)) {
		thread = *(swift_thread_t **)iter.data;
		swift_del_thread(thread);
	}

	swift_del_array(thread_pool->threads);
	swift_del_mutex(thread_pool->lock);
	SWIFT_FREE(thread_pool);
}

bool swift_thread_pool_wait(void)
{
	swift_array_iter_t iter;
	swift_thread_t *thread;

	for (iter = swift_array_first(thread_pool->threads); iter.data;
		swift_array_next(&iter)) {
		thread = *(swift_thread_t **)iter.data;
		swift_thread_wait(thread, NULL);
	}

	return true;
}

bool swift_schedule_job(swift_run_job_fn_t run_job, void *user_data)
{
	swift_job_t *job;
	bool ret;

	if (!run_job)
		return false;

	job = malloc(sizeof(swift_job_t));

	if (!job)
		return false;

	swift_init_list(&job->list);
	job->run_job = run_job;
	job->user_data = user_data;

	swift_mutex_lock(thread_pool->lock);
	ret = swift_list_push(&thread_pool->jobs, &job->list);
	swift_mutex_unlock(thread_pool->lock);

	if (!ret)
		goto err_free_job;

	return true;

err_free_job:
	SWIFT_FREE(job);
	return false;
}
