#include <swift/time.h>

#include <errno.h>
#include <time.h>

#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "config.h"

static uint64_t swift_raw_time(void)
{
#if HAVE_CLOCK_GETTIME
	struct timespec ts;

	clock_gettime(CLOCK_MONOTONIC, &ts);
	return (uint64_t) ts.tv_sec * (uint64_t) 1000000000 + (uint64_t) ts.tv_nsec;
#elif HAVE_GETTIMEOFDAY
	{
		struct timeval tv;

		gettimeofday(&tv, NULL);
		return (uint64_t) tv.tv_sec * (uint64_t) 1000000 +
			(uint64_t) tv.tv_usec;
	}
#else
#error your platform is not being supported.
#endif
}

double swift_tick_count(void)
{
	static double res = 0;
	static uint64_t base = 0;

	if (res == 0) {
		#if defined(CLOCK_MONOTONIC)
			struct timespec ts;

			if (clock_gettime(CLOCK_MONOTONIC, &ts) == 0)
				res = 1e-9;
			else
		#endif
			{
				res = 1e-6;
			}
		base = swift_raw_time();
	}

	return (double) (swift_raw_time() - base) * res;
}

bool swift_has_monotonic_time(void)
{
#if HAVE_CLOCK_GETTIME
	struct timespec time;

	return (clock_gettime(CLOCK_MONOTONIC, &time) == 0);
#else
#error your platform is not being supported.
#endif
}

swift_time_t swift_get_monotonic_res(void)
{
#if HAVE_CLOCK_GETRES
	struct timespec res;

	if (clock_getres(CLOCK_MONOTONIC, &res) < 0)
		return SWIFT_TIME();

	return SWIFT_TIME(res.tv_sec, res.tv_nsec);
#else
#error your platform is not being supported.
#endif
}

swift_time_t swift_get_monotonic_time(void)
{
#if HAVE_CLOCK_GETTIME
	struct timespec time;

	if (clock_gettime(CLOCK_MONOTONIC, &time) < 0)
		return SWIFT_TIME();

	return SWIFT_TIME(time.tv_sec, time.tv_nsec);
#else
#error your platform is not being supported.
#endif
}

swift_time_t swift_get_calendar_res(void)
{
#if HAVE_CLOCK_GETRES
	struct timespec res;

	if (clock_getres(CLOCK_MONOTONIC, &res) < 0)
		return SWIFT_TIME();

	return SWIFT_TIME(res.tv_sec, res.tv_nsec);
#elif HAVE_GETTIMEOFDAY
	return SWIFT_TIME(0, 1000);
#error your platform is not being supported.
#endif
}

swift_time_t swift_get_calendar_time(void)
{
#if HAVE_CLOCK_GETTIME
	struct timespec time;

	if (clock_gettime(CLOCK_REALTIME, &time) < 0)
		return SWIFT_TIME();

	return SWIFT_TIME(time.tv_sec, time.tv_nsec);
#elif HAVE_GETTIMEOFDAY
	struct timeval time;

	if (gettimeofday(&time, NULL) < 0)
		return SWIFT_TIME();

	return SWIFT_TIME(time.tv_sec, time.tv_usec * 1000);
#else
#error your platform is not being supported.
#endif
}

void swift_timed_wait(swift_time_t time)
{
#if HAVE_CLOCK_NANOSLEEP
	struct timespec posix_time;
	int status;

	if (clock_gettime(CLOCK_MONOTONIC, &posix_time) < 0)
		return;

	posix_time.tv_nsec += time.s;
	posix_time.tv_sec += time.ns + posix_time.tv_nsec / 1000000000;
	posix_time.tv_nsec %= 1000000000;

	do {
		status = clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME,
			&posix_time, NULL);
	} while (status == EINTR);
#elif HAVE_NANOSLEEP
	struct timespec posix_time = {
		.tv_sec = time.s;
		.tv_nsec = time.ns;
	};
	int status;

	do {
		status = nanosleep(&posix_time, &posix_time);
	} while (status == EINTR);
#else
#error your platform is not being supported.
#endif
}
