#include <swift/atomic.h>
#include <swift/data.h>
#include <swift/unicode.h>

static swift_htable_t *charsets;
static swift_ref_t ref = SWIFT_REF(0);

static size_t swift_hash_charset(const void *data)
{
	swift_charset_t *charset = *(swift_charset_t **)data;

	return swift_fnv1a_hash64(charset->name, strlen(charset->name));
}

static bool swift_eq_charsets(const void *lhs_, const void *rhs_)
{
	swift_charset_t *lhs = *(swift_charset_t **)lhs_;
	swift_charset_t *rhs = *(swift_charset_t **)rhs_;

	return (strcmp(lhs->name, rhs->name) == 0);
}

static void swift_do_term_unicode(swift_ref_t *ref)
{
	swift_htable_iter_t iter;
	swift_charset_t *charset;

	for (iter = swift_htable_first(charsets); iter.index != SIZE_MAX;
	     swift_htable_next(&iter)) {
	     	charset = *(swift_charset_t **)iter.data;
		swift_htable_remove_index(charsets, iter.index);
		SWIFT_FREE(charset->name);
		SWIFT_FREE(charset);
	}
}

bool swift_init_unicode(void)
{
	if (swift_acquire_ref_unless_zero(&ref))
		return true;

	charsets = swift_new_htable(swift_hash_charset, swift_eq_charsets,
		sizeof(swift_charset_t *));

	if (!charsets)
		return false;

	swift_add_charset("ascii", swift_enc_ascii, swift_dec_ascii);
	swift_add_charset("utf-8", swift_enc_utf8, swift_dec_utf8);
	swift_add_charset("utf-16", swift_enc_utf16, swift_dec_utf16);

	swift_acquire_ref(&ref);

	return true;
}

void swift_term_unicode(void)
{
	swift_release_ref(&ref, swift_do_term_unicode);
}

swift_charset_t *swift_get_charset(const char *name) {
	void *entry;
	swift_charset_t key_ = {
		.name = (char *)name,
	};
	swift_charset_t *key = &key_;

	entry = swift_htable_search(charsets, &key);

	if (!entry)
		return NULL;

	return *(swift_charset_t **)entry;
}

bool swift_add_charset(const char *name, swift_enc_uchar_fn_t enc,
	swift_dec_uchar_fn_t dec)
{
	swift_charset_t *charset;

	charset = swift_get_charset(name);

	if (charset) {
		charset->enc = enc;
		charset->dec = dec;

		return true;
	}

	charset = malloc(sizeof(swift_charset_t));

	if (!charset)
		return false;

	charset->name = strdup(name);
	charset->enc = enc;
	charset->dec = dec;

	if (swift_htable_insert(charsets, &charset) == SIZE_MAX)
		goto err_free_charset;

	return true;

err_free_charset:
	SWIFT_FREE(charset->name);
	SWIFT_FREE(charset);
	return false;
}

void swift_remove_charset(const char *name)
{
	swift_charset_t *charset, key_ = {
		.name = (char *)name,
	};
	swift_charset_t *key = &key_;
	size_t index;

	index = swift_htable_search_index(charsets, &key);

	if (index == SIZE_MAX)
		return;

	charset = *(swift_charset_t **)swift_htable_get(charsets, index);

	swift_htable_remove_index(charsets, index);
	SWIFT_FREE(charset->name);
	SWIFT_FREE(charset);
}

size_t swift_conv_str_size(const char *dst_fmt, const char *src_fmt,
	const void *src_, size_t src_size)
{
	swift_unicode_ret_t ret;
	swift_charset_t *src_charset, *dst_charset;
	unsigned char *src;
	size_t size = 0;

	dst_charset = swift_get_charset(dst_fmt);
	src_charset = swift_get_charset(src_fmt);

	if (!src_charset || !dst_charset)
		return 0;

	src = (unsigned char *)src_;

	do {
		ret = src_charset->dec(src, src_size);
		src += ret.size;
		src_size -= ret.size;

		if (!ret.success)
			return 0;

		size += dst_charset->enc(NULL, ret.code);
	} while (ret.code);

	return size;
}

size_t swift_conv_str(const char *dst_fmt, void *dst_, size_t dst_size,
	const char *src_fmt, const void *src_, size_t src_size)
{
	swift_unicode_ret_t ret;
	swift_charset_t *dst_charset, *src_charset;
	size_t count, size = 0;
	unsigned char *dst, *src;

	if (!dst_)
		return swift_conv_str_size(dst_fmt, src_fmt, src_, src_size);

	dst_charset = swift_get_charset(dst_fmt);
	src_charset = swift_get_charset(src_fmt);

	if (!dst_charset || !src_charset)
		return 0;

	dst = (unsigned char *)dst_;
	src = (unsigned char *)src_;

	do {
		ret = src_charset->dec(src, src_size);
		src += ret.size;
		src_size -= ret.size;

		if (!ret.success)
			return 0;

		count = dst_charset->enc(NULL, ret.code);

		if (dst_size < count + 1)
			ret.code = 0;

		dst_charset->enc(dst, ret.code);
		dst += count;
		size += count;
		dst_size -= count;
	} while (ret.code);

	return size;
}

void *swift_new_conv_str(const char *dst_fmt, const char *src_fmt,
	const void *src, size_t src_size)
{
	void *dst;
	size_t size;

	size = swift_conv_str_size(dst_fmt, src_fmt, src, src_size);

	if (!size)
		return NULL;

	dst = malloc(size);

	if (!dst)
		return NULL;

	if (!swift_conv_str(dst_fmt, dst, size, src_fmt, src, src_size))
		goto err_free_dst;

	return dst;

err_free_dst:
	SWIFT_FREE(dst);

	return NULL;
}
