#include <swift/window.h>
#include <swift/fs.h>
#include <swift/io.h>
#include <freetype2/ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H

typedef struct swift_fonts_ctx_t swift_fonts_ctx_t;

struct swift_fonts_ctx_t {
	swift_str_t *path;
	swift_list_t node;
};

static FT_Library library;
static swift_list_t *fonts;

static size_t swift_glyph_hash(const void *data)
{
	const swift_sym_t *symbol = *(const swift_sym_t **)data;

	return swift_fnv1a_hash64(&symbol->code, sizeof(symbol->code));
}

static bool swift_glyph_eq(const void *lhs_, const void *rhs_)
{
	const swift_sym_t *lhs = *(const swift_sym_t **)lhs_;
	const swift_sym_t *rhs = *(const swift_sym_t **)rhs_;

	return (lhs->code == rhs->code);
}

static void swift_tex_init(swift_font_t *font, swift_vec2f_t hw)
{
	font->hw = hw;
	font->texture.size = hw;

	glGenTextures(1, &font->texture.tex_id);
	glBindTexture(GL_TEXTURE_2D, font->texture.tex_id);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, font->hw.wh.w, font->hw.wh.h, 0,
		GL_RED, GL_UNSIGNED_BYTE, 0);

	glPixelStoref(GL_UNPACK_ALIGNMENT, (GLfloat)1);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	font->texture.name = NULL;
}

static void swift_clear_tex_area(swift_font_t *font, swift_sym_t *glyph,
	swift_vec2f_t xy)
{
	glBindTexture(GL_TEXTURE_2D, font->texture.tex_id);
	glTexSubImage2D(GL_TEXTURE_2D, 0, xy.xy.x, xy.xy.y, glyph->width,
		glyph->height, GL_RED, GL_UNSIGNED_BYTE, 0);

	swift_htable_remove(font->glyphs, glyph);
	SWIFT_FREE(glyph);
}

static swift_sym_t *swift_create_glyph(swift_font_t *font, swift_sym_t *symbol,
	FT_GlyphSlot g, const swift_uchar_t code)
{
	symbol->ax = g->advance.x >> 6;
	symbol->ay = g->advance.y >> 6;
	symbol->width = g->bitmap.width;
	symbol->height = g->bitmap.rows;
	symbol->left = g->bitmap_left;
	symbol->top = g->bitmap_top;
	symbol->x = font->xy.xy.x;
	symbol->y = font->xy.xy.y;
	symbol->mheight = g->face->size->metrics.height >> 6;
	symbol->code = code;
	font->xy.xy.x += symbol->width + 1;

	swift_htable_insert(font->glyphs, &symbol);

	return symbol;
}

static swift_sym_t *swift_load_glyph(swift_font_t *font,
	const swift_uchar_t code)
{
	FT_Face face = font->face;
	FT_GlyphSlot g = face->glyph;
	swift_sym_t *symbol = calloc(1, sizeof(swift_sym_t));

	if (!symbol)
		return NULL;

	FT_Set_Pixel_Sizes(face, 0, 32);

	if (FT_Load_Char(face, code, FT_LOAD_RENDER))
		return NULL;

	if (font->xy.xy.x + g->bitmap.width > font->hw.wh.w) {
		font->xy.xy.y += font->h + 1;
		font->xy.xy.x = font->h = 0;

		if (font->xy.xy.y > font->hw.wh.h)
			return NULL;
	}

	if (font->h < g->bitmap.rows)
		font->h = g->bitmap.rows;

	if (!g->bitmap.width && !g->bitmap.rows)
		return swift_create_glyph(font, symbol, g, code);

	swift_bind_texture(font->texture.tex_id);

	glTexSubImage2D(GL_TEXTURE_2D, 0, font->xy.xy.x, font->xy.xy.y,
		g->bitmap.width, g->bitmap.rows, GL_RED, GL_UNSIGNED_BYTE,
			g->bitmap.buffer);

	return swift_create_glyph(font, symbol, g, code);
}

static swift_sym_t *swift_try_load_glyph(swift_font_t *font,
	const swift_uchar_t code)
{
	size_t i = FT_Get_Char_Index(font->face, code);
	swift_fonts_ctx_t *font_ctx;

	if (i == 0) {
		swift_list_iter_t iter;

		for (iter = swift_list_first(fonts); iter.node;
			swift_list_next(&iter)) {
			FT_Face face_loader;
			font_ctx = SWIFT_CONTAINER_OF(iter.node, swift_fonts_ctx_t, node);

			FT_Done_Face(font->face);

			if (FT_New_Face(library, font_ctx->path->data, 0, &face_loader))
				return NULL;

			i = FT_Get_Char_Index(face_loader, code);

			font->face = face_loader;

			if (i > 0)
				break;
		}

		if (i == 0)
			return NULL;
	}

	return swift_load_glyph(font, code);
}

static void swift_get_font_paths(const char *path)
{
	swift_str_t *font_path = swift_new_str();
	swift_dir_t *dir;
	swift_fonts_ctx_t *font_ctx;
	const char *entry;

	if (!(dir = swift_open_dir(path)))
		return;

	while ((entry = swift_read_dir(dir))) {
		swift_set_strf(font_path, "%s%s", path, entry);

		if (!(swift_find_cstr(font_path, font_path->count - 5, ".ttf") ==
			SIZE_MAX)) {
			if (!(font_ctx = malloc(sizeof(swift_fonts_ctx_t))))
				return;

			font_ctx->path = swift_new_strf("%s", font_path->data);
			swift_init_list(&font_ctx->node);

			swift_list_push(fonts, &font_ctx->node);
		}
	}

	swift_del_str(font_path);
	swift_close_dir(dir);
}

swift_sym_t *swift_search_glyph(swift_font_t *font,
	const swift_uchar_t code)
{
	void *entry;
	swift_sym_t key_ = {.code = code,};
	swift_sym_t *key = &key_;

	entry = swift_htable_search(font->glyphs, &key);

	return (entry ? *(swift_sym_t **)entry : NULL);
}

swift_sym_t *swift_get_glyph(swift_font_t *font,
	const swift_uchar_t code)
{
	swift_sym_t *symbol = swift_search_glyph(font, code);

	if (!symbol) {
		if (!(symbol = swift_try_load_glyph(font, code)))
			symbol = swift_search_glyph(font, 0);
	}

	return symbol;
}

bool swift_init_fonts(const char *path)
{
	swift_str_t *_path;

	if (FT_Init_FreeType(&library)) {
		fprintf(stderr, "Could not initialize freetype 2 library\n");
		return false;
	}

	if (path == NULL)
		_path = swift_new_strf("./");
	else
		_path = swift_new_strf("%s", path);

	if (!swift_check_path(_path->data)) {
		swift_del_str(_path);
		return false;
	}

	fonts = swift_new_list();
	swift_get_font_paths(_path->data);
	swift_del_str(_path);

	return true;
}

bool swift_load_font(swift_font_t *font)
{
	swift_vec2f_t hw;
	size_t i = 32;
	GLint max;
	swift_list_iter_t iter = swift_list_first(fonts);
	swift_fonts_ctx_t *font_ctx;
	FT_Face face;

	if (!iter.node)
		return false;

	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max);

	if (max > 8192)
		max = 8192;

	hw = SWIFT_VEC2F(max, max);

	font_ctx = SWIFT_CONTAINER_OF(iter.node, swift_fonts_ctx_t, node);

	if (FT_New_Face(library, font_ctx->path->data, 0, &face))
		return false;

	swift_tex_init(font, hw);
	font->xy.xy.x = font->xy.xy.y = font->h = 0;
	font->glyphs = swift_new_htable(swift_glyph_hash, swift_glyph_eq,
		sizeof(swift_sym_t *));

	FT_Set_Pixel_Sizes(face, 0, 32);

	font->face = face;
	swift_load_glyph(font, 0); /*preload missing glyph image*/

	for(i = 32; i < 127; i++)
		swift_try_load_glyph(font, i);

	return true;
}

void swift_unload_font(swift_font_t *font)
{
	swift_htable_iter_t iter;
	swift_sym_t *symbol;

	if (glIsTexture(font->texture.tex_id))
		glDeleteTextures(1, &font->texture.tex_id);

	for (iter = swift_htable_first(font->glyphs); iter.data;
		swift_htable_next(&iter)) {
			if ((symbol = *(swift_sym_t **)iter.data))
				SWIFT_FREE(symbol);
	}

	swift_del_htable(font->glyphs);
	FT_Done_Face((FT_Face)font->face);
}

void swift_term_fonts(void)
{
	FT_Done_FreeType(library);
	swift_list_t *node = NULL;
	swift_fonts_ctx_t *font_ctx;

	while ((node = swift_list_pop(fonts))) {
		if ((font_ctx = SWIFT_CONTAINER_OF(node, swift_fonts_ctx_t, node))) {
			swift_del_str(font_ctx->path);
			SWIFT_FREE(font_ctx);
		}
	}

	swift_del_list(fonts);
}

swift_sym_t swift_sym_resize(swift_sym_t *sym, float size)
{
	swift_sym_t symbol;
	float resize = size / 32;

	symbol.ax = sym->ax * resize;
	symbol.ay = sym->ay * resize;
	symbol.width = sym->width * resize;
	symbol.height = sym->height * resize;

	if (symbol.width < 1)
		symbol.width = 3;

	if (symbol.height < 1)
		symbol.height = 1;

	symbol.left = sym->left * resize;
	symbol.top = sym->top * resize;
	symbol.x = sym->x;
	symbol.y = sym->y;
	symbol.mheight = sym->mheight;
	symbol.code = sym->code;

	return symbol;
}
