#include <swift/window.h>
#include <swift/io.h>
typedef struct swift_render_ctx_t swift_render_ctx_t;
typedef struct swift_vertex2_t swift_vertex2_t;

struct swift_render_ctx_t {
	GLuint vbo, vao;
};

struct swift_vertex2_t {
	swift_vec2f_t position, texcoord;
};

static swift_render_ctx_t *render_ctx;

static swift_vertex2_t swift_create_vertex(swift_vec2f_t position,
	swift_vec2f_t texcoord)
{
	swift_vertex2_t result;
	result.position = position;
	result.texcoord = texcoord;

	return result;
}

static void swift_init_vbo(void)
{
	swift_vertex2_t vertices[4];

	vertices[0] = swift_create_vertex(SWIFT_VEC2F(-1, -1),
		SWIFT_VEC2F(0, 0));
	vertices[1] = swift_create_vertex(SWIFT_VEC2F(-1, 1),
		SWIFT_VEC2F(0, 1));
	vertices[2] = swift_create_vertex(SWIFT_VEC2F(1, -1),
		SWIFT_VEC2F(1, 0));
	vertices[3] = swift_create_vertex(SWIFT_VEC2F(1, 1),
		SWIFT_VEC2F(1, 1));

/*bind the VAO*/
	glGenVertexArrays(1, &render_ctx->vao);
	glBindVertexArray(render_ctx->vao);

/*bind the VBO and set the vertices*/
	glGenBuffers(1, &render_ctx->vbo);
	glBindBuffer(GL_ARRAY_BUFFER, render_ctx->vbo);
	glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(swift_vertex2_t), vertices,
		GL_STATIC_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);

	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, sizeof(swift_vertex2_t), 0);
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(swift_vertex2_t),
		(GLvoid*)sizeof(swift_vec2f_t));

/*unbind then rebind the VAO*/
	glBindVertexArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(render_ctx->vao);
}

void swift_set_clip_box(swift_window_t *window, swift_vec4f_t box)
{
	window->sprite_box = box;
}

void swift_draw_state_default(void)
{
	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
	glDepthFunc(GL_LEQUAL);
}

void swift_set_draw_view(swift_window_t *window, swift_vec4f_t box)
{
	glViewport(box.xywh.x, box.xywh.y, box.xywh.w, box.xywh.h);
	window->proj_mat = swift_mat4f_ortho(box.xywh.x, box.xywh.w, box.xywh.y,
		box.xywh.h, -1, 1);
	swift_set_clip_box(window, window->screen_box);
}

void swift_clear_screen(swift_vec4f_t col)
{
	glClearColor(col.rgba.r, col.rgba.g, col.rgba.b, col.rgba.a);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

swift_sprite_t swift_init_sprite(const char *img_path)
{
	swift_sprite_t sprite = SWIFT_SPRITE(.pos = SWIFT_VEC3F(.xyz.z = 1),
		.col = SWIFT_VEC4F(1, 1, 1, 1));

	if (!(sprite.tex = swift_add_texture(img_path)) ||
		!(sprite.shader = swift_get_shader("swift_basic_shader"))) {
		eprintf("Failed to load texture");
		swift_remove_texture(sprite.tex);
	}

	return sprite;
}

swift_sprite_t swift_init_sprite_notex(swift_vec4f_t color)
{
	swift_sprite_t sprite = SWIFT_SPRITE(.pos = SWIFT_VEC3F(.xyz.z = 1),
		.col = color, .tex = NULL);

	if (!(sprite.shader = swift_get_shader("swift_notex_shader")))
		eprintf("Failed to load texture");

	return sprite;
}

swift_sprite_t *swift_new_sprite_notex(swift_vec4f_t color)
{
	swift_sprite_t *sprite = calloc(1, sizeof(swift_sprite_t));

	if (!sprite)
		return NULL;

	sprite->pos = SWIFT_VEC3F(.xyz.z = 1);
	sprite->col = color;
	sprite->tex = NULL;

	if (!(sprite->shader = swift_get_shader("swift_notex_shader"))) {
		eprintf("Failed to load texture");
		SWIFT_FREE(sprite);
		return NULL;
	}

	return sprite;
}

swift_sprite_t *swift_new_sprite(const char *img_path)
{
	swift_sprite_t *sprite = calloc(1, sizeof(swift_sprite_t));

	if (!sprite)
		return NULL;

	sprite->pos = SWIFT_VEC3F(.xyz.z = 1);
	sprite->col = SWIFT_VEC4F(1, 1, 1, 1);

	if (!(sprite->tex = swift_add_texture(img_path)) ||
		!(sprite->shader = swift_get_shader("swift_basic_shader"))) {
		eprintf("Failed to load texture");
		swift_remove_texture(sprite->tex);
		SWIFT_FREE(sprite);
		return NULL;
	}

	return sprite;
}

void swift_unload_sprite(swift_sprite_t *sprite)
{
	swift_remove_texture(sprite->tex);
}

void swift_destory_sprite(swift_sprite_t *sprite)
{
	if (sprite) {
		swift_unload_sprite(sprite);
		SWIFT_FREE(sprite);
	}
}

bool swift_init_rendering(void)
{
	if (!(render_ctx = malloc(sizeof(swift_render_ctx_t))))
		return false;

	swift_draw_state_default();
	swift_init_vbo();

	if (!swift_init_shader_sys()) {
		swift_term_render();
		return false;
	}

	return true;
}

void swift_term_render(void)
{
	if (glIsVertexArray(render_ctx->vao))
		glDeleteVertexArrays(1, &render_ctx->vao);

	if (glIsBuffer(render_ctx->vbo))
		glDeleteBuffers(1, &render_ctx->vbo);

	if (render_ctx)
		SWIFT_FREE(render_ctx);
}

void swift_bind_texture(size_t texture_id)
{
	static GLuint cached_tex_id = 0;

	if (cached_tex_id != texture_id) {
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, texture_id);
		cached_tex_id = texture_id;
	}
}

void swift_render(swift_window_t *window, swift_sprite_t *sprite)
{
	if (sprite->tex)
		swift_bind_texture(sprite->tex->tex_id);

	swift_update_shader(window, sprite);

	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}
