#include <swift/window.h>
#include <swift/maths.h>
#include <swift/string.h>
#include <swift/io.h>

static swift_htable_t *shaders;

static size_t swift_hash_shader(const void *data)
{
	swift_shader_ctx_t *shader = *(swift_shader_ctx_t **)data;

	return swift_fnv1a_hash64(shader->name, strlen(shader->name));
}

static bool swift_eq_shaders(const void *lhs_, const void *rhs_)
{
	swift_shader_ctx_t *lhs = *(swift_shader_ctx_t **)lhs_;
	swift_shader_ctx_t *rhs = *(swift_shader_ctx_t **)rhs_;

	return (strcmp(lhs->name, rhs->name) == 0);
}

static void swift_print_shader_info_log (unsigned int shader_index)
{
	int actual_length = 0;
	char log[2048];
	glGetShaderInfoLog(shader_index, 2048, &actual_length, log);
	fprintf(stderr, "shader info log for GL index %i:\n%s\n", shader_index,
		log);
}

static void swift_compile_shader(swift_shader_t *shader)
{
	GLint success = 0;
	GLchar error[1024] = { 0 };

	glLinkProgram(shader->program);

	glGetProgramiv(shader->program, GL_LINK_STATUS, &success);

	if (!success) {
		glGetProgramInfoLog(shader->program, sizeof(error), NULL, error);
		fprintf(stderr, "%s\n", error);
	}

	glValidateProgram(shader->program);
	glGetProgramiv(shader->program, GL_VALIDATE_STATUS, &success);

	if (!success) {
		glGetProgramInfoLog(shader->program, sizeof(error), NULL, error);
		fprintf(stderr, "%s\n", error);
	}
}

static void swift_create_shader(swift_shader_t *shader, const char *source,
	GLenum shader_type)
{
	GLint lengths = strlen(source);
	GLuint shader_file = glCreateShader(shader_type);
	int params = -1;

	glShaderSource(shader_file, 1, &source, &lengths);
	glCompileShader(shader_file);
	glGetShaderiv (shader_file, GL_COMPILE_STATUS, &params);

	if (GL_TRUE != params)
		swift_print_shader_info_log(shader_file);

	glAttachShader(shader->program, shader_file);
}

static void printfMatrix(swift_mat4f_t *matrix)
{
	fprintf(stderr, "___START___\n");
	fprintf(stderr,"{ %f, %f, %f, %f }\n", matrix->data[0], matrix->data[1],
		matrix->data[2], matrix->data[3]);
	fprintf(stderr,"{ %f, %f, %f, %f }\n", matrix->data[4], matrix->data[5],
		matrix->data[6], matrix->data[7]);
	fprintf(stderr,"{ %f, %f, %f, %f }\n", matrix->data[8], matrix->data[9],
		matrix->data[10], matrix->data[11]);
	fprintf(stderr,"{ %f, %f, %f, %f }\n", matrix->data[12], matrix->data[13],
		matrix->data[14], matrix->data[15]);
	fprintf(stderr, "___END___\n");
}

static swift_mat4f_t swift_get_mat4f_trans(swift_sprite_t *sprite,
	swift_window_t *window)
{
	float w = sprite->size.wh.w * 0.5, h = sprite->size.wh.h * 0.5;
	swift_mat4f_t matrix = swift_mat4f_translation(
		SWIFT_VEC3F(w + sprite->pos.xyz.x, h + sprite->pos.xyz.y,
			sprite->pos.xyz.z));
	matrix = swift_mat4f_rotate(matrix, sprite->rotation,
		SWIFT_VEC3F(0, 0, 1));
	matrix = swift_mat4f_scale(matrix, SWIFT_VEC3F(w, h, 1));

	return matrix;
}

static swift_vec4f_t swift_update_tex_pos(swift_sprite_t *sprite)
{
	swift_vec4f_t result;

	if (!sprite->tex)
		return SWIFT_VEC4F(0, 0, 1, 1);

	if (sprite->tex->size.wh.w == 0 || sprite->tex->size.wh.h == 0)
		return sprite->tex_pos;

	result.xywh.x = sprite->tex_pos.xywh.x / sprite->tex->size.wh.w;
	result.xywh.y = sprite->tex_pos.xywh.y /  sprite->tex->size.wh.h;
	result.xywh.w = (sprite->tex_pos.xywh.x + sprite->tex_pos.xywh.w) /
		sprite->tex->size.wh.w;
	result.xywh.h = (sprite->tex_pos.xywh.y + sprite->tex_pos.xywh.h) /
		sprite->tex->size.wh.h;

	return result;
}

static void swift_unload_shader(swift_shader_t *shader)
{
	if (!shader) return;

	if (glIsShader(shader->transform))
		glDeleteShader(shader->transform);

	if (glIsShader(shader->texcoords))
		glDeleteShader(shader->texcoords);

	if (glIsShader(shader->colour))
		glDeleteShader(shader->colour);

	if (glIsShader(shader->uniform_tex))
		glDeleteShader(shader->uniform_tex);

	if (glIsShader(shader->screensize))
		glDeleteShader(shader->screensize);

	if (glIsShader(shader->ambientlight))
		glDeleteShader(shader->ambientlight);

	if (shader->type == SWIFT_SHADER_POINT) {
		size_t i = 0;

		for (i = 0; i < SWIFT_MAX_POINT_LIGHTS; i++) {
			if (glIsShader(shader->lights_uni[i].pos))
				glDeleteShader(shader->lights_uni[i].pos);

			if (glIsShader(shader->lights_uni[i].colour))
				glDeleteShader(shader->lights_uni[i].colour);

			if (glIsShader(shader->lights_uni[i].range))
				glDeleteShader(shader->lights_uni[i].range);

			if (glIsShader(shader->lights_uni[i].intensity))
				glDeleteShader(shader->lights_uni[i].intensity);
		}

		if (shader->lights) SWIFT_FREE(shader->lights);
		if (shader->lights_uni) SWIFT_FREE(shader->lights_uni);
	}

	if (glIsProgram(shader->program))
		glDeleteProgram(shader->program);
}

static bool swift_init_light(swift_light_t *light)
{
		if (!light)
			return false;

	light->pos = SWIFT_VEC2F(0.0f);
	light->col = SWIFT_VEC4F(1.0f);
	light->range = 0.0f;
	light->intensity = 0.0f;

	return true;
}

static swift_shader_t *swift_set_active_shader(swift_shader_t *shader)
{
	static swift_shader_t *active_shader = NULL;

	if (shader && active_shader != shader) {
		active_shader = shader;
		glUseProgram(shader->program);
	}

	return active_shader;
}

char *swift_load_shader(const char *path)
{
	FILE *file = fopen(path, "r");
	size_t filesize = 0;
	char *shader_source = NULL;

	fseek(file, 0, SEEK_END);
	filesize = ftell(file);
	rewind(file);

	shader_source = calloc(1, sizeof(char) * (filesize + 1));
	fread(shader_source, sizeof(char), filesize, file);
	shader_source[filesize] = '\0';
	fclose(file);

	return shader_source;
}

bool swift_update_shader_name(swift_window_t *window, const char *name,
	swift_sprite_t *sprite)
{
	if (name) {
		swift_shader_ctx_t *shader_ctx = swift_get_shader_ctx(name);

		if (!shader_ctx)
			return false;

		sprite->shader = shader_ctx->shader;

		return swift_update_shader(window, sprite);
	}

	return false;
}

bool swift_update_shader(swift_window_t *window, swift_sprite_t *sprite)
{
	if (!window || !sprite)
		return false;
	else {
		swift_shader_t *shader = swift_set_active_shader(sprite->shader);
		swift_vec4f_t tex_pos = swift_update_tex_pos(sprite);
		swift_mat4f_t mat = swift_mat4f_dot(swift_get_mat4f_trans(sprite,
			window), window->proj_mat);

		glUniform4f(shader->screensize, window->sprite_box.xywh.x,
			window->sprite_box.xywh.y, window->sprite_box.xywh.w,
				window->sprite_box.xywh.h);
		glUniformMatrix4fv(shader->transform, 1, GL_FALSE, mat.data);
		glUniform4f(shader->colour, sprite->col.rgba.r, sprite->col.rgba.g,
			sprite->col.rgba.b, sprite->col.rgba.a);
		glUniform4f(shader->texcoords, tex_pos.xywh.x, tex_pos.xywh.y,
			tex_pos.xywh.w, tex_pos.xywh.h);

		if (shader->type == SWIFT_SHADER_AMBIENT)
			glUniform1f(shader->ambientlight, 1.0f);

		if (shader->type == SWIFT_SHADER_POINT) {
			size_t i = 0;

			for (i = 0; i < SWIFT_MAX_POINT_LIGHTS; i++) {
				if (shader->lights[i].intensity > 0) {
					glUniform2f(shader->lights_uni[i].pos,
						shader->lights[i].pos.xy.x, shader->lights[i].pos.xy.y);

					glUniform4f(shader->lights_uni[i].colour,
						shader->lights[i].col.rgba.r,
							shader->lights[i].col.rgba.g,
								shader->lights[i].col.rgba.b,
									shader->lights[i].col.rgba.a);

					glUniform1f(shader->lights_uni[i].range,
						shader->lights[i].range);

					glUniform1f(shader->lights_uni[i].intensity,
						shader->lights[i].intensity);
				}
			}
		}

		return true;
	}
}

bool swift_init_shader_sys(void)
{
	if (!(shaders = swift_new_htable(swift_hash_shader, swift_eq_shaders,
		sizeof(swift_shader_ctx_t *))))
		return false;

	swift_add_shader(bss, bfts, "swift_basic_shader", SWIFT_SHADER_BASIC);
	swift_add_shader(bss, bfs, "swift_notex_shader", SWIFT_SHADER_BASIC);
	swift_add_shader(bss, afs, "swift_ambient_shader", SWIFT_SHADER_AMBIENT);
	swift_add_shader(bss, lfs, "swift_point_shader", SWIFT_SHADER_POINT);
	swift_add_shader(bss, tfs, "swift_text_shader", SWIFT_SHADER_TEXT);

	swift_set_active_shader(swift_get_shader("swift_basic_shader"));

	return true;
}

bool swift_term_shader_sys(void)
{
	swift_htable_iter_t iter;
	swift_shader_ctx_t *shader_ctx;

	if (shaders) {
		for (iter = swift_htable_first(shaders); iter.index !=
			SIZE_MAX; swift_htable_next(&iter)) {
			shader_ctx = *(swift_shader_ctx_t **)iter.data;

			swift_htable_remove_index(shaders, iter.index);
			swift_unload_shader(shader_ctx->shader);
			SWIFT_FREE(shader_ctx->shader);
			SWIFT_FREE(shader_ctx->name);
			SWIFT_FREE(shader_ctx);
		}

		swift_del_htable(shaders);

		return true;
	}

	return false;
}

bool swift_add_shader(const char *base, const char *frag, const char *name,
	int type)
{
	swift_shader_t *shader = calloc(1, sizeof(swift_shader_t));
	swift_shader_ctx_t *shader_ctx;

	if (!shader || !name || !base || !frag || type < SWIFT_SHADER_BASIC ||
		type >= SWIFT_SHADER_END)
		return false;

	shader->type = type;
	shader->program = glCreateProgram();
	swift_create_shader(shader, base, GL_VERTEX_SHADER);
	swift_create_shader(shader, frag, GL_FRAGMENT_SHADER);
	swift_compile_shader(shader);

	glBindAttribLocation(shader->program, 0, "position");
	glBindAttribLocation(shader->program, 1, "texcoord");
	shader->transform = glGetUniformLocation(shader->program, "transform");
	shader->colour = glGetUniformLocation(shader->program, "colour");
	shader->texcoords = glGetUniformLocation(shader->program, "texcoords");
	shader->screensize = glGetUniformLocation(shader->program, "screensize");

	if (type == SWIFT_SHADER_AMBIENT)
		shader->ambientlight = glGetUniformLocation(shader->program,
			"ambientlight");

	if (type == SWIFT_SHADER_POINT) {
		size_t i = 0;

		shader->lights = calloc(SWIFT_MAX_POINT_LIGHTS, sizeof(swift_light_t));
		shader->lights_uni =
			calloc(SWIFT_MAX_POINT_LIGHTS, sizeof(swift_light_uniform_t));

		if (!shader->lights_uni)
			goto ERR_SHADER;

		for (i = 0; i < SWIFT_MAX_POINT_LIGHTS; i++) {
			swift_str_t *string;

			string = swift_new_strf("pointLights[%i].pos", &i);
			shader->lights_uni[i].pos =
				glGetUniformLocation(shader->program, string->data);

			swift_set_strf(string, "pointLights[%i].colour", &i);
			shader->lights_uni[i].colour =
				glGetUniformLocation(shader->program, string->data);

			swift_set_strf(string, "pointLights[%i].range", &i);
			shader->lights_uni[i].range =
				glGetUniformLocation(shader->program, string->data);

			swift_set_strf(string, "pointLights[%i].intensity", &i);
			shader->lights_uni[i].intensity =
				glGetUniformLocation(shader->program, string->data);

			swift_del_str(string);

			swift_init_light(&shader->lights[i]);
		}
	}

	if ((shader_ctx = swift_get_shader_ctx(name))) {
		swift_unload_shader(shader_ctx->shader);
		shader_ctx->shader = shader;

		return true;
	}

	if (!(shader_ctx = malloc(sizeof(swift_shader_ctx_t))))
		goto ERR_SHADER;

	shader_ctx->name = strdup(name);
	shader_ctx->shader = shader;

	if (swift_htable_insert(shaders, &shader_ctx) == SIZE_MAX)
		goto ERR_HTABLE;

	return true;

ERR_HTABLE:
	SWIFT_FREE(shader_ctx->name);
	SWIFT_FREE(shader_ctx);
ERR_SHADER:
	swift_unload_shader(shader);
	SWIFT_FREE(shader);

	return false;
}

swift_shader_ctx_t *swift_get_shader_ctx(const char *name)
{
	void *entry;
	swift_shader_ctx_t key_ = {.name = (char *)name,};
	swift_shader_ctx_t *key = &key_;

	if (!(entry = swift_htable_search(shaders, &key)))
		return NULL;

	return *(swift_shader_ctx_t **)entry;
}

swift_shader_t *swift_get_shader(const char *name)
{
	swift_shader_ctx_t *shader_ctx = swift_get_shader_ctx(name);

	if (!shader_ctx || !shader_ctx->shader)
		return NULL;

	return shader_ctx->shader;
}

bool swift_add_light(swift_shader_t *shader, swift_light_t *light)
{
	if (!shader || !light)
		return false;

	if (shader->lights) {
		size_t i = 0;

		for (i = 0; i < SWIFT_MAX_POINT_LIGHTS; i++) {
			if (shader->lights[i].intensity <= 0) {
				shader->lights[i] = *light;
				break;
			}
		}
	}

	return true;
}

bool swift_remove_shader(const char *name)
{
	swift_shader_ctx_t *shader_ctx = swift_get_shader_ctx(name);

	if (!shader_ctx)
		return false;

	swift_htable_remove(shaders, &shader_ctx);
	swift_unload_shader(shader_ctx->shader);
	SWIFT_FREE(shader_ctx->shader);
	SWIFT_FREE(shader_ctx->name);
	SWIFT_FREE(shader_ctx);

	return true;
}
