#include <swift/data.h>
#include <swift/window.h>
#include <swift/image.h>

static swift_htable_t *textures;

static size_t swift_hash_texture(const void *data)
{
	swift_tex_t *texture = *(swift_tex_t **)data;

	return swift_fnv1a_hash64(texture->name, strlen(texture->name));
}

static bool swift_eq_textures(const void *lhs_, const void *rhs_)
{
	swift_tex_t *lhs = *(swift_tex_t **)lhs_;
	swift_tex_t *rhs = *(swift_tex_t **)rhs_;

	return (strcmp(lhs->name, rhs->name) == 0);
}

swift_tex_t *swift_create_texture(const char *path)
{
	swift_image_t *img = NULL;
	swift_tex_t *tex = malloc(sizeof(swift_tex_t));

	if (!tex)
		return NULL;

	if (!(img = swift_try_read_image_from_file(path))) {
		SWIFT_FREE(tex);
		return NULL;
	}

	glGenTextures(1, &tex->tex_id);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, tex->tex_id);
	glPixelStoref(GL_UNPACK_ALIGNMENT, (GLfloat)1);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->width, img->height, 0, GL_RGBA,
		GL_UNSIGNED_BYTE, (void*) img->data);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	tex->name = strdup(path);
	tex->size.wh.w = img->width;
	tex->size.wh.h = img->height;
	tex->ref = 0;

	swift_del_image(img);

	return tex;
}

static void swift_unload_textures(void)
{
	swift_htable_iter_t iter;
	swift_tex_t *tex;

	for (iter = swift_htable_first(textures); iter.data;
		swift_htable_next(&iter)) {
		tex = *(swift_tex_t **)iter.data;

		if (tex->tex_id)
			glDeleteTextures(1, &tex->tex_id);

		swift_htable_remove_index(textures, iter.index);

		if (tex->name)
			SWIFT_FREE(tex->name);

		if (tex)
			SWIFT_FREE(tex);
	}
}

bool swift_init_textures(const char *path)
{
		if (!(textures = swift_new_htable(swift_hash_texture, swift_eq_textures,
			sizeof(swift_tex_t *))))
			return false;

		if (!swift_init_images(path)) {
			swift_del_htable(textures);
			return false;
		}

	return true;
}

void swift_term_textures(void)
{
	swift_unload_textures();
	swift_del_htable(textures);
	swift_term_images();
}

swift_tex_t *swift_get_texture(const char *name)
{
	void *entry;
	swift_tex_t key_ = {.name = (char *)name,};
	swift_tex_t *key = &key_;

	entry = swift_htable_search(textures, &key);

	return (entry ? *(swift_tex_t **)entry : NULL);
}

swift_tex_t *swift_add_texture(const char *path)
{
	swift_tex_t *tex = NULL;

	if (!path)
		return NULL;

	if (!(tex = swift_get_texture(path))) {
		if (!(tex = swift_create_texture(path)))
			return NULL;

		swift_htable_insert(textures, &tex);
	}

	tex->ref++;
	return tex;
}

void swift_remove_texture(swift_tex_t *tex)
{
	if (!tex)
		return;

	if ((tex->ref--) <= 0) {
		swift_htable_remove(textures, &tex);

		if (glIsTexture(tex->tex_id))
			glDeleteTextures(1, &tex->tex_id);

		SWIFT_FREE(tex->name);
		SWIFT_FREE(tex);
	}
}
