#include <swift/window/glad.h>
#include <swift/window.h>
#include <swift/atomic.h>
#include "callbacks.h"

static swift_ref_t ref = SWIFT_REF(0);

static void swift_do_term_windows(swift_ref_t *ref)
{
	swift_term_textures();
	swift_term_shader_sys();
	swift_term_render();
	glfwTerminate();
}

static void swift_window_error_callback(int error, const char* description)
{
    puts(description);
}

static bool swift_init_glfw(swift_window_t *window, swift_window_t *parent,
	GLFWmonitor *monitor, swift_vec2f_t hw, const char *title, bool decorated)
{
	GLFWwindow *parent_window = NULL;

	if(!window)
		return false;

	glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_ANY_PROFILE);
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 2);
	glfwSetErrorCallback(swift_window_error_callback);

	if (decorated)
		glfwWindowHint(GLFW_DECORATED, GL_FALSE);
	else
		glfwWindowHint(GLFW_DECORATED, GL_TRUE);

	if (parent) {
		parent_window = parent->screen;
		window->parent = parent;
	}
	else
		window->parent = NULL;

	if (!(window->screen = glfwCreateWindow(hw.wh.w, hw.wh.h, title, monitor,
		parent_window))) {
		fprintf(stderr, "GLFWWIN_ERROR: screen in swift_init_glfw.\n");
		return false;
	}

	swift_set_current_window(window);

	gladLoadGLLoader((GLADloadproc) glfwGetProcAddress);

	swift_handle_window_resize(window, hw.wh.w, hw.wh.h);

	if (!swift_create_callbacks(window)) {
		glfwDestroyWindow(window->screen);
		return false;
	}

	return true;
}

void swift_set_current_window(swift_window_t *window)
{
	static GLFWwindow *current_screen = NULL;

	if (window->screen != current_screen) {
		current_screen = window->screen;
		glfwMakeContextCurrent(current_screen);
	}
}

bool swift_init_windows(bool init_tex_cache, const char *path)
{
	if (swift_acquire_ref_unless_zero(&ref))
		return true;

	if (glfwInit() == GL_FALSE) {
		fprintf(stderr, "GLFW FAILED TO INIT");
		return false;
	}

	if (init_tex_cache) {
		if (!swift_init_textures(path)) {
			glfwTerminate();
			return false;
		}
	}

	swift_acquire_ref(&ref);

	return true;
}

void swift_term_windows(void)
{
	swift_release_ref(&ref, swift_do_term_windows);
}

swift_window_t *swift_create_window(swift_window_t *parent, swift_vec2f_t hw,
	swift_vec2f_t pos, const char *title, bool decorated, void *user_data)
{
	swift_window_t *window = malloc(sizeof(swift_window_t));

	if (!window)
		return NULL;

	swift_init_list(&window->children);

	if (!(swift_init_glfw(window, parent, NULL, hw, title, decorated))) {
		SWIFT_FREE(window);
		return NULL;
	}

	if (parent) {
		swift_init_list(&window->node);
		swift_list_insert_before(&parent->children, &window->node);
	}

	window->user_data = user_data;

	if (pos.xy.x < 0)
		pos.xy.x = 0;

	if (pos.xy.y < 0)
		pos.xy.y = 0;

	glfwSetWindowUserPointer(window->screen, window);
	glfwSetWindowPos(window->screen, pos.xy.x, pos.xy.y);
	glfwSetWindowSizeCallback(window->screen, &swift_callback_window_resize);
	glfwSetWindowRefreshCallback(window->screen, &swift_callback_window_refresh);
	glfwSetWindowFocusCallback(window->screen, &swift_callback_window_focus);
	glfwSetWindowIconifyCallback(window->screen, &swift_callback_iconify);
	glfwSetWindowPosCallback(window->screen, &swift_callback_window_pos);
	glfwSetFramebufferSizeCallback(window->screen,
		&swift_callback_buffer_resize);
	glfwSetCharModsCallback(window->screen, &swift_callback_mods);
	glfwSetWindowCloseCallback(window->screen, &swift_callback_window_close);
	glfwSetCharCallback(window->screen, &swift_callback_char);
	glfwSetKeyCallback(window->screen, &swift_callback_key);
	glfwSetCursorPosCallback(window->screen, &swift_callback_cursor_pos);
	glfwSetCursorEnterCallback(window->screen, &swift_callback_cursor_entered);
	glfwSetDropCallback(window->screen, &swift_callback_drops);
	glfwSetMouseButtonCallback(window->screen, &swift_callback_mouse_press);
	glfwSetScrollCallback(window->screen, &swift_callback_scroll);
	glfwSetInputMode(window->screen, GLFW_STICKY_MOUSE_BUTTONS, GL_TRUE);

	return window;
}

void swift_destroy_window(swift_window_t *window)
{
	swift_intern_cb_t *cbs;

	if(!swift_list_is_empty(&window->children)) {
		swift_list_iter_t iter;
		swift_window_t *child;

		for (iter = swift_list_last(&window->children); iter.node;
			swift_list_prev(&iter)) {
			if ((child = SWIFT_CONTAINER_OF(iter.node, swift_window_t, node))) {
				cbs = child->callbacks;

				if (cbs && cbs->win_close)
					cbs->win_close(child);

				swift_destroy_window(child);
			}
		}
	}

	if(window->parent) {
		swift_list_remove(&window->parent->children, &window->node);
		window->parent = NULL;
	}

	cbs = window->callbacks;

	if (cbs && cbs->win_close)
		cbs->win_close(window);

	swift_destroy_callbacks(window);
	window->user_data = NULL;
	glfwDestroyWindow(window->screen);
	SWIFT_FREE(window);
}

void swift_handle_window_resize(swift_window_t *window, int width, int height)
{
	if (!window || !window->screen)
		return;

	swift_set_current_window(window);
	swift_set_window_size(window, SWIFT_VEC2F(width, height));
	glViewport( 0, 0, width, height);
	window->proj_mat = swift_mat4f_ortho(0, width, 0, height, -1, 1);
	window->screen_box = SWIFT_VEC4F(0, 0, width, height);
	swift_set_clip_box(window, window->screen_box);
}

void swift_set_windows_title(swift_window_t *window, const char *title)
{
	glfwSetWindowTitle(window->screen, title);
}

swift_vec2f_t swift_get_window_pos(swift_window_t *window)
{
	int x, y;

	glfwGetWindowPos(window->screen, &x, &y);

	return SWIFT_VEC2F(x, y);
}

void swift_set_window_pos(swift_window_t *window, swift_vec2f_t pos)
{
	glfwSetWindowPos(window->screen, pos.xy.x, pos.xy.y);
}

swift_vec2f_t swift_get_window_size(swift_window_t *window)
{
	int w, h;

	glfwGetWindowSize(window->screen, &w, &h);

	return SWIFT_VEC2F(w, h);
}

void swift_set_window_size(swift_window_t *window, swift_vec2f_t size)
{
	glfwSetWindowSize(window->screen, size.wh.w, size.wh.h);
}

void swift_iconify(swift_window_t *window)
{
	glfwIconifyWindow(window->screen);
}

void swift_restore(swift_window_t *window)
{
	glfwRestoreWindow(window->screen);
}

void swift_show_window(swift_window_t *window)
{
	glfwShowWindow(window->screen);
}

void swift_hide_window(swift_window_t *window)
{
	glfwHideWindow(window->screen);
}

void swift_set_user_data(swift_window_t *window, void *data)
{
	if (!data)
		return;

	window->user_data = data;
}

void *swift_get_user_data(swift_window_t *window)
{
	return window->user_data;
}

void swift_poll_events(void)
{
	glfwPollEvents();
}

void swift_post_empty_event(void)
{
	glfwPostEmptyEvent();
}

void swift_event_wait(void)
{
	glfwWaitEvents();
}

int swift_get_input_mode(swift_window_t* window, int mode)
{
	return glfwGetInputMode(window->screen, mode);
}

void swift_set_input_mode(swift_window_t* window, int mode, int value)
{
	glfwSetInputMode(window->screen, mode, value);
}

int swift_get_key(swift_window_t* window, int key)
{
	return glfwGetKey(window->screen, key);
}

int swift_get_mouse_button(swift_window_t* window, int button)
{
	return glfwGetMouseButton(window->screen, button);
}

swift_vec2f_t swift_get_cursor_pos(swift_window_t *window)
{
	double x, y;

	glfwGetCursorPos(window->screen, &x, &y);

	return SWIFT_VEC2F(x, y);
}

void swift_setcursor_pos(swift_window_t *window, swift_vec2f_t pos)
{
	glfwSetCursorPos(window->screen, pos.xy.x, pos.xy.y);
}

void swift_swap_buffers(swift_window_t *window)
{
	glfwSwapBuffers(window->screen);
}

void swift_swap_intervals(swift_window_t *window, int interval)
{
	swift_set_current_window(window);

	glfwSwapInterval(interval);
}

bool swift_is_window_open(swift_window_t *window)
{
	if (!window || !window->screen || glfwWindowShouldClose(window->screen))
		return false;

	return true;
}

void swift_should_close_window(swift_window_t *window, bool close)
{
	glfwSetWindowShouldClose(window->screen, close);
}

swift_vec2f_t swift_get_screen_size(void)
{
    const GLFWvidmode *mode = glfwGetVideoMode(glfwGetPrimaryMonitor());

	return SWIFT_VEC2F(mode->width, mode->height);
}

void swift_center_window(swift_window_t *window)
{
    swift_vec2f_t scrn = swift_get_screen_size();
	swift_vec2f_t win = swift_get_window_size(window);

	swift_set_window_pos(window, SWIFT_VEC2F(scrn.xy.x / 2 - win.xy.x / 2,
		scrn.xy.y / 2 - win.xy.y / 2));
}
